Package: khaos
Type: Package
Title: Bayesian Sparse Polynomial Chaos Expansion
Version: 2.1.1
Authors@R: c(
    person(given = "Kellin",
           family = "Rumsey",
           role = c("aut", "cre"),
           email = "knrumsey@lanl.gov",
           comment = c(ORCID = "0000-0002-2989-965X")),
    person(given = "J. Derek", 
           family = "Tucker", 
           role = "ctb", 
           email = "jdtuck@sandia.gov", 
           comment = c(ORCID = "0000-0001-8844-2169")))
Description: Implements Bayesian polynomial chaos expansion methods for
    surrogate modeling, uncertainty quantification, and sensitivity
    analysis. Includes sparse regression-based approaches and adaptive
    Bayesian models based on reversible-jump Markov chain Monte Carlo.
    Optional screening and basis-enrichment strategies are provided to
    improve scalability in moderate to high dimensions.
Imports: glmnet, mvtnorm
Suggests: knitr, rmarkdown, lhs, testthat (>= 2.1.0)
License: BSD_3_clause + file LICENSE
Encoding: UTF-8
NeedsCompilation: no
RoxygenNote: 7.3.2
Packaged: 2026-04-22 15:22:43 UTC; knrumsey
Author: Kellin Rumsey [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-2989-965X>),
  J. Derek Tucker [ctb] (ORCID: <https://orcid.org/0000-0001-8844-2169>)
Maintainer: Kellin Rumsey <knrumsey@lanl.gov>
Repository: CRAN
Date/Publication: 2026-04-23 20:30:03 UTC
Built: R 4.5.2; ; 2026-04-23 20:58:57 UTC; unix

Package: khaos
Type: Package
Title: Bayesian Sparse Polynomial Chaos Expansion
Version: 2.1.1
Authors@R: c(
    person(given = "Kellin",
           family = "Rumsey",
           role = c("aut", "cre"),
           email = "knrumsey@lanl.gov",
           comment = c(ORCID = "0000-0002-2989-965X")),
    person(given = "J. Derek", 
           family = "Tucker", 
           role = "ctb", 
           email = "jdtuck@sandia.gov", 
           comment = c(ORCID = "0000-0001-8844-2169")))
Description: Implements Bayesian polynomial chaos expansion methods for
    surrogate modeling, uncertainty quantification, and sensitivity
    analysis. Includes sparse regression-based approaches and adaptive
    Bayesian models based on reversible-jump Markov chain Monte Carlo.
    Optional screening and basis-enrichment strategies are provided to
    improve scalability in moderate to high dimensions.
Imports: glmnet, mvtnorm
Suggests: knitr, rmarkdown, lhs, testthat (>= 2.1.0)
License: BSD_3_clause + file LICENSE
Encoding: UTF-8
NeedsCompilation: no
RoxygenNote: 7.3.2
Packaged: 2026-04-22 15:22:43 UTC; knrumsey
Author: Kellin Rumsey [aut, cre] (ORCID:
    <https://orcid.org/0000-0002-2989-965X>),
  J. Derek Tucker [ctb] (ORCID: <https://orcid.org/0000-0001-8844-2169>)
Maintainer: Kellin Rumsey <knrumsey@lanl.gov>
Repository: CRAN
Date/Publication: 2026-04-23 20:30:03 UTC
Built: R 4.6.0; ; 2026-04-26 08:20:24 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(plot,adaptive_khaos)
S3method(plot,ordinal_khaos)
S3method(plot,sparse_khaos)
S3method(predict,adaptive_khaos)
S3method(predict,ordinal_khaos)
S3method(predict,sparse_khaos)
S3method(print,sparse_khaos)
export(adaptive_khaos)
export(adaptive_khaos2)
export(adaptive_khaos_gprior)
export(adaptive_khaos_ridge)
export(ordinal_khaos)
export(sobol_khaos)
export(sparse_khaos)
importFrom(graphics,abline)
importFrom(graphics,axis)
importFrom(graphics,barplot)
importFrom(graphics,box)
importFrom(graphics,boxplot)
importFrom(graphics,hist)
importFrom(graphics,legend)
importFrom(graphics,matplot)
importFrom(graphics,par)
importFrom(stats,dgamma)
importFrom(stats,pnorm)
importFrom(stats,predict)
importFrom(stats,qnorm)
importFrom(stats,rbeta)
importFrom(stats,rgamma)
importFrom(stats,runif)
importFrom(stats,var)

#' Bayesian Sparse Polynomial Chaos Expansion
#'
#' A Bayesian sparse polynomial chaos expansion (PCE) method based on the
#' forward-selection framework of Shao et al. (2017), with optional enrichment
#' strategies that adapt the candidate basis set across degree and interaction-order
#' expansions.
#'
#' @param X A data frame or matrix of predictors scaled to lie between 0 and 1.
#' @param y A numeric response vector of length \code{nrow(X)}.
#' @param degree Integer vector of length 3 giving the polynomial-degree schedule:
#'   \code{c(d_init, d_increment, d_max)}.
#' @param order Integer vector of length 3 giving the interaction-order schedule:
#'   \code{c(q_init, q_increment, q_max)}.
#' @param prior Numeric vector of prior hyperparameters
#'   \code{c(n0, v0, s0)}, where \code{n0} controls shrinkage of the regression
#'   coefficients, \code{v0} is the degrees-of-freedom parameter for the inverse-gamma
#'   prior on \eqn{\sigma^2}, and \code{s0} is the scale parameter for that prior.
#' @param lambda Nonnegative penalty parameter used in the KIC calculation. Larger
#'   values favor sparser models.
#' @param rho Basis functions are discarded after partial-correlation screening if
#'   their squared partial correlation is smaller than \code{rho}.
#' @param regularize Logical; if \code{TRUE}, a weighted LASSO pre-screen is used
#'   to reduce the candidate basis set before partial-correlation ranking.
#' @param max_basis Maximum number of candidate basis functions allowed in any
#'   fitting round. This acts as a hard cap on computational cost.
#' @param enrichment Either an integer in \code{0:4} specifying a fixed enrichment
#'   strategy, or the character string \code{"adaptive"}, which allows the strategy
#'   to become more restrictive when the next candidate set would otherwise be too
#'   large. See Details.
#' @param adaptive_ladder Numeric vector of candidate-size thresholds used only when
#'   \code{enrichment = "adaptive"}. The ladder should have length between 1 and 4,
#'   be sorted in increasing order, and satisfy \code{max(adaptive_ladder) <= max_basis}.
#'   Internally, the ladder is used to determine when the algorithm should fall back
#'   to a more restrictive enrichment strategy before attempting the next recursion
#'   step.
#' @param verbose Logical; if \code{TRUE}, progress messages are printed.
#'
#' @details
#' For fixed maximum degree and interaction order, the method constructs a candidate
#' library of polynomial basis functions, ranks them using marginal correlations and
#' sequential partial correlations, and evaluates nested models using the Kashyap
#' information criterion (KIC). The best-ranked model is retained at that stage.
#'
#' If the selected model contains a basis function at the current maximum degree or
#' interaction order, the algorithm attempts another fitting round with increased
#' degree and/or interaction order. In later rounds, the candidate basis set can be
#' rebuilt using one of several enrichment strategies rather than constructing the
#' full basis library from scratch.
#'
#' The \code{enrichment} argument controls how the candidate set is generated after
#' the first fitting round:
#' \describe{
#'   \item{\code{0} (local enrichment)}{
#'     Starts from the previously selected basis functions and applies local
#'     modifications to each term. These include deletion, simplification,
#'     complication, and promotion moves. This is typically the fastest strategy,
#'     but it is also the most local.}
#'   \item{\code{1} (active-variable rebuild)}{
#'     Rebuilds the full candidate library using only variables that were active
#'     in the previously selected model. This is computationally efficient, but
#'     inactive variables cannot re-enter once dropped.}
#'   \item{\code{2} (full active rebuild plus sparse inactive enrichment)}{
#'     Rebuilds the candidate library on the currently active variables, then
#'     enriches around the previously selected basis functions by allowing inactive
#'     variables to re-enter through complication and promotion moves. This is a
#'     compromise between speed and flexibility.}
#'   \item{\code{3} (full active rebuild plus inactive enrichment)}{
#'     Rebuilds the candidate library on the active variables, then enriches that
#'     larger active-variable library by allowing inactive variables to re-enter
#'     through complication and promotion moves. This is more expansive than
#'     \code{enrichment = 2}, but can be substantially more expensive.}
#'   \item{\code{4} (full rebuild)}{
#'     Rebuilds the full candidate library over all variables at each enrichment
#'     step. This is the most exhaustive strategy.}
#'   \item{\code{"adaptive"}}{
#'     Begins with a relatively expansive enrichment strategy and, before the next
#'     recursion step, uses an upper bound on the candidate-set size to decide
#'     whether a more restrictive strategy should be used instead. This recovers
#'     the computational benefit of avoiding construction of candidate sets that
#'     are known in advance to be too large.}
#' }
#'
#' When \code{enrichment = "adaptive"}, the function uses \code{adaptive_ladder}
#' together with an upper bound on the next candidate-set size. If the current
#' enrichment choice appears too expensive, the algorithm steps down to a more
#' restrictive enrichment rule before proceeding. If even the most restrictive
#' admissible next step is predicted to exceed the allowed budget, the current
#' fitted model is returned.
#'
#' Strategies \code{2}, \code{3}, and \code{"adaptive"} are designed to address a
#' limitation of active-variable-only rebuilds: variables that are inactive in one
#' round are allowed to re-enter the model in later rounds.
#'
#' @return An object of class \code{"sparse_khaos"}.
#' The object is a list containing:
#' \describe{
#'   \item{B}{Matrix of selected basis functions evaluated at the training inputs.}
#'   \item{nbasis}{Number of selected basis functions (including intercept).}
#'   \item{vars}{Matrix encoding the polynomial multi-indices for each basis function.}
#'   \item{beta_hat}{MAP estimates of the regression coefficients.}
#'   \item{s2}{Estimated noise variance.}
#'   \item{X, y}{Training data (with \code{y} on the original scale).}
#'   \item{mu_y, sigma_y}{Centering and scaling constants used for \code{y}.}
#'   \item{BtB, BtBi}{Cross-product matrix and its inverse.}
#'   \item{G}{Prior scaling factors for coefficients.}
#'   \item{prior}{Prior hyperparameters.}
#'   \item{KIC}{Kashyap information criterion values for candidate models.}
#' }
#'
#' @references
#' Shao, Q., Younes, A., Fahs, M., and Mara, T. A. (2017).
#' Bayesian sparse polynomial chaos expansion for global sensitivity analysis.
#' \emph{Computer Methods in Applied Mechanics and Engineering}, \strong{318},
#' 474--496.
#'
#' Rumsey, K. N., Francom, D., Gibson, G., Tucker, J. D., and Huerta, G. (2026).
#' Bayesian adaptive polynomial chaos expansions.
#' \emph{Stat}, \strong{15}(1), e70151.
#'
#' @examples
#' X <- lhs::maximinLHS(100, 2)
#' f <- function(x) 10.391 * ((x[1] - 0.4) * (x[2] - 0.6) + 0.36)
#' y <- apply(X, 1, f) + stats::rnorm(100, 0, 0.1)
#'
#' fit <- sparse_khaos(X, y)
#'
#' @export
sparse_khaos <- function(X, y,
                         degree=c(2,2,16), order=c(1,1,5),
                         prior=c(1, 1, 1), lambda=1, rho=0, regularize=TRUE,
                         max_basis=1e6,
                         enrichment="adaptive",
                         adaptive_ladder = c(1e4, 1e5),
                         verbose=TRUE){
  if(max(X) > 1 || min(X) < 0){
    warning("X matrix should have values between 0 and 1.\n")
  }
  if(!is.matrix(X)){
    if(is.data.frame(X)){
      X <- as.matrix(X)
    }else{
      # Assuming it's a 1-d vector, try to coerce
      if(is.numeric(X)){
        X <- matrix(X, ncol=1)
      }else{
        stop("X should be a matrix, data.frame, or vector.")
      }
    }
  }
  n <- nrow(X)
  p <- ncol(X)

  d_curr <- degree[1]
  d_inc  <- degree[2]
  d_max  <- degree[3]

  o_curr <- order[1]
  o_inc  <- order[2]
  o_max  <- min(p, order[3])

  mu_y <- mean(y)
  sig_y <- stats::sd(y)
  y <- (y - mu_y)/sig_y

  if(enrichment == "adaptive"){
    # check ladder for validity
    if(length(adaptive_ladder) > 4 || length(adaptive_ladder) < 1){
      stop("adaptive_ladder has the wrong length, see documentation")
    }
    if(max(adaptive_ladder) > max_basis){
      stop("adaptive_ladder cannot exceed max_basis, see documentation")
    }
    enrichment <- length(adaptive_ladder)
    adaptive_ladder <- c(sort(adaptive_ladder), max_basis)
  }else{
    adaptive_ladder = NA
  }
  if(!(enrichment %in% 0:4)){
    stop("enrichment must be in 0:4")
  }

  A_curr <- NULL
  res <- sparse_khaos_wrapper(X, y, n, p, d_curr, d_inc, d_max, o_curr, o_inc, o_max, mu_y, sig_y, max_basis, prior, lambda, rho, regularize, verbose, enrichment, adaptive_ladder, A_curr)
  return(res)
}

sparse_khaos_wrapper <- function(X, y, n, p, d_curr, d_inc, d_max, o_curr, o_inc, o_max, mu_y, sig_y, max_basis, prior, lambda, rho, regularize, verbose, enrichment, adaptive_ladder, A_curr){
  # Create a list of p sequences from 1 to n
  if(verbose){
    cat("Starting model with max degree = ", d_curr, " and max order = ", o_curr, "\n", sep="")
  }

  if(is.null(A_curr)){
    A_num <- A_size(p, d_curr, o_curr)
    if(verbose) cat("\tFound ", A_num, " possible basis functions.\n", sep="")
    if(A_num > max_basis){
      stop("Too many basis functions. Increase max_basis or decrease initial degree/order. Consider dimension reduction approaches?")
    }
    if(verbose){
      if(A_num > 1000){
        t_est <- 4.258369e-7 * A_num ^ 1.781556
        if(t_est > 60){
          cat("\tComputing initial phi matrix (~", round(t_est/60, 2), " minutes)\n", sep = "")
        }else{
          cat("\tComputing initial phi matrix (~", round(t_est, 2), " seconds)\n", sep = "")
        }
      }else{
        cat("\tComputing initial phi matrix\n")
      }
    }
    A_set <- generate_A(p, d_curr, o_curr)
  }else{
    if(verbose){
      cat("\tBuilding enriched candidate set using strategy ", enrichment, "\n", sep = "")
    }
    A_set <- build_candidate_set(p, d_curr, o_curr, A_curr, enrichment, max_basis)
    A_num <- nrow(A_set)
    if(verbose) cat("\tFound ", A_num, " candidate basis functions after enrichment.\n", sep = "")
    if(A_num > max_basis){
      stop("Too many basis functions after enrichment. Increase max_basis or use a more restrictive enrichment strategy.")
    }
  }

  A_deg <- rowSums(A_set)
  A_ord <- rowSums(A_set > 0)
  N_alpha <- nrow(A_set)

  phi <- matrix(NA, nrow = n, ncol = N_alpha)
  rr <- rep(NA, N_alpha)
  for(i in 1:N_alpha){
    curr <- rep(1, n)
    for(j in 1:p){
      curr <- curr * ss_legendre_poly(X[,j], A_set[i, j])
    }
    phi[, i] <- curr
    rr[i] <- stats::cor(curr, y)
  }

  ord <- rev(order(rr^2))
  A_set <- A_set[ord, , drop = FALSE]
  A_deg <- A_deg[ord]
  A_ord <- A_ord[ord]
  phi <- phi[, ord, drop = FALSE]
  rr <- rr[ord]

  # Do LASSO?
  if(regularize){
    if(verbose) cat("\tRunning weighted LASSO...", sep="")
    lfit <- glmnet::glmnet(phi, y, penalty.factor=1-rr^2, dfmax=n-2)
    count_nonzero <- apply(lfit$beta, 2, function(bb) sum(bb != 0))
    valid_lambda <- which(count_nonzero < n)
    lambda_indx <- if(length(valid_lambda)) max(valid_lambda) else length(count_nonzero)
    alpha_indx <- which(lfit$beta[,lambda_indx] != 0)
    if(length(alpha_indx) == 0) alpha_indx <- 1
    N_alpha <- length(alpha_indx)
    A_set <- A_set[alpha_indx,,drop=FALSE]
    A_deg <- A_deg[alpha_indx]
    A_ord <- A_ord[alpha_indx]
    phi <- phi[,alpha_indx,drop=FALSE]
    rr <- rr[alpha_indx]
    if(verbose) cat(" Keeping ", N_alpha, " basis functions\n", sep="")
  }

  # Get partial correlation coefficients
  if(verbose) cat("\tComputing partial correlation coefficients\n")
  if(verbose) cat("\t\tFitting linear models: 0/", N_alpha, ", ", sep="")
  for(i in 2:N_alpha){
    if(N_alpha > 20 && ((i %% round(N_alpha/5)) == 0)){
      if(verbose) cat(i, "/", N_alpha, ", ",sep="")
    }
    eps_y <- stats::lm(y ~ phi[,1:(i-1)])$residuals
    eps_p <- stats::lm(phi[,i] ~ phi[,1:(i-1)])$residuals
    rr[i] <- stats::cor(eps_y, eps_p)
  }

  if(any(abs(rr) > 1)){
    rr <- rr/max(abs(rr))
  }
  ord <- rev(order(rr^2))
  A_set <- A_set[ord,,drop=FALSE]
  A_deg <- A_deg[ord]
  A_ord <- A_ord[ord]
  phi <- phi[,ord]
  rr <- rr[ord]

  # Get KIC for various models
  if(verbose) cat("\n\tRanking models based on KIC\n")
  # Add coefficient column in
  phi <- cbind(rep(1, n), phi)
  A_set <- rbind(rep(0, p), A_set)
  K_trunc <- 1 + max(c(0, which(rr^2 >= rho)))

  if(verbose){
    if(rho > 0) cat("\t\t Throwing out bases with low partial correlation. ", K_trunc, " remain.\n", sep="")
  }

  # Get g prior info
  Gdiag <- c(1, 1/(1 + A_ord*(A_ord + A_deg - 2)))
  n0 <- prior[1]
  G1 <- n*Gdiag/(n*Gdiag + n0)
  G2 <- n0/(n0 + n*Gdiag)

  # Get current BtB and BtBinv values (B is the phi matrix)
  B    <- phi[,1,drop=FALSE]
  BtB  <- crossprod(phi[,1,drop=FALSE])
  BtBi <- solve(BtB)
  Bty  <- crossprod(phi[,1,drop=FALSE], y)

  # Get MAP estimates
  a_mle  <- BtBi%*%Bty
  y_mle  <- B%*%a_mle
  s2_mle <- crossprod(y - y_mle)
  a_map  <- a_mle*G1[1]
  peff   <- 1
  s2_map <- (s2_mle + prior[3]^2 + G2[1]*crossprod(y_mle))/(n + 2 + prior[2] - 1)

  # Evaluate KIC for intercept-only model
  Sigma_prior <- sqrt(diag(BtBi) * Gdiag[1] * prior[3]^2 / prior[2] * n / n0)
  Sigma_post  <- (G1[1] * BtBi) * as.numeric(s2_mle)

  KIC <- rep(NA, K_trunc)
  KIC[1] <- -2 * sum(stats::dnorm(y, y_mle, sqrt(s2_map), log = TRUE)) -
    2 * sum(stats::dnorm(as.numeric(a_map), 0, Sigma_prior, log = TRUE)) -
    lambda * (peff + 1) * log(2 * pi) + log(det(Sigma_post))

  best <- list(
    k     = 1,
    KIC   = KIC[1],
    coeff = a_map,
    s2    = s2_map,
    BtB   = BtB,
    BtBi  = BtBi,
    G     = Gdiag[1]
  )
  if(K_trunc >= 2){
    for(k in 2:K_trunc){
      # Update BtB quantities
      B    <- phi[,1:k,drop=FALSE]
      bnew <- phi[,k,drop=FALSE]
      BtB_curr <- update_BtB(B, BtB, BtBi, bnew)
      BtB  <- BtB_curr$BtB
      BtBi <- BtB_curr$BtBi
      Bty  <- rbind(Bty, sum(bnew*y))

      # Get MAP estimates
      a_mle  <- BtBi%*%Bty
      y_mle  <- B%*%a_mle
      s2_mle <- crossprod(y - y_mle)
      a_map  <- a_mle*G1[1:k]
      y_reg  <- B%*%(a_mle * sqrt(G2[1:k]))
      peff   <- k
      s2_map <- (s2_mle + prior[3]^2 + crossprod(y_reg))/(n + 2 + prior[2] - peff)

      # Evaluate KIC
      yhat_k <- B%*%a_map
      Sigma_prior <- sqrt(diag(BtBi)*Gdiag[1:k]*prior[3]^2/prior[2]*n/n0) # Take the diagonal for simplicity
      Sigma_post  <- (G1[1:k] * BtBi) * as.numeric(s2_mle) # equivalent to (diag(G1) %*% BtBi)

      KIC[k] <- -2*sum(stats::dnorm(y, yhat_k, sqrt(s2_map), log=TRUE)) -
        2*sum(stats::dnorm(as.numeric(a_map), 0, Sigma_prior, log=TRUE)) -
        lambda*(k+1)*log(2*pi) + log(det(Sigma_post))

      if(KIC[k] <= best$KIC){
        best$k     <- k
        best$KIC   <- KIC[k]
        best$coeff <- a_map
        best$s2    <- s2_map
        best$BtB   <- BtB
        best$BtBi  <- BtBi
        best$G     <- Gdiag[1:k]
      }
    }
  }

  if(best$k <= 1){
    max_degree_curr <- 0
    max_order_curr <- 0
  } else {
    max_degree_curr <- max(A_deg[1:(best$k-1)])
    max_order_curr <- max(A_ord[1:(best$k-1)])
  }
  at_capacity_degree  <- (max_degree_curr == d_curr) #Note, Shao et al. suggest a +1 on the LHS of ineq.
  at_capacity_order   <- (max_order_curr == o_curr)
  over_max_degree     <- (d_curr >= d_max)
  over_max_order      <- (o_curr >= o_max)
  found_final_model   <- (!at_capacity_degree | over_max_degree) & (!at_capacity_order | over_max_order) # Neither degree nor order is at capacity (unless they're at the max)
  order_bigger_than_p <- (o_curr + o_inc > p) & (d_inc == 0)
  over_max_model      <- over_max_degree & over_max_order
  # Check whether the next candidate set would be too large.
  # For adaptive enrichment, try decreasing enrichment before giving up.
  d_next <- min(d_curr + d_inc, d_max)
  o_next <- min(o_curr + o_inc, o_max)

  if(best$k > 1){
    A_next <- A_set[2:best$k, , drop = FALSE]
  } else {
    A_next <- matrix(0, nrow = 0, ncol = p)
  }

  enrichment_next <- enrichment
  next_basis_too_big <- FALSE

  if(!all(is.na(adaptive_ladder))){
    p_active <- sum(colSums(A_next) > 0)
    A_next_num <- nrow(A_next)

    A_num_ub <- A_size_ub(p, d_next, o_next, p_active, A_next_num, enrichment_next)

    # adaptive_ladder has already had max_basis appended to the end
    # Example: c(1e4, 1e5, max_basis)
    # Then enrichment 2 -> 1e4, enrichment 1 -> 1e5, enrichment 0 -> max_basis
    while(enrichment_next > 0 &&
          A_num_ub > adaptive_ladder[length(adaptive_ladder) - enrichment_next]){
      enrichment_next <- enrichment_next - 1
      A_num_ub <- A_size_ub(p, d_next, o_next, p_active, A_next_num, enrichment_next)
    }

    if(A_num_ub > adaptive_ladder[length(adaptive_ladder) - enrichment_next]){
      next_basis_too_big <- TRUE
    }

    if(verbose && enrichment_next < enrichment){
      cat("\tAdaptive enrichment: reducing from ",
          enrichment, " to ", enrichment_next,
          " (new estimated size = ", format(A_num_ub, scientific = TRUE),
          ")\n", sep = "")
    }
  } else {
    p_active <- sum(colSums(A_next) > 0)
    A_next_num <- nrow(A_next)
    A_num_ub <- A_size_ub(p, d_next, o_next, p_active, A_next_num, enrichment_next)
    next_basis_too_big <- (A_num_ub > max_basis)
  }
  if(found_final_model | over_max_model | order_bigger_than_p | next_basis_too_big){
    if(verbose){
      if(next_basis_too_big) cat("Note: max_basis was reached. Consider dimension reduction?\n")
      if(over_max_model) cat("Note: Maximum degree and order were both reached. Consider increasing these values?\n")
    }
    obj <- list(B        = phi[, 1:best$k, drop = FALSE],
                nbasis   = best$k,
                vars     = A_set[1:best$k, , drop = FALSE],
                s2       = best$s2,
                beta_hat = best$coeff,

                X       = X,
                y       = y * sig_y + mu_y,

                mu_y    = mu_y,
                sigma_y = sig_y,

                BtB     = best$BtB,
                BtBi    = best$BtBi,
                G       = best$G,

                prior   = prior,

                KIC = KIC[1:best$k])

    class(obj) <- "sparse_khaos"
  }else{
    res <- sparse_khaos_wrapper(X, y, n, p, d_next, d_inc, d_max, o_next, o_inc, o_max, mu_y, sig_y, max_basis, prior, lambda, rho, regularize, verbose, enrichment_next, adaptive_ladder, A_next)
    return(res)
  }
  return(obj)
}

#' Predict Method for class sparse_khaos
#'
#' See \code{sparse_khaos()} for details.
#'
#' @param object An object returned by the \code{sparse_khaos()} function.
#' @param newdata A dataframe of the same dimension as the training data.
#' @param samples How many posterior samples should be taken at each test point? If 0 or FALSE, then the MAP estimate is returned.
#' @param nugget logical; should predictions include error? If FALSE, predictions will be for mean.
#' @param ... Additional arguments to predict
#' @details Predict function for sparse_khaos object.
#' @return A numeric matrix of predictive samples (or a vector if \code{samples = FALSE}),
#' with rows corresponding to samples and columns corresponding to observations.
#' @references Shao, Q., Younes, A., Fahs, M., & Mara, T. A. (2017). Bayesian sparse polynomial chaos expansion for global sensitivity analysis. Computer Methods in Applied Mechanics and Engineering, 318, 474-496.
#' @examples
#' X <- lhs::maximinLHS(100, 2)
#' f <- function(x) 10.391*((x[1]-0.4)*(x[2]-0.6) + 0.36)
#' y <- apply(X, 1, f) + stats::rnorm(100, 0, 0.1)
#' fit <- sparse_khaos(X, y)
#' predict(fit)
#'
#' @export
predict.sparse_khaos <- function(object, newdata=NULL, samples=1000, nugget=FALSE, ...){
  if(is.null(newdata)){
    newdata <- object$X
  }
  XX <- newdata
  n <- nrow(XX)
  p <- ncol(XX)
  N_alpha <- object$nbasis
  phi <- matrix(NA, nrow=n, ncol=N_alpha)
  for(i in 1:N_alpha){
    curr <- rep(1, n)
    for(j in 1:p){
      curr <- curr * ss_legendre_poly(XX[,j], object$vars[i,j])
    }
    phi[,i] <- curr
  }

  n0 <- object$prior[1]
  v0 <- object$prior[2]
  s0 <- object$prior[3]
  a_hat <- object$beta_hat
  peff <- length(a_hat)
  ntrain <- length(object$y)
  if(samples){
    pred <- matrix(NA, nrow=samples, ncol=n)
    for(i in 1:samples){
      shape <-  (ntrain + v0 - peff)/2
      sigma2 <- 1/stats::rgamma(1, shape=shape, rate=object$s2*shape)
      a_Sigma <- sigma2 * (object$G * object$BtBi)
      coeff <- stats::rnorm(peff, a_hat, sqrt(diag(a_Sigma)))
      y_hat <- phi%*%coeff
      if(nugget){
        y_hat <- y_hat + stats::rnorm(n, 0, sqrt(sigma2))
      }
      pred[i,] <- y_hat
    }
  }else{
    pred <- phi%*%object$beta_hat
  }
  pred <- object$mu_y + object$sigma_y * pred
  return(pred)
}

#' Plot Method for class sparse_khaos
#'
#' See \code{sparse_khaos()} for details.
#'
#' @param x An object returned by the \code{sparse_khaos()} function.
#' @param ... additional arguments passed to \code{plot}
#' @details Plot function for sparse_khaos object.
#' @return returns invisible(x)
#' @references Shao, Q., Younes, A., Fahs, M., & Mara, T. A. (2017). Bayesian sparse polynomial chaos expansion for global sensitivity analysis. Computer Methods in Applied Mechanics and Engineering, 318, 474-496.
#' @examples
#' X <- lhs::maximinLHS(100, 2)
#' f <- function(x) 10.391*((x[1]-0.4)*(x[2]-0.6) + 0.36)
#' y <- apply(X, 1, f) + stats::rnorm(100, 0, 0.1)
#' fit <- sparse_khaos(X, y)
#' plot(fit)
#' @export
plot.sparse_khaos <- function(x, ...){
  dots <- list(...)

  nugget <- if("nugget" %in% names(dots)) dots$nugget else FALSE
  samples <- if("samples" %in% names(dots)) dots$samples else 1000

  dots$nugget <- NULL
  dots$samples <- NULL

  pred <- predict(x, x$X, samples = samples, nugget = nugget)
  yhat <- colMeans(pred)

  do.call(graphics::plot, c(list(x = x$y, y = yhat, xlab="observed", ylab="predicted"), dots))
  graphics::abline(0, 1, lwd = 2, col = "orange")

  ci <- apply(pred, 2, function(yy) stats::quantile(yy, c(0.025, 0.975)))
  for(i in 1:ncol(ci)){
    graphics::segments(x$y[i], ci[1, i], x$y[i], ci[2, i])
  }
}

#' Print Method for class sparse_khaos
#'
#' See \code{sparse_khaos()} for details.
#'
#' @param x An object returned by the \code{sparse_khaos()} function.
#' @param ... additional arguments passed to or from other methods.
#' @details Print function for sparse_khaos object.
#' @return The input object \code{x}, invisibly.
#' @references Shao, Q., Younes, A., Fahs, M., & Mara, T. A. (2017). Bayesian sparse polynomial chaos expansion for global sensitivity analysis. Computer Methods in Applied Mechanics and Engineering, 318, 474-496.
#' @examples
#' X <- lhs::maximinLHS(100, 2)
#' f <- function(x) 10.391*((x[1]-0.4)*(x[2]-0.6) + 0.36)
#' y <- apply(X, 1, f) + stats::rnorm(100, 0, 0.1)
#' fit <- sparse_khaos(X, y)
#' print(fit)
#' @export
print.sparse_khaos <- function(x, ...){
  vars_used <- colSums(x$vars)
  names(vars_used) <- paste("V", seq_along(vars_used), sep="")

  bf_degrees <- table(rowSums(x$vars))
  names(bf_degrees) <- paste("deg", seq_along(bf_degrees), sep="")

  bf_orders <- table(rowSums(x$vars > 0))
  names(bf_orders) <- paste("ord", seq_along(bf_orders), sep="")

  cat("Fitted model contains", x$nbasis, "basis functions\n\n")
  cat("KIC =", round(min(x$KIC), 3), "\n\n")

  cat("Variables used:\n")
  print(vars_used)

  cat("\nBasis function degrees:\n")
  print(bf_degrees)

  cat("\nBasis function orders:\n")
  print(bf_orders)

  invisible(x)
}

legendre_poly <- function(x, j){
  if(j == 0){
    return(rep(1, length(x)))
  }
  if(j == 1){
    return(x)
  }
  n <- j - 1
  res <- ((2*n+1)*x*legendre_poly(x, n) - n*legendre_poly(x, n-1))/(n+1)
  return(res)
}

ss_legendre_poly <- function(x, j){
  sqrt(2*j+1)*legendre_poly(2*x-1, j)
}

generate_A <- function(p, d, q) {
  # Initialize the result matrix
  res <- NULL
  #res <- list()

  # Recursive function
  generate <- function(current_set, current_sum, current_nonzero, index){
    if (current_sum > d || current_nonzero > q) return()
    if (index > p){
      if(current_nonzero > 0){
        res <<- rbind(res, current_set)
        #res <- c(res, list(current_set))
      }
      return()
    }

    for (value in 0:(d - current_sum)) {
      new_set <- current_set
      new_set[index] <- value
      generate(new_set,
               current_sum + value,
               current_nonzero + as.numeric(value > 0),
               index + 1)
    }
  }

  # Start the generation process
  generate(integer(p), 0, 0, 1)
  return(res)
}

update_BtB <- function(B, BtB, BtBi, bnew){
  k <- nrow(BtB)
  BtB_new <- BtBi_new <- matrix(NA, nrow=k+1, ncol=k+1)

  # Update BtB
  Btb <- t(B[,1:k])%*%as.matrix(bnew, ncol=1)
  BtB_new[1:k, 1:k] <- BtB
  BtB_new[k+1,1:k]  <- Btb
  BtB_new[1:k,k+1]  <- Btb
  BtB_new[k+1,k+1]  <- sum(bnew^2)

  # Update BtB inverse
  A <- BtBi%*%Btb
  S <- as.numeric(sum(bnew^2) - t(Btb)%*%A)
  BtBi_new[1:k,1:k] <- BtBi + tcrossprod(A)/S
  BtBi_new[1:k,k+1] <- -A/S
  BtBi_new[k+1,1:k] <- -A/S
  BtBi_new[k+1,k+1] <- 1/S

  out <- list(BtB=BtB_new, BtBi=BtBi_new)
  return(out)
}



estimate_map <- function(y, phi, Cinv, sig0=1, tol=1e-3, max_iter=1000){
  n <- length(y)
  p <- ncol(phi)
  sig_curr <- sig0
  Cinv <- Cinv
  flag <- TRUE
  cnt <- 1
  while(flag){
    Cinv_curr <-  crossprod(phi)/sig_curr^2 + Cinv
    tmp <- tryCatch({
      solve(Cinv_curr)
    }, error=function(e){ NULL })
    if(is.null(tmp)){
      #Cinv_curr <- as.matrix(Matrix::nearPD(Cinv_curr)$mat)
      stop("Matrix can't be inverted. Is regularize = TRUE set?")
    }
    C_curr <- solve(Cinv_curr)
    a_curr <- (tcrossprod(C_curr, phi)%*%y)/sig_curr^2
    yhat   <- phi%*%a_curr
    sig_new <- sqrt(mean((y-yhat)^2))
    delta <- abs(sig_new-sig_curr)
    sig_curr <- sig_new

    if(delta < tol){
      flag <- FALSE
    }
    if(cnt >= max_iter){
      flag <- FALSE
    }
    cnt <- cnt + 1
  }
  out <- list(a=a_curr, sig=sig_curr, Caa_inv=Cinv_curr)
  return(out)
}


A_size <- function(p, d, q){
  res <- 0
  for(qq in 1:q){
    for(dd in 1:d){
      res <- res + choose(p, qq)*choose(dd-1, qq-1)
    }
  }
  return(res)
}

A_size_ub <- function(p, d, q, p_active, A_size_curr, enrichment){
  if(enrichment == 0){
    return(A_size_curr * (4 * p + 1))
  }
  if(enrichment == 1){
    return(A_size(p_active, d, q))
  }
  if(enrichment == 2){
    return(A_size(p_active, d, q) + 2 * A_size_curr * (p - p_active))
  }
  if(enrichment == 3){
    return(A_size(p_active, d, q) * (1 + 2 * (p - p_active)))
  }
  if(enrichment == 4){
    return(A_size(p, d, q))
  }
  stop("enrichment number not recognized")
}

# HELPERS FOR ADAPTIVE VERSION

make_weights <- function(eta, p0, epsilon, alpha, num_passes){
  p <- length(eta)
  v <- (eta^alpha + epsilon)/sum((eta^alpha + epsilon))*p0/p
  delta <- 0
  for(i in 1:num_passes){
    beta <- mean((log(p0 + delta) - log(p))/log(v))
    delta <- delta + p0 - sum(v^beta)
  }
  return(v^beta)
}

# Breaks d into q elements d_1, ... d_q >= 1 s.t d1+...+dq = d.
# Probability of getting any particular sequence is 1/choose(d-1, q-1)
random_partition <- function(d, q) {
  if (d < q) stop("d must be greater than or equal to q")
  partition <- rep(1, q)
  remaining <- d - q

  # Intuitive but slower
  #extra     <- sample(factor(1:q), remaining, replace=TRUE)
  #partition <- partition + as.numeric(table(extra))

  # Less intuitive but faster
  cuts <- sort(sample(0:remaining, q - 1, replace = TRUE))
  additions <- diff(c(0, cuts, remaining))
  partition <- partition + additions
  return(partition)
}


make_basis<-function(vv,dd,XX){ # function to make a basis function given vars and degs
  n <- nrow(XX)
  curr <- rep(1, n)
  for(j in seq_along(vv)){
    curr <- curr * ss_legendre_poly(XX[,vv[j]], dd[j])
  }
  return(curr)
}

# Function to sample from multivariate normal distribution
sample_mvn <- function(n, m, S) {
  p <- length(m)
  L <- chol(S)
  Z <- matrix(stats::rnorm(n * p), nrow = n)
  samples <- tcrossprod(Z, L) + matrix(m, nrow = n, ncol = length(m), byrow = TRUE)
  return(samples)
}

myTimestamp <-function(){
  x<-Sys.time()
  paste('#--',format(x,"%b %d %X"),'--#')
}

test_iid_unif <- function(X, Nsims=1000, alpha=0.01){
  # Test uniformity
  p_unif <- min(apply(X, 2, function(xx) stats::ks.test(xx, "punif")$p.value))

  # Test pairwise independence
  nn <- nrow(X)
  pp <- ncol(X)
  NN <- Nsims

  S0 <- diag(rep(1/12), pp)
  Sx <- stats::cov(X)
  Tx <- sum((S0-Sx)^2)
  T0 <- rep(NA, NN)
  for(ii in 1:NN){
    Xi <- matrix(stats::runif(nn*pp), nrow=nn, ncol=pp)
    Si <- stats::cov(Xi)
    T0[ii] <- sum((S0-Si)^2)
  }
  p_ind <- mean(T0 > Tx)


  T_obs <- norm(stats::cov(X) - (1/12) * diag(pp), type = "F")^2
  T_null <- replicate(NN, {
    norm(stats::cov(matrix(stats::runif(nn * pp), nn, pp)) - (1/12) * diag(pp), type = "F")^2
  })
  p_val <- mean(T_null > T_obs)  # one-sided

  if(min(p_ind, p_unif) < alpha){
    return(FALSE)
  }else{
    return(TRUE)
  }
}



rg0sq_laplace_orth <- function(n, a, b, g, n_iter=5, c=0){
  # Set initial value
  theta_u_init <- b/a

  # Iterate to find map
  for(i in seq_len(n_iter)){
    Gl <- 0.5 * sum(g^2/(1 + theta_u_init*g^2))
    theta_u_init <- max(0, (-a + sqrt(a^2 + 4*Gl*b) ) / (2 * Gl))
  }
  # Set the MAP
  m_theta <- theta_u_init

  # Estimate variance (with curvature)
  s2_theta <- 1 / (2*b/m_theta^3 - a/(2*m_theta^2) - 0.5*sum(g^4/(1 + m_theta*g^2)^2))

  # Get the inverse gamma parameters
  a_theta <- 2 + m_theta^2 / s2_theta
  b_theta <- m_theta * (a_theta + c) # c=-1 is moment matching, c=1 matches modes (technically not correct but gives better results in practice?)

  return(1/rgamma(n, a_theta, b_theta))
}

rg0sq_laplace_full <- function(n, a, b, g, BtB, n_iter=5, n_newtonsteps=3, c=0){
  M <- length(g)

  # Set initial value
  theta_u_init <- b/a

  # Iterate to find map
  for(i in seq_len(n_iter)){
    Gl <- 0.5 * sum(g^2/(1 + theta_u_init*g^2))
    theta_u_init <- max(0, (-a + sqrt(a^2 + 4*Gl*b) ) / (2 * Gl))
  }
  # Set the MAP
  m_theta <- theta_u_init

  # Define first and second derivative functions
  hh <- function(xx, a, b, M, PinvR){
    -(a + M/2)/xx + b/xx^2 + 1/(2*xx^2)*sum(diag(PinvR))
  }
  hhprime <- function(xx, a, b, M, PinvR, PinvR_sq){
    (a + M/2)/xx^2 - 2*b/xx^3 - 1/xx^3*sum(diag(PinvR)) + 1/(2*xx^4)*sum(diag(PinvR_sq))
  }

  # Start newton raphson (log-scale)
  theta_curr <- m_theta
  ggt <- tcrossprod(1/g)
  R0 <- ggt * BtB
  for(i in 1:n_newtonsteps){
    P_theta <- (matrix(1, nrow=M, ncol=M) + ggt/theta_curr) * BtB
    PinvR <- solve(P_theta)%*%R0
    PinvR_sq <- PinvR %*% PinvR

    h_curr <- hh(theta_curr, a, b, M, PinvR)
    hp_curr <- hhprime(theta_curr, a, b, M, PinvR, PinvR_sq)
    #step <- h_curr / (h_curr + theta_curr*hp_curr)
    #theta_curr <- exp(log(theta_curr) - step)

    step <- h_curr / (theta_curr * hp_curr)
    theta_curr <- theta_curr * exp(-step)
  }
  m_theta <- theta_curr

  # Estimate variance (with curvature)
  s2_theta <- -1 / hp_curr

  # Get the inverse gamma parameters
  a_theta <- 2 + m_theta^2 / s2_theta
  b_theta <- m_theta * (a_theta + c) # c=-1 is moment matching, c=1 matches modes (technically not correct but gives better results in practice?)

  return(1/rgamma(n, a_theta, b_theta))
}

dgsq_orth <- function(theta, a=1, b=1, gm_sq=NULL){
  if(is.null(gm_sq)){
    gm_sq <- 1
  }
  M <- length(gm_sq)

  res <- rep(0, length(theta))
  for(i in seq_along(theta)){
    term1 <- theta[i]^(-a-M/2)
    term2 <- exp(-b/theta[i])
    term3 <- sqrt(prod(theta[i]*gm_sq/(1+theta[i]*gm_sq)))
    if(term2 > 0){
      res[i] <- term1 * term2 * term3
    }
  }
  return(res)
}

dgsq_full <- function(theta, a=1, b=1, gm=NULL, BtB=NULL){
  if(is.null(gm)){
    gm <- rep(1, nrow(BtB))
  }
  M <- length(gm)

  res <- rep(0, length(theta))
  for(i in seq_along(theta)){
    term1 <- theta[i]^(-a-M/2)
    term2 <- exp(-b/theta[i])

    G <- matrix(1, nrow=M, ncol=M) + tcrossprod(1/gm) / theta[i]
    Sigma_inv <- G * BtB
    term3 <- det(Sigma_inv)^(-0.5)
    if(term2 > 0){
      res[i] <- term1 * term2 * term3
    }
  }
  return(res)
}

rg0sq_laplace_orth <- function(n, a, b, g, n_iter=5, c=0){
  # Set initial value
  theta_u_init <- b/a

  # Iterate to find map
  for(i in seq_len(n_iter)){
    Gl <- 0.5 * sum(g^2/(1 + theta_u_init*g^2))
    theta_u_init <- max(0, (-a + sqrt(a^2 + 4*Gl*b) ) / (2 * Gl))
  }
  # Set the MAP
  m_theta <- theta_u_init

  # Estimate variance (with curvature)
  s2_theta <- 1 / (2*b/m_theta^3 - a/(2*m_theta^2) - 0.5*sum(g^4/(1 + m_theta*g^2)^2))

  # Get the inverse gamma parameters
  a_theta <- 2 + m_theta^2 / s2_theta
  b_theta <- m_theta * (a_theta + c) # c=-1 is moment matching, c=1 matches modes (technically not correct but gives better results in practice?)

  if(n <= 0){
    return(list(a=a_theta, b=b_theta))
  }
  return(1/rgamma(n, a_theta, b_theta))
}

rg0sq_laplace_full <- function(n, a, b, g, BtB, n_iter=5, n_newtonsteps=3, c=0){
  M <- length(g)

  # Set initial value
  theta_u_init <- b/a

  # Iterate to find map
  for(i in seq_len(n_iter)){
    Gl <- 0.5 * sum(g^2/(1 + theta_u_init*g^2))
    theta_u_init <- max(0, (-a + sqrt(a^2 + 4*Gl*b) ) / (2 * Gl))
  }
  # Set the MAP
  m_theta <- theta_u_init

  # Define first and second derivative functions
  hh <- function(xx, a, b, M, PinvR){
    -(a + M/2)/xx + b/xx^2 + 1/(2*xx^2)*sum(diag(PinvR))
  }
  hhprime <- function(xx, a, b, M, PinvR, PinvR_sq){
    (a + M/2)/xx^2 - 2*b/xx^3 - 1/xx^3*sum(diag(PinvR)) + 1/(2*xx^4)*sum(diag(PinvR_sq))
  }

  # Start newton raphson (log-scale)
  theta_curr <- m_theta
  ggt <- tcrossprod(1/g)
  R0 <- ggt * BtB
  for(i in 1:n_newtonsteps){
    P_theta <- (matrix(1, nrow=M, ncol=M) + ggt/theta_curr) * BtB
    PinvR <- solve(P_theta)%*%R0
    PinvR_sq <- PinvR %*% PinvR

    h_curr <- hh(theta_curr, a, b, M, PinvR)
    hp_curr <- hhprime(theta_curr, a, b, M, PinvR, PinvR_sq)
    #step <- h_curr / (h_curr + theta_curr*hp_curr)
    #theta_curr <- exp(log(theta_curr) - step)

    step <- h_curr / (theta_curr * hp_curr)
    theta_curr <- theta_curr * exp(-step)
  }
  m_theta <- theta_curr

  # Estimate variance (with curvature)
  s2_theta <- -1 / hp_curr

  # Get the inverse gamma parameters
  a_theta <- 2 + m_theta^2 / s2_theta
  b_theta <- m_theta * (a_theta + c) # c=-1 is moment matching, c=1 matches modes (technically not correct but gives better results in practice?)

  if(n <= 0){
    return(list(a=a_theta, b=b_theta))
  }
  return(1/rgamma(n, a_theta, b_theta))
}

log_evidence_fast_core <- function(yty, v, g_diag, g0_sq, nu0, s0_sq, n){
  M <- length(v)
  logdet <- 0.5 * (sum(log(g0_sq*g_diag^2 / (g0_sq*g_diag^2 + 1))) - M*log(n))
  quad <- max(1e-16, 0.5 * (yty - sum(g0_sq*g_diag^2*v^2 / (g0_sq*g_diag^2 + 1)) / n))
  llhd <- -(nu0 + n)/2 * log(nu0*s0_sq/2 + quad)
  out <- logdet + llhd
}

log_evidence_full_core <- function(BtB, BtBi, Bty,
                                   g_diag, g0_sq,
                                   nu0, s0_sq, yty, n)
{
  #KR: CHANGE 1
  #G <- outer(rep(1, length(g_diag)), rep(1, length(g_diag))) +
  #          (1 / g0_sq) * tcrossprod(1 / sqrt(g_diag))
  G <- matrix(1, nrow=length(g_diag), ncol=length(g_diag)) + tcrossprod(1/g_diag) /g0_sq
  Sigma_i <- G * BtB
  L       <- chol(Sigma_i)
  logdet  <- -2 * sum(log(diag(L)))

  a_hat   <- BtBi %*% Bty
  resid2  <- yty - crossprod(a_hat, Bty)

  quad <- (resid2 + nu0 * s0_sq) / 2
  out  <- lgamma((nu0 + n)/2) - lgamma(nu0/2) -
    (nu0 + n)/2 * log(quad) +
    0.5 * logdet - n/2 * log(2*pi)


  as.numeric(out)
}

## ---------- public wrapper -----------------------------------
log_evidence <- function(mode = c("fast","full"),
                         yty, v = NULL,              # fast needs v
                         BtB = NULL, BtBi = NULL, Bty = NULL,  # full needs these
                         g_diag, g0_sq,
                         nu0, s0_sq, n){
  mode <- match.arg(mode)
  if (mode == "fast") {
    l_ev <- log_evidence_fast_core(yty, v, g_diag, g0_sq,
                           nu0, s0_sq, n)
  } else {
    l_ev <- log_evidence_full_core(BtB, BtBi, Bty,
                           g_diag, g0_sq,
                           nu0, s0_sq, yty, n)
  }
  if(is.na(l_ev)){
    return(-Inf)
  }
  return(l_ev)
}

#' Bayesian Adaptive Polynomial Chaos Expansion (Ridge Prior)
#'
#' Adaptive Bayesian polynomial chaos expansion using a fixed Gaussian (ridge)
#' prior on coefficients. This implementation corresponds to a simplified
#' version of the adaptive Bayesian PCE methodology described in Rumsey et al. (2026)
#'
#'
#' @param X A data frame or matrix of predictors scaled to be between 0 and 1
#' @param y a response vector
#' @param degree Maximum polynomial degree for each basis function.
#' @param order Maximum order of interaction for each basis function.
#' @param nmcmc Number of MCMC iterations.
#' @param nburn Number of initial MCMC iterations to discard.
#' @param thin Keep every \code{thin} samples.
#' @param max_basis Maximum number of basis functions.
#' @param tau2 Prior variance for coefficients
#' @param s2_lower Lower bound on process variance (numerically useful for deterministic functions).
#' @param g1,g2 Shape/rate parameters for the IG prior on process variance (default is Jefffrey's prior)
#' @param h1,h2 Shape/rate parameters for the Gamma prior on expected number of basis functions.
#' @param move_probs A 3-vector with probabilities for (i) birth, (ii) death, and (iii) mutation.
#' @param coin_pars A list of control parameters for coinflip proposal
#' @param degree_penalty Increasing this value encourages lower order polynomial terms (0 is no penalization).
#' @param legacy Logical. If TRUE, mimics original implementation behavior (from the emulator comparison paper),
#'        which may retain invalid basis function structures across iterations. Defaults to FALSE.
#' @param verbose Logical. Should progress information be printed?
#' @details Implements the RJMCMC algorithm described by Francom & Sanso (2020) for BMARS, modifying it for polynomial chaos basis functions.
#' As an alternative the NKD procedure of Nott et al. (2005), we use a coinflipping procedure to identify useful variables. See writeup (coming soon) for details.
#' The coin_pars argument is an ordered list with elements:
#' \enumerate{
#' \item{a function giving proposal probability for q_0, the expected order of interaction,}
#' \item{epsilon, the base weight for each variable (epsilon=Inf corresponds to uniform sampling),}
#' \item{alpha, the exponent-learning rate,}
#' \item{num_passes a numerical parameter only.}
#' }
#'
#' Other proposal notes:
#'    Expected order q0 is chosen from 1:order with weights `coin_pars[[1]](1:order)`
#'    Degree is chosen from q0:degree with weights `(1:(degree-q0+1))^(-degree_penalty)` and a random partition is created (see helper function).
#'
#' Two types of change steps are possible (and equally likely)
#'    A re-ordering of the degrees
#'    A single variable is swapped out with another one
#'
#' @return An object of class \code{"adaptive_khaos"}.
#' The object is a list containing:
#' \describe{
#'   \item{B}{Matrix of basis functions evaluated at the training inputs.}
#'   \item{vars, degs}{Arrays encoding the variables and polynomial degrees for each basis function across MCMC iterations.}
#'   \item{nint, dtot}{Interaction order and total degree for each basis function.}
#'   \item{nbasis}{Number of basis functions per MCMC iteration.}
#'   \item{beta}{Posterior samples of regression coefficients.}
#'   \item{s2}{Posterior samples of the noise variance.}
#'   \item{lam}{Posterior samples of the Poisson rate parameter controlling model size.}
#'   \item{sum_sq}{Residual sum of squares at each iteration.}
#'   \item{eta}{Variable inclusion counts used in the proposal distribution.}
#'   \item{count_accept, count_propose}{Acceptance and proposal counts for RJMCMC moves.}
#'   \item{prior_type}{Character string indicating the ridge prior.}
#'   \item{X, y}{Training data.}
#' }
#'
#' @references Francom, Devin, and Bruno Sanso. "BASS: An R package for fitting and performing sensitivity analysis of Bayesian adaptive spline surfaces." Journal of Statistical Software 94.LA-UR-20-23587 (2020).
#'
#' Nott, David J., Anthony YC Kuk, and Hiep Duc. "Efficient sampling schemes for Bayesian MARS models with many predictors." Statistics and Computing 15 (2005): 93-101.
#'
#' Rumsey, K.N., Francom, D., Gibson, G., Derek Tucker, J. and Huerta, G., 2026. Bayesian Adaptive Polynomial Chaos Expansions. Stat, 15(1), p.e70151.
#' @examples
#' \donttest{
#' X <- lhs::maximinLHS(100, 2)
#' f <- function(x) 10.391*((x[1]-0.4)*(x[2]-0.6) + 0.36)
#' y <- apply(X, 1, f) + stats::rnorm(100, 0, 0.1)
#' fit <- adaptive_khaos(X, y)
#' }
#' @export
adaptive_khaos_ridge <-function(X, y,
                          degree=15, order=5,
                          nmcmc=10000,
                          nburn=9000,
                          thin=1,
                          max_basis=1000,
                          tau2=10^5,
                          s2_lower=0,
                          g1=0,g2=0,
                          h1=4,h2=40/length(y),
                          move_probs=rep(1/3, 3),
                          coin_pars=list(function(j) 1/j, 1, 2, 3),
                          degree_penalty=0,
                          legacy = TRUE,
                          verbose=TRUE
){
  # Define a safer solve
  # Note: Under iid uniform X, columns of B are uncorrelated.
  #       So the tolerance should be pretty safe unless there's an
  #       extreme amount of correlation in X.
  safe_solve <- function(A, tol = 1e-9) {
    rc <- rcond(A)
    if (is.na(rc) || rc < tol) return(FALSE)
    solve(A)
  }

  # Make sure X is a matrix
  if(!is.matrix(X)){
    if(is.data.frame(X)){
      X <- as.matrix(X)
    }else if(is.numeric(X)){
      X <- matrix(X, ncol = 1)
    }else{
      stop("X must be a matrix, data frame, or numeric vector.")
    }
  }
  # Check inputs and assign globals
  if(max(X) > 1 || min(X) < 0) warning("Inputs are expected to be scaled on (0, 1). Is this intentional?")
  if(length(y) != nrow(X)) stop("length(y) must equal nrow(X)")
  n<-length(y)
  p<-ncol(X)
  ssy<-sum(y^2)
  if(p < order){
    order <- p
  }
  if(degree < order){
    stop("degree must be at least as large as order")
  }
  #KR: generalized harmonic number
  J_probs <- coin_pars[[1]](1:order)
  Hjs <- sum(J_probs)
  A_total <- A_size(p, degree, order)

  # storage
  degs<-vars<-array(dim=c(nmcmc,max_basis,order)) # this is the largest possible, our filling will be ragged

  nint<-dtot<-matrix(nrow=nmcmc,ncol=max_basis) # order of interaction J, and total degree, again filling will be ragged
  beta<-matrix(nrow=nmcmc,ncol=max_basis+1) # +1 for intercept, again filling will be ragged
  s2<-lam<-nbasis<-rep(NA,nmcmc) # error variance, poisson hyperprior, and number of basis functions
  sum_sq <- rep(NA, nmcmc)
  eta_vec <- rep(0, p) # KR: count how many times each variable is added to the model

  # initialize
  nbasis[1]<-0
  s2[1]<-stats::var(y) + 1e-9
  lam[1]<-1
  B.curr<-matrix(rep(1,n)) # matrix of current basis functions, so that yhat = B.curr %*% beta
  BtB.curr <- crossprod(B.curr)
  BtBi.curr <- solve(BtB.curr)

  Vinv.curr<-crossprod(B.curr)+1/tau2 # V^(-1) from DMS
  bhat<-solve(Vinv.curr)%*%t(B.curr)%*%y
  d.curr <- g2 + ssy - t(bhat)%*%Vinv.curr%*%bhat # d from DMS

  count_accept<-c(0,0,0,0)  # count how many times we accept birth, death, change degree, change vars
  count_propose<-c(0,0,0,0) # count how many times we propose each step
  names(count_accept) <- c("Birth", "Death", "Mutate (degree)", "Mutate (vars)")
  names(count_propose) <- c("Birth", "Death", "Mutate (degree)", "Mutate (vars)")
  beta[1,1]<-bhat

  if(verbose){
    pr<-c('MCMC iteration',0,myTimestamp(),'nbasis:', 0)
    cat(pr,'\n')
  }

  # start MCMC
  for(i in 2:nmcmc){
    ## Reversible jump step

    move.type<-sample(c('birth','death','change'),1,prob=move_probs)
    if(nbasis[i-1]==0)
      move.type<-'birth'
    if(nbasis[i-1]==max_basis)
      move.type<-sample(c('death','change'),1,prob=move_probs[2:3])

    # set all of this iterations values to last iteration values...we'll change them if we accept a move below
    nbasis[i]<-nbasis[i-1]
    nint[i,]<-nint[i-1,]
    dtot[i,]<-dtot[i-1,]
    #knots[i,,]<-knots[i-1,,]
    #signs[i,,]<-signs[i-1,,]
    degs[i,,]<-degs[i-1,,]
    vars[i,,]<-vars[i-1,,]

    if(move.type=='birth'){
      # KR:
      q0  <- sample(1:order, 1, prob=J_probs)
      wts <- make_weights(eta_vec, q0, coin_pars[[3]], coin_pars[[2]], coin_pars[[4]])
      # Delayed rejection component
      chi.cand <- 0
      delayed_reject_term <- 0
      while(sum(chi.cand) == 0 || sum(chi.cand) > order){
        chi.cand  <- stats::rbinom(p, 1, wts)
        vars.cand <- which(chi.cand == 1)
        res <- 0
        for(jj in 1:order){
          wts.j <- make_weights(eta_vec, jj, coin_pars[[3]], coin_pars[[2]], coin_pars[[4]])
          term_j <- log(J_probs[jj]) - log(Hjs) + sum(log(wts.j[vars.cand])) + sum(log(1 - wts.j[-vars.cand]))
          res <- res + exp(term_j)
        }
        delayed_reject_term <- delayed_reject_term + log(res)
      }
      #vars.cand <- which(chi.cand == 1)
      nint.cand <- length(vars.cand)
      d_probs <- seq_along(nint.cand:degree)^(-degree_penalty)
      d_probs <- d_probs / sum(d_probs)
      dtot.cand <- sample(nint.cand:degree, 1, prob=d_probs)
      degs.cand <- random_partition(dtot.cand, nint.cand)
      #nint.cand<-sample(order,1) # sample degree of interaction for new basis function (different from DMS)
      #vars.cand<-sample(p,nint.cand,replace = F) # variables to use in new basis function (different from DMS)
      #knots.cand<-stats::runif(nint.cand) # sample knots for new basis function
      #signs.cand<-sample(c(-1,1),nint.cand,replace = T) # signs for new basis function

      # KR:


      if(nint.cand == 0) stop("Why is nint.cand zero? This shouldn't be possible.")
      basis.cand <- make_basis(vars.cand, degs.cand, X)
      B.cand <- cbind(B.curr,basis.cand) # add the new basis function to the basis functions we already have
      BtB.cand <- crossprod(B.cand)
      BtBi.cand <- tryCatch({
        safe_solve(BtB.cand)
      }, error = function(e) {
        FALSE
      })

      if(!isFALSE(BtBi.cand)){ # Otherwise reject immediately

        Vinv.cand<-crossprod(B.cand)+diag(nbasis[i-1]+2)/tau2 # +2: one for intercept and one for birth
        bhat.cand<-solve(Vinv.cand)%*%t(B.cand)%*%y
        d.cand <- g2 + ssy - t(bhat.cand)%*%Vinv.cand%*%bhat.cand

        # calculate the log likelihood ratio (sort of) after integrating out beta and s2
        llik.alpha <- (
                      0.5*log(1/tau2) +
                      (determinant(Vinv.curr)$mod/2 - determinant(Vinv.cand)$mod/2) +
                      (g1+n/2)*(log(d.curr) - log(d.cand))
        )

        # log prior ratio
        lprior.alpha <- (
                          log(lam[i-1]) - log(nbasis[i-1] + 1) # number of basis functions
                        - log(A_total)                         # Prior over vars/degs
                        + log(nbasis[i-1] + 1)                 # Account for ordering (a historical artifact?)
        )

        # log proposal ratio (proposed to current via death) - (current to proposed)
        lprop.alpha <- (
             (log(move_probs[2])                    # probability of selection a death step
             + log(1/(nbasis[i-1]+1)))              # probability that this basis function is selected to kill
             - (log(move_probs[1])                  # probability of selecting a birth step
             + log(d_probs[1+dtot.cand-nint.cand])  # KR: Probability of total degree
             - lchoose(dtot.cand, nint.cand)        # KR: Probability of degree partition
             + delayed_reject_term)                 # KR: probability of nint and vars based on coin flipping (delayed reject, see above)
        )

        alpha <- llik.alpha + lprior.alpha + lprop.alpha


        if(log(stats::runif(1))<alpha){ # accept, update current values
          B.curr<-B.cand
          BtB.curr <- BtB.cand
          BtBi.curr <- BtBi.cand
          bhat<-bhat.cand
          Vinv.curr<-Vinv.cand
          d.curr<-d.cand
          nbasis[i]<-nbasis[i-1]+1
          nint[i,nbasis[i]]<-nint.cand
          dtot[i,nbasis[i]]<-dtot.cand
          if(!legacy){
            vars[i,nbasis[i],] <- NA
            degs[i,nbasis[i],] <- NA
          }
          vars[i,nbasis[i],1:nint.cand]<-vars.cand
          degs[i,nbasis[i],1:nint.cand]<-degs.cand

          eta_vec[vars.cand] <- eta_vec[vars.cand] + 1
          count_accept[1] <- count_accept[1] + 1
        }
      }
      # Auto reject birth
      count_propose[1] <- count_propose[1] + 1

    } else if(move.type=='death'){
      tokill<-sample(nbasis[i-1],1) # which basis function we will delete
      B.cand<-B.curr[,-(tokill+1),drop=FALSE]  # + 1 to skip the intercept
      BtB.cand <- crossprod(B.cand)
      # There really shouldn't be any issues here,
      # but for some reason we still need tryCatch
      BtBi.cand <- tryCatch({
        safe_solve(BtB.cand)
      }, error = function(e) {
        FALSE
      })

      if(!isFALSE(BtBi.cand)){ #Otherwise reject immediately
        Vinv.cand<-crossprod(B.cand)+diag(nbasis[i-1])/tau2
        bhat.cand<-solve(Vinv.cand)%*%crossprod(B.cand,y)
        d.cand <- g2 + ssy - crossprod(bhat.cand,Vinv.cand%*%bhat.cand)

        # KR:
        vars.cand <- vars[i-1, tokill, 1:nint[i-1,tokill]]
        degs.cand <- degs[i-1, tokill, 1:nint[i-1,tokill]]
        dtot.cand <- sum(degs.cand)
        nint.cand <- length(vars.cand)
        d_probs <- seq_along(nint.cand:degree)^(-degree_penalty)
        d_probs <- d_probs / sum(d_probs)

        eta.cand <- eta_vec
        eta.cand[vars.cand] <- eta.cand[vars.cand] - 1
        proposal_nint   <- 0 # Probability of proposing the sum that we did
        proposal_nint_0 <- 0 # Probability of proposing a sum that equals 0
        proposal_nint_q <- 0 # Probability of proposing a sum that exceeds q=order
        for(jj in 1:order){
          wts.j <- make_weights(eta.cand, jj, coin_pars[[3]], coin_pars[[2]], coin_pars[[4]])
          term_j <- log(J_probs[jj]) - log(Hjs) + sum(log(wts.j[vars.cand])) + sum(log(1 - wts.j[-vars.cand]))
          proposal_nint <- proposal_nint + exp(term_j)

          # Include the delayed rejection stuff
          # P(sum(chi) = 0) = P(sum(chi) = 0 | q0=1)P(q0=1) + ... + P(sum(chi) = 0 | q0=q)P(q0=q)
          term_0_j <- log(J_probs[jj]) - log(Hjs) + sum(log(1 - wts.j))
          proposal_nint_0 <- proposal_nint_0 + exp(term_0_j)

          # if order == p, then it is not possible to overshoot
          term_q_j <- -Inf
          if(order < p){
            # Use a normal approx for this part
            mu_chi   <- sum(wts.j)
            sig_chi  <- sqrt(sum(wts.j*(1-wts.j)))
            term_q_j <- log(J_probs[jj]) - log(Hjs) +
                        stats::pnorm(order+1/2,mu_chi,sig_chi,lower.tail=FALSE,log.p=TRUE)
          }
          proposal_nint_q <- proposal_nint_q + exp(term_q_j)
        }
        proposal_reject <- min(1, proposal_nint_0 + proposal_nint_q) # The 1 should never trigger, but I added it to be safe since we're using a normal approximation


        # calculate the log likelihood ratio (sort of) after integrating out beta and s2
        llik.alpha <- (
          -0.5*log(1/tau2) +
            (determinant(Vinv.curr)$mod/2 - determinant(Vinv.cand)$mod/2) +
            (g1+n/2)*(log(d.curr) - log(d.cand))
        )

        # log prior ratio
        lprior.alpha <- (
          - log(lam[i-1]) + log(nbasis[i-1])     # number of basis functions
          + log(A_total)                         # Prior over vars/degs
          - log(nbasis[i-1] + 1)                 # Account for ordering (a historical artifact?)
        )

        # log proposal ratio (proposed to current via birth) - (current to proposed)
        lprop.alpha <- (
           (log(move_probs[1])                               # probability of selection a birth step
           + log(proposal_nint) - log(1 - proposal_reject)   # probability that we proposed the variables we did (accounting for rejection shenanigans)
           + log(d_probs[1+dtot.cand-nint.cand])             # probability that we proposed the degree that we did
           - lchoose(dtot.cand, nint.cand))                  # probability that we proposed the degree partition that we did
           - (log(move_probs[2])                             # probability of selecting a death step
           + log(1/nbasis[i-1]))                             # probability of selecting this particular basis function to kill
        )

        alpha <- llik.alpha + lprior.alpha + lprop.alpha

        if(log(stats::runif(1))<alpha){ # accept, update
          B.curr<-B.cand
          BtB.curr <- BtB.cand
          BtBi.curr <- BtBi.cand
          bhat<-bhat.cand
          Vinv.curr<-Vinv.cand
          d.curr<-d.cand
          nbasis[i]<-nbasis[i-1]-1

          nint[i,]<-dtot[i,]<-NA
          vars[i,,]<-degs[i,,]<-NA
          if(nbasis[i] > 0){
            nint[i,1:nbasis[i]]<-nint[i-1,(1:nbasis[i-1])[-tokill]]
            dtot[i,1:nbasis[i]]<-dtot[i-1,(1:nbasis[i-1])[-tokill]]

            vars[i,1:nbasis[i],]<-vars[i-1,(1:nbasis[i-1])[-tokill],]
            degs[i,1:nbasis[i],]<-degs[i-1,(1:nbasis[i-1])[-tokill],]
          }
          eta_vec <- eta.cand
          count_accept[2] <- count_accept[2] + 1
        }
      }
      count_propose[2] <- count_propose[2] + 1

    } else{ # KR: mutate

      # Adapt which mutation should get used.
      if(p <= 3){
        # Second mutation type isn't really needed for small p
        mutate_type <- 1
      }else{
        mutate_eps <- 0.1 # hard coded (worst case probability for either type)
        success_rates <- count_accept[3:4]/(0.01 + count_propose[3:4])
        mutate_prob <- mutate_eps + (1 - 2*mutate_eps) * stats::pnorm(-diff(pmin(5,pmax(-5,stats::qnorm(success_rates)))))
        mutate_type <- stats::rbinom(1, 1, mutate_prob)
      }

      # MUTATION TYPE 1: Re-sample the degrees
      if(mutate_type == 1){
        tochange<-sample(nbasis[i-1],1)                        # which basis function we will change

        # Get current info
        vars.curr <- vars[i-1,tochange,1:nint[i-1,tochange]]
        degs.curr <- degs[i-1,tochange,1:nint[i-1,tochange]]
        dtot.curr <- sum(degs.curr)    # equivalently dtot[i-1,tochange]
        nint.curr <- length(vars.curr) # equivalently nint[i-1,tochange]
        d_probs <- seq_along(nint.curr:degree)^(-degree_penalty)
        d_probs <- d_probs / sum(d_probs)

        # Generate new degree partition
        dtot.cand <- sample(nint.curr:degree, 1, prob=d_probs)
        degs.cand <- random_partition(dtot.cand, nint.curr)

        # Generate candidate basis function
        basis <- make_basis(vars.curr, degs.cand, X)
        B.cand <- B.curr
        B.cand[,tochange+1] <- basis # replace with our new basis function (+1 for intercept)
        BtB.cand <- crossprod(B.cand)
        BtBi.cand <- tryCatch({
          safe_solve(BtB.cand)
        }, error = function(e) {
          FALSE
        })

        if(!isFALSE(BtBi.cand)){ #Otherwise reject immediately
          Vinv.cand<-crossprod(B.cand)+diag(nbasis[i-1]+1)/tau2
          bhat.cand<-solve(Vinv.cand)%*%crossprod(B.cand,y)
          d.cand <- g2 + ssy - crossprod(bhat.cand,Vinv.cand%*%bhat.cand)

          # Compute acceptance ratio
          # Assume that "interaction" between the two mutation types is negligible
          llik.alpha <- determinant(Vinv.curr)$mod/2 - determinant(Vinv.cand)$mod/2 +
                        (g1+n/2)*(log(d.curr) - log(d.cand))

          # Variation in the degree proposal
          lprop.alpha <- log(d_probs[1+dtot.curr-nint.curr]) - log(d_probs[1+dtot.cand-nint.curr]) +
                         (-lchoose(dtot.curr, nint.curr) + lchoose(dtot.cand, nint.curr))

          # Uniform priors + no dimension change <=> no term for the prior
          alpha <- llik.alpha + lprop.alpha

          if(log(stats::runif(1))<alpha){ # accept, update
            B.curr<-B.cand
            BtB.curr <- BtB.cand
            BtBi.curr <- BtBi.cand
            bhat<-bhat.cand
            Vinv.curr<-Vinv.cand
            d.curr<-d.cand

            vars[i,tochange,1:nint[i-1,tochange]] <- vars.curr # no change
            degs[i,tochange,1:nint[i-1,tochange]] <- degs.cand
            nint[i,tochange] <- nint.curr                      # no change
            dtot[i,tochange] <- dtot.cand

            count_accept[3] <- count_accept[3] + 1
          }
        }
        count_propose[3] <- count_propose[3] + 1

      }else{
      # MUTATION TYPE 2: Swap out a variable
        tochange<-sample(nbasis[i-1],1)           # which basis function we will change

        # Get current info
        vars.curr <- vars[i-1,tochange,1:nint[i-1,tochange]]
        degs.curr <- degs[i-1,tochange,1:nint[i-1,tochange]]
        dtot.curr <- sum(degs.curr)    # equivalently dtot[i-1,tochange]
        nint.curr <- length(vars.curr) # equivalently nint[i-1,tochange]

        # Choose a variable to delete
        ind_tochange <- sample(nint.curr, 1)
        var_tochange <- vars.curr[ind_tochange]
        newvar_probs.curr <- 1e-9 + coin_pars[[2]] + eta_vec
        newvar_probs.curr[var_tochange] <- 0
        newvar_probs.curr <- newvar_probs.curr / sum(newvar_probs.curr)
        newvar.curr <- sample(p, 1, prob=newvar_probs.curr)

        vars.cand <- vars.curr
        vars.cand[ind_tochange] <- newvar.curr

        # Construct probabilities for the reverse case
        eta.cand <- eta_vec
        eta.cand[var_tochange] <- eta.cand[var_tochange] - 1
        eta.cand[newvar.curr] <- eta.cand[newvar.curr] + 1
        newvar_probs.cand <- 1e-9 + coin_pars[[2]] + eta.cand
        newvar_probs.cand[newvar.curr] <- 0
        newvar_probs.cand <- newvar_probs.cand / sum(newvar_probs.cand)
        newvar.cand <- var_tochange

        # Generate candidate basis function
        basis <- make_basis(vars.cand, degs.curr, X)
        B.cand <- B.curr
        B.cand[,tochange+1] <- basis # replace with our new basis function (+1 for intercept)
        BtB.cand <- crossprod(B.cand)
        BtBi.cand <- tryCatch({
          safe_solve(BtB.cand)
        }, error = function(e) {
          FALSE
        })

        if(!isFALSE(BtBi.cand)){ # Otherwise reject immediately
          Vinv.cand<-crossprod(B.cand)+diag(nbasis[i-1]+1)/tau2
          bhat.cand<-solve(Vinv.cand)%*%crossprod(B.cand,y)
          d.cand <- g2 + ssy - crossprod(bhat.cand,Vinv.cand%*%bhat.cand)

          # Compute acceptance ratio
          # Assume that "interaction" between the two mutation types is negligible
          llik.alpha <- determinant(Vinv.curr)$mod/2 - determinant(Vinv.cand)$mod/2 +
            (g1+n/2)*(log(d.curr) - log(d.cand))

          # Difference in probability of the variables.
          lprop.alpha <- log(newvar_probs.cand[newvar.cand]) - log(newvar_probs.curr[newvar.curr])

          # Uniform priors + no dimension change <=> no term for the prior
          alpha <- llik.alpha + lprop.alpha

          if(log(stats::runif(1))<alpha){ # accept, update
            B.curr<-B.cand
            BtB.curr <- BtB.cand
            BtBi.curr <- BtBi.cand
            bhat<-bhat.cand
            Vinv.curr<-Vinv.cand
            d.curr<-d.cand

            vars[i,tochange,1:nint[i,tochange]] <- vars.cand
            degs[i,tochange,1:nint[i,tochange]] <- degs.curr # no change
            nint[i,tochange] <- nint.curr                    # no change
            dtot[i,tochange] <- dtot.curr                    # no change

            eta_vec <- eta.cand

            count_accept[4] <- count_accept[4] + 1
          }
        }
        count_propose[4] <- count_propose[4] + 1
      }
    }

    ## Gibbs steps
    Lambda_n <- BtB.curr + diag(nbasis[i] + 1)/tau2
    Lambda_i_n <- solve(Lambda_n)
    beta_ls <- BtBi.curr %*% crossprod(B.curr, y)
    mu_n <- Lambda_i_n %*% (BtB.curr %*% beta_ls)
    Sigma_n <- s2[i-1] * Lambda_i_n
    beta_curr <- t(mvtnorm::rmvnorm(1, mu_n, Sigma_n))
    beta[i,1:(nbasis[i]+1)] <- beta_curr
    yhat_curr <- B.curr %*% beta_curr
    resid <- y-yhat_curr


    s2[i] <- max(s2_lower, 1/stats::rgamma(1, n/2+g1, rate = g2+.5*sum(resid^2)))
    lam[i] <- stats::rgamma(1,h1+nbasis[i],h2+1)                                               # update lambda
    sum_sq[i] <- sum(resid^2)

    # S_cov <- solve(crossprod(B.curr)/s2[i-1]+diag(nbasis[i]+1)/tau2)                    # covariance matrix for beta update
    # beta[i,1:(nbasis[i]+1)] <- sample_mvn(1, S_cov%*%t(B.curr)%*%y/s2[i-1], S_cov)      # update beta
    # s2[i] <- 1/stats::rgamma(1,n/2+g1,rate=g2+.5*sum((y-B.curr%*%beta[i,1:(nbasis[i]+1)])^2))  # update s2
    # #if(s2[i] > 1000) browser()

    if(verbose & i%%1000 == 0){
      pr<-c('MCMC iteration',i,myTimestamp(),'nbasis:', nbasis[i])
      cat(pr,'\n')
    }
  }

  # Trim down data structures
  mcmc_iter <- seq(nburn+1, nmcmc, by=thin)
  max_basis_used <- max(nbasis, na.rm = TRUE)
  basis_high <- if(max_basis_used > 0) 1:max_basis_used else integer(0)
  max_inter_used <- max(nint, na.rm = TRUE)
  inter_high <- if(is.finite(max_inter_used) && max_inter_used > 0) 1:max_inter_used else integer(0)

  vars <- vars[mcmc_iter, basis_high, inter_high, drop=FALSE]
  degs <- degs[mcmc_iter, basis_high, inter_high, drop=FALSE]
  nint <- nint[mcmc_iter, basis_high, drop=FALSE]
  dtot <- dtot[mcmc_iter, basis_high, drop=FALSE]

  beta <- beta[mcmc_iter, 1:(1+max(nbasis))]
  nbasis <- nbasis[mcmc_iter]
  s2 <- s2[mcmc_iter]
  lam <- lam[mcmc_iter]
  sum_sq <- sum_sq[mcmc_iter]


  out <- list(B=B.curr,
              vars=vars,degs=degs,
              nint=nint,dtot=dtot,
              nbasis=nbasis,beta=beta,
              s2=s2,lam=lam,sum_sq=sum_sq,
              eta=eta_vec,
              count_accept=count_accept,
              count_propose=count_propose,
              prior_type="ridge",
              X=X, y=y)
  class(out) <- "adaptive_khaos"
  return(out)
}

#' Predict Method for class adaptive_khaos
#'
#' See \code{adaptive_khaos()} for details.
#'
#' @param object An object returned by the \code{adaptive_khaos()} function.
#' @param newdata A dataframe of the same dimension as the training data.
#' @param mcmc.use Which posterior samples should be used for prediction?
#' @param nugget Logical. Should predictions include error?
#' @param nreps How many predictions should be taken for each mcmc sample (ignored when \code{nugget = FALSE}).
#' @param ... Additional arguments to predict
#' @details Predict function for adaptive_khaos object.
#' @return A numeric matrix of posterior predictive samples, with rows corresponding
#' to samples and columns corresponding to observations.
#' @references Francom, Devin, and Bruno Sanso. "BASS: An R package for fitting and performing sensitivity analysis of Bayesian adaptive spline surfaces." Journal of Statistical Software 94.LA-UR-20-23587 (2020).
#' @examples
#' \donttest{
#' X <- lhs::maximinLHS(100, 2)
#' f <- function(x) 10.391*((x[1]-0.4)*(x[2]-0.6) + 0.36)
#' y <- apply(X, 1, f) + stats::rnorm(100, 0, 0.1)
#' fit <- adaptive_khaos(X, y)
#' predict(fit)
#' }
#' @export
predict.adaptive_khaos<-function(object, newdata=NULL, mcmc.use=NULL, nugget=FALSE, nreps=1, ...){ # prediction function, gets a prediction for each MCMC iteration
  if(is.null(newdata)){
    newdata <- object$X
  }
  if(is.null(mcmc.use)){
    mcmc.use <- seq_along(object$nbasis)
  }
  if(!nugget){
    nreps <- 1
  }
  pred <- matrix(NA, nrow=length(mcmc.use)*nreps, ncol=nrow(newdata))
  cnt <- 1
  for(i in mcmc.use){
    B <- matrix(1, nrow=nrow(newdata),ncol=object$nbasis[i]+1)
    for(j in seq_len(object$nbasis[i])){
      B[,j+1] <- make_basis(object$vars[i,j,1:object$nint[i,j]], object$degs[i,j,1:object$nint[i,j]], newdata)
    }
    beta_curr <- object$beta[i,1:(object$nbasis[i]+1)]
    if(nugget){
      mu <- matrix(rep(B %*% beta_curr, each=nreps),
                   nrow=nreps, ncol=nrow(newdata))
      sigma <- sqrt(object$s2[i])
      pred[(1 + (cnt-1)*nreps):(cnt*nreps),] <- mu + stats::rnorm(nrow(newdata)*nreps, 0, sigma)
    }else{
      pred[cnt,] <- B %*% beta_curr
    }
    cnt <- cnt + 1
  }
  return(pred)
}


#' Plot Method for class adaptive_khaos
#'
#' See \code{adaptive_khaos()} for details.
#'
#' @param x An object returned by the \code{adaptive_khaos()} function.
#' @param ... Additional arguments for plotting
#' @details Plot function for adaptive_khaos object.
#' @return No return value; called for its side effects (plotting).
#' @references Francom, Devin, and Bruno Sanso. "BASS: An R package for fitting and performing sensitivity analysis of Bayesian adaptive spline surfaces." Journal of Statistical Software 94.LA-UR-20-23587 (2020).
#' @examples
#' \donttest{
#' X <- lhs::maximinLHS(100, 2)
#' f <- function(x) 10.391*((x[1]-0.4)*(x[2]-0.6) + 0.36)
#' y <- apply(X, 1, f) + stats::rnorm(100, 0, 0.1)
#' fit <- adaptive_khaos(X, y)
#' plot(fit)
#' }
#' @export
plot.adaptive_khaos <- function(x, ...){
  #opar <- graphics::par(no.readonly = TRUE)
  oldpar <- par(no.readonly = TRUE)
  on.exit(par(oldpar))
  graphics::par(mfrow = c(2, 2), mar = c(4, 4, 2, 1), oma = c(0, 0, 0, 0))

  preds <- stats::predict(x, nugget=TRUE, nreps=10)
  yhat <- apply(preds, 2, mean)
  ci <- 2*apply(preds, 2, stats::sd)

  stats::ts.plot(x$nbasis, ylab="nbasis")
  stats::ts.plot(x$s2, ylab="s2")
  plot(x$y, yhat, pch=16, xlab="y")
  graphics::segments(x0=x$y, y0=yhat-ci, y1=yhat+ci, col='orange')

  graphics::points(x$y, yhat, pch=16)
  graphics::abline(0,1,col='dodgerblue')
  rr <- x$y-yhat
  rr_sd <- stats::sd(rr)
  if(!is.finite(rr_sd) || rr_sd <= 0){
    num_breaks <- 10
  } else {
    num_breaks <- max(1, ceiling(length(rr)^0.33 * diff(range(rr)) / (3.5 * rr_sd)))
  }
  graphics::hist(rr, breaks=num_breaks, freq=F)
  x = seq(range(rr)[1], range(rr)[2], length.out=100)
  graphics::curve(stats::dnorm(x, mean(rr), stats::sd(rr)), add=TRUE, col='orange')

  #graphics::par(opar)
  return(NULL)
}

#' Ordinal Probit Regression with BA-PCE
#'
#' The emulation approach of Francom et al. (2020) for BMARS, modified for polynomial chaos.
#'
#' @param X A data frame or matrix of predictors scaled to be between 0 and 1
#' @param y a response vector
#' @param degree Maximum polynomial degree for each basis function.
#' @param order Maximum order of interaction for each basis function.
#' @param nmcmc Number of MCMC iterations.
#' @param nburn Number of initial MCMC iterations to discard.
#' @param thin Keep every \code{thin} samples.
#' @param max_basis Maximum number of basis functions.
#' @param tau2 Prior variance for coefficients
#' @param sigma_thresh Prior variance for the threshold parameters.
#' @param h1,h2 Shape/rate parameters for the Gamma prior on expected number of basis functions.
#' @param move_probs A 3-vector with probabilities for (i) birth, (ii) death, and (iii) mutation.
#' @param coin_pars A list of control parameters for coinflip proposal
#' @param degree_penalty Increasing this value encourages lower order polynomial terms (0 is no penalization).
#' @param verbose Logical. Should progress information be printed?
#' @details Implements the RJMCMC algorithm described by Francom & Sanso (2020) for BMARS, modifying it for polynomial chaos basis functions.
#' As an alternative the NKD procedure of Nott et al. (2005), we use a coinflipping procedure to identify useful variables. See writeup (coming soon) for details.
#' The coin_pars argument is an ordered list with elements:
#' \enumerate{
#' \item{a function giving proposal probability for q_0, the expected order of interaction,}
#' \item{epsilon, the base weight for each variable (epsilon=Inf corresponds to uniform sampling),}
#' \item{alpha, the exponent-learning rate,}
#' \item{num_passes a numerical parameter only.}
#' }
#'
#' Other proposal notes:
#'    Expected order q0 is chosen from 1:order with weights `coin_pars[[1]](1:order)`
#'    Degree is chosen from q0:degree with weights `(1:(degree-q0+1))^(-degree_penalty)` and a random partition is created (see helper function).
#'
#' Two types of change steps are possible (and equally likely)
#'    A re-ordering of the degrees
#'    A single variable is swapped out with another one
#'
#' @return An object of class \code{"ordinal_khaos"}.
#' The object is a list containing:
#' \describe{
#'   \item{B}{Matrix of basis functions evaluated at the training inputs.}
#'   \item{vars, degs}{Arrays encoding the variables and polynomial degrees for each basis function across MCMC iterations.}
#'   \item{nint, dtot}{Interaction order and total degree for each basis function.}
#'   \item{nbasis}{Number of basis functions per MCMC iteration.}
#'   \item{beta}{Posterior samples of regression coefficients.}
#'   \item{thresh}{Posterior samples of the ordinal threshold parameters.}
#'   \item{s2z}{Estimated latent variance across iterations.}
#'   \item{lam}{Posterior samples of the Poisson rate parameter controlling model size.}
#'   \item{eta}{Variable inclusion counts used in the proposal distribution.}
#'   \item{count_accept, count_propose}{Acceptance and proposal counts for RJMCMC moves.}
#'   \item{X, y}{Training data.}
#' }
#'
#' @references Francom, Devin, and Bruno Sanso. "BASS: An R package for fitting and performing sensitivity analysis of Bayesian adaptive spline surfaces." Journal of Statistical Software 94.LA-UR-20-23587 (2020).
#'
#' Nott, David J., Anthony YC Kuk, and Hiep Duc. "Efficient sampling schemes for Bayesian MARS models with many predictors." Statistics and Computing 15 (2005): 93-101.
#' @examples
#' \donttest{
#' set.seed(1)
#' X <- matrix(runif(200), ncol = 1)
#' y <- sample(1:4, 200, replace = TRUE)
#' fit <- ordinal_khaos(X, y, nmcmc = 2000, nburn = 1000, thin = 2, verbose = FALSE)
#' plot(fit)
#' }
#' @export
ordinal_khaos <-function(X, y,
                          degree=15, order=5,
                          nmcmc=10000,
                          nburn=9000,
                          thin=1,
                          max_basis=1000,
                          tau2=10^5,
                          sigma_thresh=10,
                          h1=4,h2=40/length(y),
                          move_probs=rep(1/3, 3),
                          coin_pars=list(function(j) 1/j, 1, 2, 3),
                          degree_penalty=0,
                          verbose=TRUE
){
  # Define a safer solve
  # Note: Under iid uniform X, columns of B are uncorrelated.
  #       So the tolerance should be pretty safe unless there's an
  #       extreme amount of correlation in X.
  safe_solve <- function(A, tol = 1e-9) {
    rc <- rcond(A)
    if (is.na(rc) || rc < tol) return(FALSE)
    solve(A)
  }
  is_odd <- function(n) (n %% 2) == 1

  # Make sure X is a matrix
  if(!is.matrix(X)){
    if(is.data.frame(X)){
      X <- as.matrix(X)
    }else if(is.numeric(X)){
      X <- matrix(X, ncol = 1)
    }else{
      stop("X must be a matrix, data frame, or numeric vector.")
    }
  }
  # Check inputs and assign globals
  if(max(X) > 1 || min(X) < 0) warning("Inputs are expected to be scaled on (0, 1). Is this intentional?")
  if(length(y) != nrow(X)) stop("length(y) must equal nrow(X)")
  n<-length(y)
  p<-ncol(X)
  if(p < order){
    order <- p
  }
  if(degree < order){
    stop("degree must be at least as large as order")
  }

  if(!all(unlist(lapply(y, function(yy) (yy %% 1) == 0)))){
    stop("y must be integer valued.")
  }
  K <- max(y)
  if(K < 2) stop("Need at least 2 categories")
  if(!all(y %in% 1:K)){
    stop("y must have values 1, 2, ... K")
  }
  if(length(unique(y)) != K){
    stop("All K classes must be represented. Consider re-defining class levels.")
  }
  thresh_curr <- mu_thresh <- 1:K - (K + is_odd(K)) / 2
  # Initialize z
  z <- thresh_curr[y] + rbeta(n, 5, 5) - 1
  ssz<-sum(z^2)
  thresh_curr[K] <- Inf

  s2z <- rep(NA, nmcmc)
  s2z[1] <- var(z)

  #KR: generalized harmonic number
  J_probs <- coin_pars[[1]](1:order)
  Hjs <- sum(J_probs)
  A_total <- A_size(p, degree, order)

  # storage
  degs<-vars<-array(dim=c(nmcmc,max_basis,order)) # this is the largest possible, our filling will be ragged

  nint<-dtot<-matrix(nrow=nmcmc,ncol=max_basis) # order of interaction J, and total degree, again filling will be ragged
  thresh_mat <- matrix(nrow=nmcmc, ncol=K-1)
  beta<-matrix(nrow=nmcmc,ncol=max_basis+1) # +1 for intercept, again filling will be ragged
  lam<-nbasis<- rep(NA,nmcmc) # error variance, poisson hyperprior, and number of basis functions
  #sum_sq <- rep(NA, nmcmc)
  eta_vec <- rep(0, p) # KR: count how many times each variable is added to the model

  # initialize
  nbasis[1]<-0
  s2 <- 1
  lam[1]<-1
  B.curr<-matrix(rep(1,n)) # matrix of current basis functions, so that yhat = B.curr %*% beta
  BtB.curr <- crossprod(B.curr)
  BtBi.curr <- solve(BtB.curr)
  thresh_mat[1,] <- thresh_curr[1:(K-1)]

  Vinv.curr<-crossprod(B.curr)+1/tau2 # V^(-1) from DMS
  bhat<-1

  count_accept<-c(0,0,0,0)  # count how many times we accept birth, death, change degree, change vars
  count_propose<-c(0,0,0,0) # count how many times we propose each step
  names(count_accept) <- c("Birth", "Death", "Mutate (degree)", "Mutate (vars)")
  names(count_propose) <- c("Birth", "Death", "Mutate (degree)", "Mutate (vars)")
  beta[1,1]<-bhat

  if(verbose){
    pr<-c('MCMC iteration',0,myTimestamp(),'nbasis:', 0)
    cat(pr,'\n')
  }

  # start MCMC
  for(i in 2:nmcmc){
    ## Reversible jump step

    move.type<-sample(c('birth','death','change'),1,prob=move_probs)
    if(nbasis[i-1]==0)
      move.type<-'birth'
    if(nbasis[i-1]==max_basis)
      move.type<-sample(c('death','change'),1,prob=move_probs[2:3])

    # set all of this iterations values to last iteration values...we'll change them if we accept a move below
    nbasis[i]<-nbasis[i-1]
    nint[i,]<-nint[i-1,]
    dtot[i,]<-dtot[i-1,]
    #knots[i,,]<-knots[i-1,,]
    #signs[i,,]<-signs[i-1,,]
    degs[i,,]<-degs[i-1,,]
    vars[i,,]<-vars[i-1,,]

    if(move.type=='birth'){
      # KR:
      q0  <- sample(1:order, 1, prob=J_probs)
      wts <- make_weights(eta_vec, q0, coin_pars[[3]], coin_pars[[2]], coin_pars[[4]])
      # Delayed rejection component
      chi.cand <- 0
      delayed_reject_term <- 0
      while(sum(chi.cand) == 0 || sum(chi.cand) > order){
        chi.cand  <- stats::rbinom(p, 1, wts)
        vars.cand <- which(chi.cand == 1)
        res <- 0
        for(jj in 1:order){
          wts.j <- make_weights(eta_vec, jj, coin_pars[[3]], coin_pars[[2]], coin_pars[[4]])
          term_j <- log(J_probs[jj]) - log(Hjs) + sum(log(wts.j[vars.cand])) + sum(log(1 - wts.j[-vars.cand]))
          res <- res + exp(term_j)
        }
        delayed_reject_term <- delayed_reject_term + log(res)
      }
      #vars.cand <- which(chi.cand == 1)
      nint.cand <- length(vars.cand)
      d_probs <- seq_along(nint.cand:degree)^(-degree_penalty)
      d_probs <- d_probs / sum(d_probs)
      dtot.cand <- sample(nint.cand:degree, 1, prob=d_probs)
      degs.cand <- random_partition(dtot.cand, nint.cand)

      if(nint.cand == 0) stop("Why is nint.cand zero? This shouldn't be possible.")
      basis.cand <- make_basis(vars.cand, degs.cand, X)
      B.cand <- cbind(B.curr,basis.cand) # add the new basis function to the basis functions we already have
      BtB.cand <- crossprod(B.cand)
      BtBi.cand <- tryCatch({
        safe_solve(BtB.cand)
      }, error = function(e) {
        FALSE
      })

      if(!isFALSE(BtBi.cand)){ # Otherwise reject immediately
        Vinv.cand<-crossprod(B.cand)+diag(nbasis[i-1]+2)/tau2 # +2: one for intercept and one for birth

        b.curr <- crossprod(B.curr, z)
        q.curr <- as.numeric(crossprod(b.curr, solve(Vinv.curr, b.curr)))
        logdet.curr <- as.numeric(determinant(Vinv.curr)$mod)

        b.cand <- crossprod(B.cand, z)
        q.cand <- as.numeric(crossprod(b.cand, solve(Vinv.cand, b.cand)))
        logdet.cand <- as.numeric(determinant(Vinv.cand)$mod)

        # calculate the log likelihood ratio (sort of) after integrating out beta and s2
        llik.alpha <- 0.5 * (q.cand - q.curr) - 0.5 * (logdet.cand - logdet.curr) + 0.5 * log(1/tau2)

        # log prior ratio
        lprior.alpha <- (
          log(lam[i-1]) - log(nbasis[i-1] + 1) # number of basis functions
          - log(A_total)                         # Prior over vars/degs
          + log(nbasis[i-1] + 1)                 # Account for ordering (a historical artifact?)
        )

        # log proposal ratio (proposed to current via death) - (current to proposed)
        lprop.alpha <- (
          (log(move_probs[2])                    # probability of selection a death step
           + log(1/(nbasis[i-1]+1)))              # probability that this basis function is selected to kill
          - (log(move_probs[1])                  # probability of selecting a birth step
             + log(d_probs[1+dtot.cand-nint.cand])  # KR: Probability of total degree
             - lchoose(dtot.cand, nint.cand)        # KR: Probability of degree partition
             + delayed_reject_term)                 # KR: probability of nint and vars based on coin flipping (delayed reject, see above)
        )

        alpha <- llik.alpha + lprior.alpha + lprop.alpha


        if(log(stats::runif(1))<alpha){ # accept, update current values
          B.curr     <-B.cand
          BtB.curr   <- BtB.cand
          BtBi.curr  <- BtBi.cand
          Vinv.curr<-Vinv.cand

          nbasis[i]<-nbasis[i-1]+1
          nint[i,nbasis[i]]<-nint.cand
          dtot[i,nbasis[i]]<-dtot.cand
          vars[i,nbasis[i],1:nint.cand]<-vars.cand
          degs[i,nbasis[i],1:nint.cand]<-degs.cand

          eta_vec[vars.cand] <- eta_vec[vars.cand] + 1
          count_accept[1] <- count_accept[1] + 1
        }
      }
      # Auto reject birth
      count_propose[1] <- count_propose[1] + 1

    } else if(move.type=='death'){
      tokill<-sample(nbasis[i-1],1) # which basis function we will delete
      B.cand<-B.curr[,-(tokill+1)]  # + 1 to skip the intercept
      BtB.cand <- crossprod(B.cand)
      # There really shouldn't be any issues here,
      # but for some reason we still need tryCatch
      BtBi.cand <- tryCatch({
        safe_solve(BtB.cand)
      }, error = function(e) {
        FALSE
      })

      if(!isFALSE(BtBi.cand)){ #Otherwise reject immediately
        Vinv.cand<-crossprod(B.cand)+diag(nbasis[i-1])/tau2

        b.curr <- crossprod(B.curr, z)
        q.curr <- as.numeric(crossprod(b.curr, solve(Vinv.curr, b.curr)))
        logdet.curr <- as.numeric(determinant(Vinv.curr)$mod)

        b.cand <- crossprod(B.cand, z)
        q.cand <- as.numeric(crossprod(b.cand, solve(Vinv.cand, b.cand)))
        logdet.cand <- as.numeric(determinant(Vinv.cand)$mod)

        # KR:
        vars.cand <- vars[i-1, tokill, 1:nint[i-1,tokill]]
        degs.cand <- degs[i-1, tokill, 1:nint[i-1,tokill]]
        dtot.cand <- sum(degs.cand)
        nint.cand <- length(vars.cand)
        d_probs <- seq_along(nint.cand:degree)^(-degree_penalty)
        d_probs <- d_probs / sum(d_probs)

        eta.cand <- eta_vec
        eta.cand[vars.cand] <- eta.cand[vars.cand] - 1
        proposal_nint   <- 0 # Probability of proposing the sum that we did
        proposal_nint_0 <- 0 # Probability of proposing a sum that equals 0
        proposal_nint_q <- 0 # Probability of proposing a sum that exceeds q=order
        for(jj in 1:order){
          wts.j <- make_weights(eta.cand, jj, coin_pars[[3]], coin_pars[[2]], coin_pars[[4]])
          term_j <- log(J_probs[jj]) - log(Hjs) + sum(log(wts.j[vars.cand])) + sum(log(1 - wts.j[-vars.cand]))
          proposal_nint <- proposal_nint + exp(term_j)

          # Include the delayed rejection stuff
          # P(sum(chi) = 0) = P(sum(chi) = 0 | q0=1)P(q0=1) + ... + P(sum(chi) = 0 | q0=q)P(q0=q)
          term_0_j <- log(J_probs[jj]) - log(Hjs) + sum(log(1 - wts.j))
          proposal_nint_0 <- proposal_nint_0 + exp(term_0_j)

          # if order == p, then it is not possible to overshoot
          term_q_j <- -Inf
          if(order < p){
            # Use a normal approx for this part
            mu_chi   <- sum(wts.j)
            sig_chi  <- sqrt(sum(wts.j*(1-wts.j)))
            term_q_j <- log(J_probs[jj]) - log(Hjs) +
              stats::pnorm(order+1/2,mu_chi,sig_chi,lower.tail=FALSE,log.p=TRUE)
          }
          proposal_nint_q <- proposal_nint_q + exp(term_q_j)
        }
        proposal_reject <- min(1, proposal_nint_0 + proposal_nint_q) # The 1 should never trigger, but I added it to be safe since we're using a normal approximation


        # calculate the log likelihood ratio (sort of) after integrating out beta and s2
        llik.alpha <- 0.5 * (q.cand - q.curr) - 0.5 * (logdet.cand - logdet.curr) - 0.5 * log(1/tau2)

        # log prior ratio
        lprior.alpha <- (
          - log(lam[i-1]) + log(nbasis[i-1])     # number of basis functions
          + log(A_total)                         # Prior over vars/degs
          - log(nbasis[i-1] + 1)                 # Account for ordering (a historical artifact?)
        )

        # log proposal ratio (proposed to current via birth) - (current to proposed)
        lprop.alpha <- (
          (log(move_probs[1])                               # probability of selection a birth step
           + log(proposal_nint) - log(1 - proposal_reject)   # probability that we proposed the variables we did (accounting for rejection shenanigans)
           + log(d_probs[1+dtot.cand-nint.cand])             # probability that we proposed the degree that we did
           - lchoose(dtot.cand, nint.cand))                  # probability that we proposed the degree partition that we did
          - (log(move_probs[2])                             # probability of selecting a death step
             + log(1/nbasis[i-1]))                             # probability of selecting this particular basis function to kill
        )

        alpha <- llik.alpha + lprior.alpha + lprop.alpha

        if(log(stats::runif(1))<alpha){ # accept, update
          B.curr<-B.cand
          BtB.curr <- BtB.cand
          BtBi.curr <- BtBi.cand
          Vinv.curr<-Vinv.cand

          nbasis[i]<-nbasis[i-1]-1
          nint[i,]<-dtot[i,]<-NA
          vars[i,,]<-degs[i,,]<-NA
          if(nbasis[i] > 0){
            nint[i,1:nbasis[i]]<-nint[i-1,(1:nbasis[i-1])[-tokill]]
            dtot[i,1:nbasis[i]]<-dtot[i-1,(1:nbasis[i-1])[-tokill]]

            vars[i,1:nbasis[i],]<-vars[i-1,(1:nbasis[i-1])[-tokill],]
            degs[i,1:nbasis[i],]<-degs[i-1,(1:nbasis[i-1])[-tokill],]
          }
          eta_vec <- eta.cand
          count_accept[2] <- count_accept[2] + 1
        }
      }
      count_propose[2] <- count_propose[2] + 1

    } else{ # KR: mutate

      # Adapt which mutation should get used.
      if(p <= 3){
        # Second mutation type isn't really needed for small p
        mutate_type <- 1
      }else{
        mutate_eps <- 0.1 # hard coded (worst case probability for either type)
        success_rates <- count_accept[3:4]/(0.01 + count_propose[3:4])
        mutate_prob <- mutate_eps + (1 - 2*mutate_eps) * stats::pnorm(-diff(pmin(5,pmax(-5,stats::qnorm(success_rates)))))
        mutate_type <- stats::rbinom(1, 1, mutate_prob)
      }

      # MUTATION TYPE 1: Re-sample the degrees
      if(mutate_type == 1){
        tochange<-sample(nbasis[i-1],1)                        # which basis function we will change

        # Get current info
        vars.curr <- vars[i-1,tochange,1:nint[i-1,tochange]]
        degs.curr <- degs[i-1,tochange,1:nint[i-1,tochange]]
        dtot.curr <- sum(degs.curr)    # equivalently dtot[i-1,tochange]
        nint.curr <- length(vars.curr) # equivalently nint[i-1,tochange]
        d_probs <- seq_along(nint.curr:degree)^(-degree_penalty)
        d_probs <- d_probs / sum(d_probs)

        # Generate new degree partition
        dtot.cand <- sample(nint.curr:degree, 1, prob=d_probs)
        degs.cand <- random_partition(dtot.cand, nint.curr)

        # Generate candidate basis function
        basis <- make_basis(vars.curr, degs.cand, X)
        B.cand <- B.curr
        B.cand[,tochange+1] <- basis # replace with our new basis function (+1 for intercept)
        BtB.cand <- crossprod(B.cand)
        BtBi.cand <- tryCatch({
          safe_solve(BtB.cand)
        }, error = function(e) {
          FALSE
        })

        if(!isFALSE(BtBi.cand)){ #Otherwise reject immediately
          Vinv.cand<-crossprod(B.cand)+diag(nbasis[i-1]+1)/tau2

          b.curr <- crossprod(B.curr, z)
          q.curr <- as.numeric(crossprod(b.curr, solve(Vinv.curr, b.curr)))
          logdet.curr <- as.numeric(determinant(Vinv.curr)$mod)

          b.cand <- crossprod(B.cand, z)
          q.cand <- as.numeric(crossprod(b.cand, solve(Vinv.cand, b.cand)))
          logdet.cand <- as.numeric(determinant(Vinv.cand)$mod)

          # Compute acceptance ratio
          # Assume that "interaction" between the two mutation types is negligible
          llik.alpha <- 0.5 * (q.cand - q.curr) - 0.5 * (logdet.cand - logdet.curr)

          # Variation in the degree proposal
          lprop.alpha <- log(d_probs[1+dtot.curr-nint.curr]) - log(d_probs[1+dtot.cand-nint.curr]) +
            (-lchoose(dtot.curr, nint.curr) + lchoose(dtot.cand, nint.curr))

          # Uniform priors + no dimension change <=> no term for the prior
          alpha <- llik.alpha + lprop.alpha

          if(log(stats::runif(1))<alpha){ # accept, update
            B.curr<-B.cand
            BtB.curr <- BtB.cand
            BtBi.curr <- BtBi.cand
            Vinv.curr<-Vinv.cand

            vars[i,tochange,1:nint[i-1,tochange]] <- vars.curr # no change
            degs[i,tochange,1:nint[i-1,tochange]] <- degs.cand
            nint[i,tochange] <- nint.curr                      # no change
            dtot[i,tochange] <- dtot.cand

            count_accept[3] <- count_accept[3] + 1
          }
        }
        count_propose[3] <- count_propose[3] + 1

      }else{
        # MUTATION TYPE 2: Swap out a variable
        tochange<-sample(nbasis[i-1],1)           # which basis function we will change

        # Get current info
        vars.curr <- vars[i-1,tochange,1:nint[i-1,tochange]]
        degs.curr <- degs[i-1,tochange,1:nint[i-1,tochange]]
        dtot.curr <- sum(degs.curr)    # equivalently dtot[i-1,tochange]
        nint.curr <- length(vars.curr) # equivalently nint[i-1,tochange]

        # Choose a variable to delete
        ind_tochange <- sample(nint.curr, 1)
        var_tochange <- vars.curr[ind_tochange]
        newvar_probs.curr <- 1e-9 + coin_pars[[2]] + eta_vec
        newvar_probs.curr[var_tochange] <- 0
        newvar_probs.curr <- newvar_probs.curr / sum(newvar_probs.curr)
        newvar.curr <- sample(p, 1, prob=newvar_probs.curr)

        vars.cand <- vars.curr
        vars.cand[ind_tochange] <- newvar.curr

        # Construct probabilities for the reverse case
        eta.cand <- eta_vec
        eta.cand[var_tochange] <- eta.cand[var_tochange] - 1
        eta.cand[newvar.curr] <- eta.cand[newvar.curr] + 1
        newvar_probs.cand <- 1e-9 + coin_pars[[2]] + eta.cand
        newvar_probs.cand[newvar.curr] <- 0
        newvar_probs.cand <- newvar_probs.cand / sum(newvar_probs.cand)
        newvar.cand <- var_tochange

        # Generate candidate basis function
        basis <- make_basis(vars.cand, degs.curr, X)
        B.cand <- B.curr
        B.cand[,tochange+1] <- basis # replace with our new basis function (+1 for intercept)
        BtB.cand <- crossprod(B.cand)
        BtBi.cand <- tryCatch({
          safe_solve(BtB.cand)
        }, error = function(e) {
          FALSE
        })

        if(!isFALSE(BtBi.cand)){ # Otherwise reject immediately
          Vinv.cand<-crossprod(B.cand)+diag(nbasis[i-1]+1)/tau2

          b.curr      <- crossprod(B.curr, z)
          q.curr      <- as.numeric(crossprod(b.curr, solve(Vinv.curr, b.curr)))
          logdet.curr <- as.numeric(determinant(Vinv.curr)$mod)

          b.cand      <- crossprod(B.cand, z)
          q.cand      <- as.numeric(crossprod(b.cand, solve(Vinv.cand, b.cand)))
          logdet.cand <- as.numeric(determinant(Vinv.cand)$mod)

          # Compute acceptance ratio
          # Assume that "interaction" between the two mutation types is negligible
          llik.alpha <- 0.5 * (q.cand - q.curr) - 0.5 * (logdet.cand - logdet.curr)

          # Difference in probability of the variables.
          lprop.alpha <- log(newvar_probs.cand[newvar.cand]) - log(newvar_probs.curr[newvar.curr])

          # Uniform priors + no dimension change <=> no term for the prior
          alpha <- llik.alpha + lprop.alpha

          if(log(stats::runif(1))<alpha){ # accept, update
            B.curr<-B.cand
            BtB.curr <- BtB.cand
            BtBi.curr <- BtBi.cand
            Vinv.curr<-Vinv.cand

            vars[i,tochange,1:nint[i,tochange]] <- vars.cand
            degs[i,tochange,1:nint[i,tochange]] <- degs.curr # no change
            nint[i,tochange] <- nint.curr                    # no change
            dtot[i,tochange] <- dtot.curr                    # no change

            eta_vec <- eta.cand

            count_accept[4] <- count_accept[4] + 1
          }
        }
        count_propose[4] <- count_propose[4] + 1
      }
    }

    ## Gibbs steps

    # SAMPLE COEFFICIENTS
    Lambda_n <- BtB.curr + diag(nbasis[i] + 1)/tau2
    Lambda_i_n <- solve(Lambda_n)
    beta_ls <- BtBi.curr %*% crossprod(B.curr, z)
    mu_n <- Lambda_i_n %*% (BtB.curr %*% beta_ls)
    Sigma_n <- s2 * Lambda_i_n
    beta_curr <- t(mvtnorm::rmvnorm(1, mu_n, Sigma_n))
    beta[i,1:(nbasis[i]+1)] <- beta_curr
    zhat_curr <- B.curr %*% beta_curr
    resid <- z-zhat_curr

    # SAMPLE LAMBDA
    lam[i] <- stats::rgamma(1,h1+nbasis[i],h2+1)                                               # update lambda

    # SAMPLE LATENT Zs
    for(j in 1:n){
      ez <- zhat_curr[j]
      a <- max(-Inf, thresh_curr[y[j]-1], na.rm=TRUE)
      b <- thresh_curr[y[j]]  # min(thresh_curr[y[i]], Inf, na.rm=TRUE)
      u <- runif(1, pnorm(a-ez), pnorm(b-ez))
      z[j] <- ez + qnorm(u)
    }
    ssz <- sum(z^2)
    s2z[i] <- (ssz - sum(z)^2/n) / (n-1)

    # SAMPLE THRESHOLDS
    for(k in 2:(K-1)){
      a <- max(z[which(y == k)])
      b <- min(z[which(y == k+1)])
      u <- runif(1,
                 pnorm(a, mu_thresh[k], sigma_thresh),
                 pnorm(b, mu_thresh[k], sigma_thresh))
      thresh_curr[k] <- mu_thresh[k] + sigma_thresh * qnorm(u)
    }
    thresh_mat[i,] <- thresh_curr[1:(K-1)]

    if(verbose & i%%1000 == 0){
      pr<-c('MCMC iteration',i,myTimestamp(),'nbasis:', nbasis[i])
      cat(pr,'\n')
    }
  }

  # Trim down data structures
  mcmc_iter <- seq(nburn+1, nmcmc, by=thin)
  max_basis_used <- max(nbasis, na.rm = TRUE)
  basis_high <- if(max_basis_used > 0) 1:max_basis_used else integer(0)
  max_inter_used <- max(nint, na.rm = TRUE)
  inter_high <- if(is.finite(max_inter_used) && max_inter_used > 0) 1:max_inter_used else integer(0)


  vars <- vars[mcmc_iter, basis_high, inter_high, drop=FALSE]
  degs <- degs[mcmc_iter, basis_high, inter_high, drop=FALSE]
  nint <- nint[mcmc_iter, basis_high, drop=FALSE]
  dtot <- dtot[mcmc_iter, basis_high, drop=FALSE]

  beta <- beta[mcmc_iter, 1:(1+max(nbasis))]
  nbasis <- nbasis[mcmc_iter]
  s2z <- s2z[mcmc_iter]
  #s2 <- s2[mcmc_iter]
  lam <- lam[mcmc_iter]
  thresh_mat <- thresh_mat[mcmc_iter,,drop=FALSE]

  out <- list(B=B.curr,
              vars=vars,degs=degs,
              nint=nint,dtot=dtot,
              thresh=thresh_mat,
              nbasis=nbasis,beta=beta,
              lam=lam,
              eta=eta_vec,
              s2z=s2z,
              count_accept=count_accept,
              count_propose=count_propose,
              X=X, y=y)
  class(out) <- "ordinal_khaos"
  return(out)
}

#' Predict Method for class \code{ordinal_khaos}
#'
#' Compute posterior predictive quantities for an ordinal probit BA-PCE model.
#' For each selected MCMC draw, the method builds the polynomial chaos basis,
#' forms the linear predictor \eqn{f(x) = B(x)\beta}, reconstructs the ordered
#' cutpoints, and returns either category probabilities, MAP classes, or latent
#' means. Results can be averaged across draws or returned per-draw.
#'
#' @param object An object of class \code{ordinal_khaos} returned by \code{ordinal_khaos()}.
#' @param newdata A data frame or matrix of predictors on the same scale and with
#'   the same columns as used for fitting. If \code{NULL}, uses \code{object$X}.
#' @param mcmc.use Integer vector of posterior draw indices to use. If \code{NULL},
#'   uses all stored draws (i.e., \code{seq_along(object$nbasis)}).
#' @param type Character; one of \code{"prob"}, \code{"class"}, or \code{"latent"}.
#'   \code{"prob"} returns category probabilities \eqn{P(Y=k \mid x)} for
#'   \eqn{k=1,\dots,K}. \code{"class"} returns MAP classes. \code{"latent"} returns
#'   the latent mean \eqn{f(x)} (i.e., \eqn{B(x)\beta}) for each draw.
#' @param aggregate Logical (default \code{FALSE}); if \code{TRUE}, averages results across the
#'   selected MCMC draws (probabilities are averaged; classes are aggregated by
#'   simple voting; latent means are averaged). If \code{FALSE}, returns per-draw
#'   arrays.
#' @param ... Unused; included for method compatibility.
#'
#' @details
#' Let \eqn{Z = f_\beta(X) + \varepsilon} with \eqn{\varepsilon \sim \mathcal{N}(0,1)} and
#' ordered cutpoints \eqn{\gamma_0=-\infty < \gamma_1 < \cdots < \gamma_{K-1} < \gamma_K=\infty}.
#' For a new input \eqn{x}, the category probabilities are
#' \deqn{P(Y=k\mid x) = \Phi(\gamma_k - f(x)) - \Phi(\gamma_{k-1} - f(x)).}
#' The function reconstructs the full cutpoint vector internally. If the fitted
#' object stored \eqn{\gamma_1,\dots,\gamma_{K-1}} then these are used directly;
#' if it stored only \eqn{\gamma_2,\dots,\gamma_{K-1}} (with \eqn{\gamma_1 \equiv 0})
#' the missing fixed cutpoint is inserted automatically.
#'
#' @return
#' If \code{type = "prob"} and \code{aggregate = TRUE}, an \code{nrow(newdata) x K}
#' numeric matrix of posterior mean category probabilities. If \code{aggregate = FALSE},
#' a 3-D array of dimension \code{(#draws) x nrow(newdata) x K}.
#'
#' If \code{type = "class"} and \code{aggregate = TRUE}, an integer vector of length
#' \code{nrow(newdata)} with MAP classes aggregated by voting across draws. If
#' \code{aggregate = FALSE}, a \code{(#draws) x nrow(newdata)} integer matrix of
#' per-draw MAP classes.
#'
#' If \code{type = "latent"}, returns the latent mean(s) \eqn{f(x)}: a numeric vector
#' of length \code{nrow(newdata)} if \code{aggregate = TRUE}, or a
#' \code{(#draws) x nrow(newdata)} numeric matrix otherwise.
#'
#' @return A numeric matrix of posterior predictive samples, with rows corresponding
#' to samples and columns corresponding to observations.
#'
#' @references
#' Hoff, Peter D. (2009). *A First Course in Bayesian Statistical Methods*.
#' Springer. (See ordinal probit data augmentation.)
#'
#' Francom, Devin, and Bruno Sanso (2020). "BASS: An R package for fitting and
#' performing sensitivity analysis of Bayesian adaptive spline surfaces."
#' \emph{Journal of Statistical Software} 94. LA-UR-20-23587.
#'
#' @seealso \code{\link{ordinal_khaos}} for model fitting.
#'
#' @examples
#' \donttest{
#' # Toy example (X scaled to [0,1])
#' set.seed(1)
#' X <- lhs::maximinLHS(200, 3)
#' # Suppose y has K=4 ordered categories:
#' fit <- ordinal_khaos(X, sample(1:4, 200, replace=TRUE),
#'                      nmcmc = 2000, nburn = 1000, thin = 2, verbose = FALSE)
#'
#' # Posterior mean probabilities for each class
#' p_hat <- predict(fit, newdata = X, type = "prob", aggregate = TRUE)
#'
#' # MAP classes (aggregated across draws)
#' y_hat <- predict(fit, newdata = X, type = "class")
#'
#' # Latent mean f(x)
#' f_hat <- predict(fit, newdata = X, type = "latent")
#' }
#'
#' @export
predict.ordinal_khaos <- function(object, newdata = NULL,
                                  mcmc.use = NULL,
                                  type = c("class","prob","latent"),
                                  aggregate = FALSE, ...) {
  type <- match.arg(type)
  #print(type)
  if (is.null(newdata)) newdata <- object$X
  K <- max(object$y)
  n <- nrow(newdata)

  # which posterior draws to use
  if (is.null(mcmc.use)) mcmc.use <- seq_along(object$nbasis)
  I <- length(mcmc.use)

  # helper: build B for a given draw i
  build_B <- function(i) {
    nb <- object$nbasis[i]
    B <- matrix(1, nrow = n, ncol = nb + 1)  # intercept
    if (nb > 0) {
      for (j in seq_len(nb)) {
        J <- object$nint[i, j]
        if (is.na(J) || J == 0) next
        vars_ij <- object$vars[i, j, 1:J]
        degs_ij <- object$degs[i, j, 1:J]
        B[, j + 1] <- make_basis(vars_ij, degs_ij, newdata)
      }
    }
    B
  }

  # helper: extend thresholds to (-Inf, ..., +Inf)
  make_gamma_ext <- function(i) {
    thr <- object$thresh[i, ]
    thr <- as.numeric(thr)
    # stored length could be K-1 (gamma1..gamma_{K-1}) or K-2 (gamma2..gamma_{K-1} with gamma1 fixed at 0)
    if (length(thr) == (K - 1)) {
      # assume thr = (gamma1, ..., gamma_{K-1})
      c(-Inf, thr, Inf)
    } else if (length(thr) == (K - 2)) {
      # assume gamma1 == 0 is fixed
      c(-Inf, 0, thr, Inf)
    } else {
      stop("object$thresh has incompatible number of columns.")
    }
  }

  # allocate outputs
  if (type == "prob") {
    if (aggregate) {
      out <- matrix(0, nrow = n, ncol = K)
    } else {
      out <- array(NA_real_, dim = c(I, n, K))
    }
  } else if (type == "class") {
    out <- integer(n)  # aggregated MAP class
    if (!aggregate) {
      out <- matrix(NA_integer_, nrow = I, ncol = n)
    }
  } else { # latent
    if (aggregate) {
      out <- numeric(n)
    } else {
      out <- matrix(NA_real_, nrow = I, ncol = n)
    }
  }

  # loop over draws
  for (idx in seq_along(mcmc.use)) {
    i <- mcmc.use[idx]
    B <- build_B(i)
    beta_i <- object$beta[i, 1:ncol(B)]
    f <- as.numeric(B %*% beta_i)

    if (type == "latent") {
      if (aggregate) {
        out <- out + f / I
      } else {
        out[idx, ] <- f
      }
      next
    }

    # probs from cutpoints
    gamma_ext <- make_gamma_ext(i)  # length K+1
    # P(y=k|x) = Phi(gamma_k - f) - Phi(gamma_{k-1} - f)
    # vectorize over k by recycling f
    pk <- matrix(NA_real_, nrow = n, ncol = K)
    for (k in 1:K) {
      pk[, k] <- stats::pnorm(gamma_ext[k + 1] - f) - stats::pnorm(gamma_ext[k] - f)
    }

    if (type == "prob") {
      if (aggregate) {
        out <- out + pk / I
      } else {
        out[idx, , ] <- pk
      }
    } else { # class
      cls <- max.col(pk, ties.method = "first")
      if (aggregate) {
        # running vote: accumulate probabilities then decide at end is better,
        # but for simplicity we vote by MAP per draw
        if (idx == 1) vote <- matrix(0, nrow = n, ncol = K)
        vote[cbind(seq_len(n), cls)] <- vote[cbind(seq_len(n), cls)] + 1
        if (idx == I) out <- max.col(vote, ties.method = "first")
      } else {
        out[idx, ] <- cls
      }
    }
  }

  return(out)
}


#' Plot Method for class \code{ordinal_khaos}
#'
#' Produces a 2x2 diagnostic plot: (1) trace of the number of basis functions;
#' (2) jittered predicted vs. actual classes with point size indicating posterior
#' support; (3) posterior mean class probabilities on \code{x$X} (lines if 1D,
#' otherwise barplot of averages); (4) histogram of latent means \eqn{f(x)} with
#' posterior-average thresholds overlaid.
#'
#' @param x An object of class \code{ordinal_khaos}.
#' @param ... additional arguments passed to \code{plot}
#'
#' @details
#' Panel (2) uses \code{predict(x, type="class", aggregate=FALSE)} to compute
#' per-draw MAP classes, plotting the modal class vs. observed class; point size
#' reflects the fraction of draws supporting the modal class. Panel (3) uses
#' \code{predict(x, type="prob")} to plot posterior mean class probabilities.
#' Panel (4) uses \code{predict(x, type="latent", aggregate=FALSE)} to pool
#' latent means across draws and overlays posterior-average cutpoints (handles
#' either stored \eqn{\gamma_1,\dots,\gamma_{K-1}} or \eqn{\gamma_2,\dots,\gamma_{K-1}}
#' with \eqn{\gamma_1\equiv 0}).
#'
#' @return Invisibly returns \code{NULL}.
#'
#' @examples
#' \donttest{
#' set.seed(1)
#' X <- matrix(runif(200), ncol = 1)
#' y <- sample(1:4, 200, replace = TRUE)
#' fit <- ordinal_khaos(X, y, nmcmc = 2000, nburn = 1000, thin = 2, verbose = FALSE)
#' plot(fit)
#' }
#' @export
plot.ordinal_khaos <- function(x, ...){
  oldpar <- par(no.readonly = TRUE)
  on.exit(par(oldpar))
  par(mfrow = c(2, 2))

  K <- max(x$y)

  ## --- (1) Trace of nbasis ---
  plot(seq_along(x$nbasis), x$nbasis, type = "l", lwd = 1,
       xlab = "MCMC iteration", ylab = "Number of basis functions",
       main = "Trace: nbasis", ...)

  ## --- (2) Predicted class (uncertainty) vs actual class ---
  cls_draws <- predict(x, newdata = x$X, type = "class", aggregate = FALSE)
  if (is.vector(cls_draws)) cls_draws <- matrix(cls_draws, nrow = 1)
  n <- ncol(cls_draws)

  modal_class <- integer(n)
  support <- numeric(n)
  for (j in seq_len(n)) {
    tabj <- tabulate(cls_draws[, j], nbins = K)
    modal_class[j] <- which.max(tabj)
    support[j] <- max(tabj) / nrow(cls_draws)
  }
  # map support to point size
  cex_j <- 0.2 + 1.6 * (support - min(support)) /
    max(1e-12, (max(support) - min(support)))

  plot(jitter(x$y, 0.15), jitter(modal_class, 0.15),
       xlab = "Actual class", ylab = "Predicted class (modal)",
       xlim = c(0.5, K + 0.5), ylim = c(0.5, K + 0.5),
       axes = FALSE, pch = 19, col = "black", cex = cex_j,
       main = "Predicted vs actual (size = support)")
  abline(0,1,col='red')
  axis(1, at = 1:K, labels = 1:K); axis(2, at = 1:K, labels = 1:K, las = 1); box()

  ## --- (3) Posterior mean class probabilities on training X ---
  probs <- predict(x, newdata = x$X, type = "prob", aggregate = TRUE)
  if (ncol(x$X) == 1) {
    ord <- order(x$X[, 1])
    matplot(x$X[ord, 1], probs[ord, ], type = "l", lty = 1, lwd = 2,
            xlab = "X", ylab = "Predicted probability",
            main = "Posterior mean class probabilities", col = seq_len(K))
    legend("topright", legend = paste("Class", 1:K),
           col = seq_len(K), lty = 1, lwd = 2, bty = "n")
  } else {
    barplot(colMeans(probs), names.arg = paste("Class", 1:K),
            ylab = "Mean predicted probability",
            main = "Avg class probabilities (train)", col = seq_len(K))
  }

  ## --- (4) Histogram of latent means f(x) with avg thresholds ---
  f_draws <- predict(x, newdata = x$X, type = "latent", aggregate = FALSE)  # I x n
  f_vec <- as.numeric(f_draws)  # pool across draws & cases
  hist(f_vec, breaks = "FD", freq = TRUE, col = "grey85", border = "black",
       xlab = "Latent mean f(x)", main = "Latent means with avg thresholds")

  # compute posterior-average thresholds
  thresh <- x$thresh
  thr_len <- ncol(thresh)
  if (thr_len == (K - 1)) {
    # assume thresh = (gamma1..gamma_{K-1})
    gamma_mean <- c(-Inf, colMeans(thresh), Inf)
  } else if (thr_len == (K - 2)) {
    # assume gamma1 == 0 is fixed
    gamma_mean <- c(-Inf, 0, colMeans(thresh), Inf)
  } else {
    gamma_mean <- c(-Inf, 0, Inf)  # fallback to avoid errors
  }

  # draw interior thresholds only (exclude +/-Inf)
  if (length(gamma_mean) >= 3) {
    g_int <- gamma_mean[is.finite(gamma_mean)]
    abline(v = g_int, col = "red", lwd = 2, lty = 2)
    legend("topright", legend = "Avg thresholds", lty = 2, lwd = 2, col = "red", bty = "n")
  }

  invisible(NULL)
}

test_that("simple ordinal_khaos test", {
  cat("simple ordinal_khaos test\n")

  set.seed(1)

  n <- 200
  p <- 2
  X <- matrix(runif(n * p), ncol = p)

  # latent function
  f <- function(x) 2 * x[1] - x[2]
  eta <- apply(X, 1, f)

  # generate latent z and discretize into K=3 classes
  z <- eta + rnorm(n, 0, 0.3)

  # choose thresholds to ensure all classes appear
  brks <- quantile(z, probs = c(0, 1/3, 2/3, 1))
  y <- cut(z, breaks = brks, include.lowest = TRUE, labels = FALSE)

  # sanity check (should always pass with quantiles)
  expect_true(length(unique(y)) == 3)

  fit <- ordinal_khaos(
    X, y,
    nmcmc = 500,
    nburn = 250,
    thin = 2,
    verbose = FALSE
  )

  # predicted probabilities
  probs <- predict(fit, newdata = X, type = "prob", aggregate = TRUE)

  # predicted class
  yhat <- max.col(probs)

  # classification accuracy
  acc <- mean(yhat == y)

  # not too strict — just ensure model learns something
  expect_true(acc > 0.5)
})

build_candidate_set <- function(p, d, q, A_curr, enrichment, max_basis){
  if(is.null(A_curr) || nrow(A_curr) == 0){
    return(generate_A(p, d, q))
  }

  # Check size for full enrichment
  if(enrichment == 4){
    A_num <- A_size(p, d, q)
    if(A_num > max_basis){
      stop("Too many basis functions for full rebuild. Increase max_basis or use a more restrictive enrichment strategy.")
    }
    return(generate_A(p, d, q))
  }

  enrich_fun <- get(paste0("enrichment_", enrichment), mode = "function")
  enrich_fun(p, d, q, A_curr)
}

enrichment_0 <- function(p, d, q, A_curr){
  A_new <- list(rep(0, p)) # Initialize with intercept

  # Check to see if we need to filter based on q (Yes, if q_curr >= q)
  q_curr <- max(apply(A_curr, 1, function(alpha) sum(alpha > 0)))

  # Loop over alpha vectors
  for(i in seq_len(nrow(A_curr))){
    alpha <- A_curr[i,]

    # Start tracking the enriched versions
    enriched_list <- list(alpha)

    for(j in seq_len(p)){
      # Do deletion and bind to A_new (set to 0 all values >= 1)
      if(alpha[j] >= 1){
        tmp <- alpha
        tmp[j] <- 0
        enriched_list <- append(enriched_list, list(tmp))
      }

      # Do simplification and bind to A_new (-1 to all values >= 2)
      if(alpha[j] >= 2){
        tmp <- alpha
        tmp[j] <- tmp[j] - 1
        enriched_list <- append(enriched_list, list(tmp))
      }

      # Do promotion and bind to A_new (set to d, all values < d)
      if(alpha[j] <= d-1){
        tmp <- alpha
        tmp[j] <- d
        enriched_list <- append(enriched_list, list(tmp))
      }

      # Do complication and bind to A_new (+1 to all values < d-1)
      if(alpha[j] <= d-2){
        tmp <- alpha
        tmp[j] <- tmp[j] + 1
        enriched_list <- append(enriched_list, list(tmp))
      }
    }
    A_new <- append(A_new, enriched_list)
  }

  A_new_mat <- unique(do.call(rbind, A_new))   # Remove duplicates
  A_new_mat <- A_new_mat[-1,,drop=FALSE]       # Remove intercept

  # Screen for max order only if necesarry
  # This works because effective q can be at most q_curr + 1.
  if(q_curr >= q){
    valid_idx <- rowSums(A_new_mat > 0) <= q
    A_new_mat <- A_new_mat[valid_idx, , drop = FALSE]
  }
  return(A_new_mat)
}

# Helper to generate the candidates only using certain active vars
generate_A_active <- function(p, d, q, active_vars){
  p_active <- length(active_vars)

  # Generate basis functions in reduced active variable space
  A_reduced <- generate_A(p_active, d, q)
  M <- nrow(A_reduced)

  # Embed into full p-dimensional space
  A_full <- matrix(0L, nrow = M, ncol = p)
  A_full[, active_vars] <- A_reduced

  return(A_full)
}

# Shao version
enrichment_1 <- function(p, d, q, A_curr){
  # Find the active vars
  activity <- apply(A_curr, 2, sum)
  index_active <- which(activity > 0)

  A_new <- generate_A_active(p, d, q, index_active)
  return(A_new)
}

enrichment_2 <- function(p, d, q, A_curr){
  # Find the active vars
  activity <- apply(A_curr, 2, sum)
  index_active <- which(activity > 0)
  index_inactive <- (1:p)[-index_active]

  # Check to see if we need to filter based on q (Yes, if q_curr >= q)
  q_curr <- max(apply(A_curr, 1, function(alpha) sum(alpha > 0)))

  # Build full-active
  A_new <- list(generate_A_active(p, d, q, index_active))

  # Return early if everything is active
  if(length(index_inactive) == 0)
    return(A_new[[1]])

  # Enrich inactives around the previously accepted terms
  for(i in seq_len(nrow(A_curr))){
    alpha <- A_curr[i,]
    for(j in index_inactive){
      # Do promotion and bind to A_new (set to d, all values < d)
      if(alpha[j] <= d-1){
        tmp <- alpha
        tmp[j] <- d
        A_new <- append(A_new, list(tmp))
      }

      # Do complication and bind to A_new (+1 to all values < d-1)
      if(alpha[j] <= d-2){
        tmp <- alpha
        tmp[j] <- tmp[j] + 1
        A_new <- append(A_new, list(tmp))
      }
    }
  }

  # Combine original + enriched
  A_new_mat <- do.call(rbind, A_new)

  # Screen for max order only if necessary
  # This works because effective-q can be at most q_curr + 1.
  if(q_curr >= q){
    valid_idx <- rowSums(A_new_mat > 0) <= q
    A_new_mat <- A_new_mat[valid_idx, , drop = FALSE]
  }
  return(A_new_mat)
}

enrichment_3 <- function(p, d, q, A_curr){
  # Find the active vars
  activity <- apply(A_curr, 2, sum)
  index_active <- which(activity > 0)
  index_inactive <- (1:p)[-index_active]

  # Check to see if we need to filter based on q (Yes, if q_curr >= q)
  q_curr <- max(apply(A_curr, 1, function(alpha) sum(alpha > 0)))

  # Build full-active
  A_base <- generate_A_active(p, d, q, index_active)
  A_new <- list(A_base)

  # Return early if everything is active
  if(length(index_inactive) == 0)
    return(A_base)

  # Enrich inactives around the previously accepted terms
  for(i in seq_len(nrow(A_base))){
    alpha <- A_base[i,]
    for(j in index_inactive){
      # Do promotion and bind to A_new (set to d, all values < d)
      if(alpha[j] <= d-1){
        tmp <- alpha
        tmp[j] <- d
        A_new <- append(A_new, list(tmp))
      }

      # Do complication and bind to A_new (+1 to all values < d-1)
      if(alpha[j] <= d-2){
        tmp <- alpha
        tmp[j] <- tmp[j] + 1
        A_new <- append(A_new, list(tmp))
      }
    }
  }

  # Combine original + enriched
  A_new_mat <- do.call(rbind, A_new)

  # Screen for max order only if necessary
  # This works because effective-q can be at most q_curr + 1.
  if(q_curr >= q){
    valid_idx <- rowSums(A_new_mat > 0) <= q
    A_new_mat <- A_new_mat[valid_idx, , drop = FALSE]
  }
  return(A_new_mat)
}

# Generate everything; A_curr is ignored
enrichment_4 <- function(p, d, q, A_curr){
  generate_A(p, d, q)
}

#' Compute Sobol Sensitivity Indices for khaos Models
#'
#' Computes first-order, total-effect, and full interaction Sobol indices
#' from an adaptive KHAOS model fit. Returns MCMC samples of sensitivity
#' indices, including all active interaction terms used in the model.
#'
#' @param obj An object of class \code{"adaptive_khaos"} returned by \code{adaptive_khaos()} or of class \code{"sparse_khaos"} returned by \code{sparse_khaos()}.
#' @param plot_it Logical; if `TRUE`, boxplots of the sensitivity indices are shown.
#' @param samples How many samples from the posterior (only used for sparse class).
#' @return A list with the following components:
#' \describe{
#'   \item{S}{A data frame of Sobol indices for all main effects and interactions (MCMC samples).}
#'   \item{T}{A data frame of total-effect Sobol indices (MCMC samples).}
#'   \item{leftover}{A numeric vector of unexplained variance proportions (residuals).}
#' }
#'
#' @examples
#' \donttest{
#' X <- lhs::maximinLHS(100, 2)
#' f <- function(x) 10.391*((x[1]-0.4)*(x[2]-0.6) + 0.36)
#' y <- apply(X, 1, f) + stats::rnorm(100, 0, 0.1)
#' fit <- adaptive_khaos(X, y)
#' sob <- sobol_khaos(fit)
#' head(sob$S)
#' }
#'
#' @export
sobol_khaos <- function(obj, plot_it = TRUE, samples=1000) {
  if (inherits(obj, "sparse_khaos")) {
    return(sobol_sparse_khaos(obj, samples, plot_it))
  } else {
    if (!inherits(obj, "ordinal_khaos") && !inherits(obj, "adaptive_khaos")) {
      stop("obj must be an object of class sparse_khaos, adaptive_khaos, or ordinal_khaos.")
    }
  }

  beta <- obj$beta
  vars <- obj$vars
  degs <- obj$degs
  nint <- obj$nint
  X    <- obj$X
  if (inherits(obj, "ordinal_khaos")) {
    s2 <- obj$s2z
  }else{
    s2 <- obj$s2
  }

  n_iter <- dim(beta)[1]
  M <- dim(beta)[2] - 1
  p <- ncol(X)
  var_names <- paste0("x", seq_len(p))

  first_order_mat <- matrix(NA, nrow = n_iter, ncol = p)
  total_effect_mat <- matrix(NA, nrow = n_iter, ncol = p)
  leftover <- rep(NA, n_iter)

  interaction_list <- vector("list", n_iter)
  all_labels_set <- new.env(hash = TRUE)

  for (iter in 1:n_iter) {
    b <- beta[iter, -1]  # exclude intercept
    b2 <- b^2
    explained <- sum(b2, na.rm = TRUE)
    residual <- s2[iter]
    total_var <- explained + residual
    if (total_var < 1e-12 || is.na(total_var)) next

    alpha_mat <- matrix(0, nrow = M, ncol = p)
    interact_map <- list()

    for (m in 1:M) {
      k <- nint[iter, m]
      if (is.na(k)) next
      if (k == 0) next
      idxs <- vars[iter, m, 1:k]
      degrees <- degs[iter, m, 1:k]
      alpha_mat[m, idxs] <- degrees

      idxs_sorted <- sort(idxs)
      label <- paste0("x", idxs_sorted, collapse = ":")

      # Store label globally
      all_labels_set[[label]] <- TRUE

      if (label %in% names(interact_map)) {
        interact_map[[label]] <- interact_map[[label]] + b2[m]
      } else {
        interact_map[[label]] <- b2[m]
      }
    }

    # Normalize interaction values
    interaction_list[[iter]] <- sapply(interact_map, function(val) val / total_var)

    for (j in 1:p) {
      only_j <- which(alpha_mat[, j] > 0 & rowSums(alpha_mat[, -j, drop = FALSE]) == 0)
      any_j  <- which(alpha_mat[, j] > 0)
      first_order_mat[iter, j] <- sum(b2[only_j], na.rm = TRUE) / total_var
      total_effect_mat[iter, j] <- sum(b2[any_j], na.rm = TRUE) / total_var
    }

    leftover[iter] <- residual / total_var
  }

  # Construct full set of labels
  all_labels <- ls(all_labels_set)
  # Sort by interaction order, then alphabetically
  if (length(all_labels) > 0) {
    label_order <- order(sapply(strsplit(all_labels, ":"), length), all_labels)
    all_labels <- all_labels[label_order]
  }

  # Construct full matrix for S
  S_mat <- matrix(0, nrow = n_iter, ncol = length(all_labels))
  colnames(S_mat) <- all_labels

  for (i in 1:n_iter) {
    terms <- interaction_list[[i]]
    if (length(terms) > 0) {
      matched <- match(names(terms), all_labels)
      S_mat[i, matched] <- terms
    }
  }

  # Output
  colnames(first_order_mat) <- var_names
  colnames(total_effect_mat) <- var_names

  out <- list(
    S = as.data.frame(S_mat),
    T = as.data.frame(total_effect_mat),
    leftover = leftover
  )

  if (plot_it) {
    oldpar <- par(no.readonly = TRUE) # code line i
    on.exit(par(oldpar)) # code line i + 1
    # Add plotting later
    tmp <- cbind(out$S, leftover)
    boxplot(tmp, main="Sensitivity")
    tmp <- cbind(out$T, leftover)
    boxplot(tmp, main="Total Sensitivity")
  }

  return(out)
}


sobol_sparse_khaos <- function(obj, samples = 1000, plot_it = TRUE) {
  X <- obj$X
  y <- obj$y
  beta_hat <- obj$beta_hat
  G <- obj$G
  BtBi <- obj$BtBi
  s2 <- obj$s2
  prior <- obj$prior
  nbasis <- obj$nbasis
  vars <- obj$vars
  p <- ncol(X)
  n <- nrow(X)

  # Setup for sampling from posterior over coefficients
  v0 <- prior[2]
  ntrain <- length(y)
  shape <- (ntrain + v0 - nbasis) / 2
  coeff_samples <- matrix(NA, nrow = samples, ncol = nbasis)
  sigma2_samples <- numeric(samples)

  for (i in 1:samples) {
    sigma2 <- 1 / stats::rgamma(1, shape, rate=s2 * shape)
    sigma2_samples[i] <- sigma2
    cov_mat <- sigma2 * (G * BtBi)
    coeff_samples[i, ] <- stats::rnorm(nbasis, mean = beta_hat, sd = sqrt(diag(cov_mat)))
  }

  # Prepare for Sobol decomposition
  var_names <- paste0("x", seq_len(p))
  first_order_mat <- matrix(NA, nrow = samples, ncol = p)
  total_effect_mat <- matrix(NA, nrow = samples, ncol = p)
  leftover <- numeric(samples)
  interaction_list <- vector("list", samples)
  all_labels_set <- new.env(hash = TRUE)

  for (iter in 1:samples) {
    b <- coeff_samples[iter, ]
    b2 <- b^2
    explained <- sum(b2)
    residual <- sigma2_samples[iter]
    total_var <- explained + residual
    if (total_var < 1e-12 || is.na(total_var)) next

    alpha_mat <- vars
    alpha_mat[is.na(alpha_mat)] <- 0

    interact_map <- list()

    for (m in 1:nbasis) {
      idxs <- which(alpha_mat[m, ] > 0)
      if (length(idxs) == 0) next

      label <- paste0("x", sort(idxs), collapse = ":")
      all_labels_set[[label]] <- TRUE

      if (label %in% names(interact_map)) {
        interact_map[[label]] <- interact_map[[label]] + b2[m]
      } else {
        interact_map[[label]] <- b2[m]
      }
    }

    interaction_list[[iter]] <- sapply(interact_map, function(val) val / total_var)

    for (j in 1:p) {
      only_j <- which(alpha_mat[, j] > 0 & rowSums(alpha_mat[, -j, drop = FALSE]) == 0)
      any_j  <- which(alpha_mat[, j] > 0)
      first_order_mat[iter, j] <- sum(b2[only_j]) / total_var
      total_effect_mat[iter, j] <- sum(b2[any_j]) / total_var
    }

    leftover[iter] <- residual / total_var
  }

  all_labels <- ls(all_labels_set)
  if (length(all_labels) > 0) {
    label_order <- order(sapply(strsplit(all_labels, ":"), length), all_labels)
    all_labels <- all_labels[label_order]
  }

  S_mat <- matrix(0, nrow = samples, ncol = length(all_labels))
  colnames(S_mat) <- all_labels

  for (i in 1:samples) {
    terms <- interaction_list[[i]]
    if (length(terms) > 0) {
      matched <- match(names(terms), all_labels)
      S_mat[i, matched] <- terms
    }
  }

  colnames(first_order_mat) <- var_names
  colnames(total_effect_mat) <- var_names

  out <- list(
    S = as.data.frame(S_mat),
    T = as.data.frame(total_effect_mat),
    leftover = leftover
  )

  if (plot_it) {
    oldpar <- par(no.readonly = TRUE)
    on.exit(par(oldpar))
    tmp <- cbind(out$S, leftover)
    boxplot(tmp, main = "Sensitivity")
    tmp <- cbind(out$T, leftover)
    boxplot(tmp, main = "Total Sensitivity")
  }

  return(out)
}

#' Bayesian Adaptive Polynomial Chaos Expansion
#'
#' A wrapper for adaptive Bayesian polynomial chaos expansion models with
#' different coefficient prior formulations.
#'
#' Depending on \code{prior_type}, this function calls either
#' \code{adaptive_khaos_ridge()} or \code{adaptive_khaos_gprior()}.
#'
#' @param X A data frame or matrix of predictors scaled to lie between 0 and 1.
#' @param y A numeric response vector of length \code{nrow(X)}.
#' @param prior_type Character string specifying the coefficient-prior formulation.
#'   Currently one of \code{"ridge"} or \code{"gprior"}.
#' @param nmcmc Number of MCMC iterations.
#' @param nburn Number of initial MCMC iterations to discard.
#' @param thin Keep every \code{thin}-th retained MCMC sample.
#' @param legacy Logical. If \code{TRUE}, reproduces the historical storage behavior
#'   of earlier implementations, which may retain stale values in unused portions of
#'   the internal \code{vars} and \code{degs} arrays. If \code{FALSE}, these unused
#'   entries are cleared more carefully.
#' @param verbose Logical; if \code{TRUE}, progress messages are printed.
#' @param ... Additional arguments passed to the prior-specific fitting routine.
#'
#' @details
#' This function dispatches to one of the prior-specific adaptive KHAOS fitting
#' routines:
#' \describe{
#'   \item{\code{prior_type = "ridge"}}{Calls \code{\link{adaptive_khaos_ridge}}.}
#'   \item{\code{prior_type = "gprior"}}{Calls \code{\link{adaptive_khaos_gprior}}.}
#' }
#'
#' The wrapper exposes only the most commonly used arguments. Additional
#' prior-specific tuning parameters can be supplied through \code{...}; see
#' \code{\link{adaptive_khaos_ridge}} and \code{\link{adaptive_khaos_gprior}}
#' for full documentation.
#'
#' @return An object of class \code{"adaptive_khaos"} containing the fitted model.
#' The object is a list including sampled basis functions, coefficient samples,
#' variance parameters, and additional MCMC output. The specific structure depends
#' on the chosen \code{prior_type}.
#'
#' @references
#' Francom, Devin, and Bruno Sanso. "BASS: An R package for fitting and performing
#' sensitivity analysis of Bayesian adaptive spline surfaces."
#' \emph{Journal of Statistical Software} 94.LA-UR-20-23587 (2020).
#'
#' Rumsey, K.N., Francom, D., Gibson, G., Tucker, J. D., and Huerta, G. (2026).
#' Bayesian Adaptive Polynomial Chaos Expansions. \emph{Stat}, 15(1), e70151.
#'
#' @examples
#' \donttest{
#' X <- lhs::maximinLHS(100, 2)
#' f <- function(x) 10.391 * ((x[1] - 0.4) * (x[2] - 0.6) + 0.36)
#' y <- apply(X, 1, f) + stats::rnorm(100, 0, 0.1)
#'
#' fit1 <- adaptive_khaos(X, y)
#' fit2 <- adaptive_khaos(X, y, prior_type = "gprior")
#' }
#' @export
adaptive_khaos <- function(X, y,
                           prior_type = "ridge",
                           nmcmc = 10000,
                           nburn = 9000,
                           thin = 1,
                           legacy = TRUE,
                           verbose = TRUE,
                           ...) {
  prior_type <- match.arg(prior_type, c("ridge", "gprior"))

  if (prior_type == "ridge") {
    adaptive_khaos_ridge(
      X = X,
      y = y,
      nmcmc = nmcmc,
      nburn = nburn,
      thin = thin,
      legacy = legacy,
      verbose = verbose,
      ...
    )
  } else {
    adaptive_khaos_gprior(
      X = X,
      y = y,
      nmcmc = nmcmc,
      nburn = nburn,
      thin = thin,
      legacy = legacy,
      verbose = verbose,
      ...
    )
  }
}

#' Bayesian Adaptive Polynomial Chaos Expansion (Modified g-Prior)
#'
#' Adaptive Bayesian polynomial chaos expansion using a modified g-prior. This
#' implementation corresponds to the adaptive Bayesian PCE methodology described
#' in Rumsey et al. (2026)
#'
#' @param X A data frame or matrix of predictors scaled to be between 0 and 1
#' @param y a response vector
#' @param degree Maximum polynomial degree for each basis function.
#' @param order Maximum order of interaction for each basis function.
#' @param nmcmc Number of MCMC iterations.
#' @param nburn Number of initial MCMC iterations to discard.
#' @param thin Keep every \code{thin} samples.
#' @param max_basis Maximum number of basis functions.
#' @param a_g,b_g Prior parameters for global g-prior term (on precision scale)
#' @param zeta Hyperparameter for modified g-prior. Setting zeta = 0 reduces to the usual g-prior.
#' @param g2_sample Character string specifying the method used to sample or update the prior scaling parameter \code{g0^2}.
#' @param g2_init Initial calue of g2 (global precision for g-prior). Becomes the only value when \code{g2_sample = "fixed"}.
#' @param s2_lower Lower bound on process variance (numerically useful for deterministic functions).
#' @param a_sigma,b_sigma Shape/rate parameters for the IG prior on process variance (default is Jefffrey's prior)
#' @param a_M,b_M Shape/rate parameters for the Gamma prior on expected number of basis functions.
#' @param move_probs A 3-vector with probabilities for (i) birth, (ii) death, and (iii) mutation.
#' @param coin_pars A list of control parameters for coinflip proposal
#' @param degree_penalty Increasing this value encourages lower order polynomial terms (0 is no penalization).
#' @param legacy Logical. If TRUE, mimics original implementation behavior (from the emulator comparison paper),
#'        which may retain invalid basis function structures across iterations. Defaults to FALSE.
#' @param verbose Logical. Should progress information be printed?
#' @details Implements the RJMCMC algorithm described by Francom & Sanso (2020) for BMARS, modifying it for polynomial chaos basis functions.
#' As an alternative the NKD procedure of Nott et al. (2005), we use a coinflipping procedure to identify useful variables. See writeup (coming soon) for details.
#' The coin_pars argument is an ordered list with elements:
#' \enumerate{
#' \item{a function giving proposal probability for q_0, the expected order of interaction,}
#' \item{epsilon, the base weight for each variable (epsilon=Inf corresponds to uniform sampling),}
#' \item{alpha, the exponent-learning rate,}
#' \item{num_passes a numerical parameter only.}
#' }
#'
#' Sampling method for g0^2 depends on \code{g2_sample}
#' \describe{
#'   \item{\code{"f"}}{Fixed: Do not update \code{g0^2}; carry forward the previous value (\code{g2[i] <- g2[i-1]}).}
#'   \item{\code{"lf"}}{Laplace-Full: Use a Laplace approximation to sample \code{g0^2} under the full marginal likelihood (no orthogonality assumption).}
#'   \item{\code{"lo"}}{Laplace-Orthogonal: Use a Laplace approximation to sample \code{g0^2} assuming the orthogonality approximation applies.}
#'   \item{\code{"mh"}}{Metropolis-Hastings with full Laplace approximation as proposal. Target is the full posterior density of \code{g0^2}.}
#'   \item{\code{"mho"}}{Metropolis-Hastings with orthogonal Laplace approximation as proposal. Target is the full posterior density of \code{g0^2}. under the orthogonality assumption.}
#'   \item{\code{"mhoo"}}{Metropolis-Hastings with orthogonal Laplace approximation as proposal. Target is the posterior density of \code{g0^2} under the orthogonality assumption.}
#' }
#'
#' Other proposal notes:
#'    Expected order q0 is chosen from 1:order with weights `coin_pars[[1]](1:order)`
#'    Degree is chosen from q0:degree with weights `(1:(degree-q0+1))^(-degree_penalty)` and a random partition is created (see helper function).
#'
#' Two types of change steps are possible (and equally likely)
#'    A re-ordering of the degrees
#'    A single variable is swapped out with another one (only used when \code{ncol(X) > 3})
#'
#' @return An object of class \code{"adaptive_khaos"}.
#' The object is a list containing:
#' \describe{
#'   \item{B}{Matrix of basis functions evaluated at the training inputs.}
#'   \item{vars, degs}{Arrays encoding the variables and polynomial degrees for each basis function across MCMC iterations.}
#'   \item{nint, dtot}{Interaction order and total degree for each basis function.}
#'   \item{nbasis}{Number of basis functions per MCMC iteration.}
#'   \item{beta}{Posterior samples of regression coefficients.}
#'   \item{s2}{Posterior samples of the noise variance.}
#'   \item{lam}{Posterior samples of the Poisson rate parameter controlling model size.}
#'   \item{g2}{Posterior samples of the global g-prior scaling parameter.}
#'   \item{sum_sq}{Residual sum of squares at each iteration.}
#'   \item{eta}{Variable inclusion counts used in the proposal distribution.}
#'   \item{count_accept, count_propose}{Acceptance and proposal counts for RJMCMC moves.}
#'   \item{prior_type}{Character string indicating the g-prior formulation.}
#'   \item{X, y}{Training data.}
#' }
#' @references
#' Francom, Devin, and Bruno Sanso. "BASS: An R package for fitting and performing sensitivity analysis of Bayesian adaptive spline surfaces." Journal of Statistical Software 94.LA-UR-20-23587 (2020).
#'
#' Nott, David J., Anthony YC Kuk, and Hiep Duc. "Efficient sampling schemes for Bayesian MARS models with many predictors." Statistics and Computing 15 (2005): 93-101.
#'
#' Rumsey, K.N., Francom, D., Gibson, G., Derek Tucker, J. and Huerta, G., 2026. Bayesian Adaptive Polynomial Chaos Expansions. Stat, 15(1), p.e70151.
#' @examples
#' \donttest{
#' X <- lhs::maximinLHS(100, 2)
#' f <- function(x) 10.391*((x[1]-0.4)*(x[2]-0.6) + 0.36)
#' y <- apply(X, 1, f) + stats::rnorm(100, 0, 0.1)
#' fit <- adaptive_khaos(X, y)
#' }
#' @export
adaptive_khaos_gprior <-function(X, y,
                          degree=15, order=5,
                          nmcmc=10000,
                          nburn=9000,
                          thin=1,
                          max_basis=1000,
                          a_g = 1e-3, b_g=1e3, zeta=1,
                          g2_sample="mh",
                          g2_init=NULL,
                          s2_lower=0,
                          a_sigma=0,b_sigma=0,
                          a_M=4,b_M=4/length(y),
                          move_probs=rep(1/3, 3),
                          coin_pars=list(function(j) 1/j, 1, 2, 3),
                          degree_penalty=0,
                          legacy=TRUE,
                          verbose=TRUE
){
  # Define a safer solve
  # Note: Under iid uniform X, columns of B are uncorrelated.
  #       So the tolerance should be pretty safe unless there's an
  #       extreme amount of correlation in X.
  # safe_solve <- function(A, tol = 1e-9) {
  #   rc <- rcond(A)
  #   if (is.na(rc) || rc < tol) return(FALSE)
  #   solve(A)
  # }
  #
  safe_solve <- function(A, tol = 1e-9) {
    rc <- rcond(A)
    if (is.na(rc) || rc < tol) return(FALSE)
    chol_A <- tryCatch(chol(A), error = function(e) return(FALSE))
    if (isFALSE(chol_A)) return(FALSE)
    return(chol2inv(chol_A))
  }

  # Make sure X is a matrix
  if(!is.matrix(X)){
    if(is.data.frame(X)){
      X <- as.matrix(X)
    }else if(is.numeric(X)){
      X <- matrix(X, ncol = 1)
    }else{
      stop("X must be a matrix, data frame, or numeric vector.")
    }
  }
  # Check inputs and assign globals
  if(max(X) > 1 || min(X) < 0) warning("Inputs are expected to be scaled on (0, 1). Is this intentional?")
  if(length(y) != nrow(X)) stop("length(y) must equal nrow(X)")
  n<-length(y)
  p<-ncol(X)
  ssy<-sum(y^2)
  if(p < order){
    order <- p
  }
  if(degree < order){
    stop("degree must be at least as large as order")
  }
  #KR: generalized harmonic number
  J_probs <- coin_pars[[1]](1:order)
  Hjs <- sum(J_probs)
  A_total <- A_size(p, degree, order)


  # storage
  degs<-vars<-array(dim=c(nmcmc,max_basis,order)) # this is the largest possible, our filling will be ragged

  nint<-dtot<-matrix(nrow=nmcmc,ncol=max_basis) # order of interaction J, and total degree, again filling will be ragged
  beta<-matrix(nrow=nmcmc,ncol=max_basis+1) # +1 for intercept, again filling will be ragged
  s2<-lam<-nbasis<-rep(NA,nmcmc) # error variance, poisson hyperprior, and number of basis functions
  sum_sq <- rep(NA, nmcmc)
  eta_vec <- rep(0, p) # KR: count how many times each variable is added to the model

  # initialize
  nbasis[1]<-0
  s2[1]<-stats::var(y) + 1e-9
  lam[1]<-1
  B.curr<-matrix(rep(1,n)) # matrix of current basis functions, so that yhat = B.curr %*% beta
  BtB.curr <- crossprod(B.curr)
  #BtBi.curr <- solve(BtB.curr)
  #==================================================
  # Modified G-Prior Stuff
  g2 <- rep(NA, nmcmc)
  if(is.null(g2_init)){
    g2[1] <- b_g / (a_g + 1) # Prior mode
  }else{
    g2[1] <- g2_init
  }
  g_vec.curr <- 1          # Intercept only right now
  G.curr <- 1 + tcrossprod(1/g_vec.curr) / g2[1]
  #==================================================
  Sigma.curr <- solve(BtB.curr * G.curr)
  v.curr <- crossprod(B.curr, y)
  quad.curr <- ssy - t(v.curr) %*% Sigma.curr %*% v.curr
  Q.curr <- b_sigma + 0.5 * quad.curr
  ldet.curr <- determinant(Sigma.curr, logarithm = TRUE)$modulus

  #Vinv.curr<-crossprod(B.curr)+1/tau2 # V^(-1) from DMS
  #bhat<-solve(Vinv.curr)%*%t(B.curr)%*%y
  #d.curr <- b_sigma + ssy - t(bhat)%*%Vinv.curr%*%bhat # d from DMS

  count_accept<-c(0,0,0,0)  # count how many times we accept birth, death, change degree, change vars
  count_propose<-c(0,0,0,0) # count how many times we propose each step
  names(count_accept) <- c("Birth", "Death", "Mutate (degree)", "Mutate (vars)")
  names(count_propose) <- c("Birth", "Death", "Mutate (degree)", "Mutate (vars)")
  beta[1,1] <- mean(y)

  if(verbose){
    pr<-c('MCMC iteration',0,myTimestamp(),'nbasis:', 0)
    cat(pr,'\n')
  }

  # start MCMC
  for(i in 2:nmcmc){
    ## Reversible jump step
    move.type<-sample(c('birth','death','change'),1,prob=move_probs)
    if(nbasis[i-1]==0)
      move.type<-'birth'
    if(nbasis[i-1]==max_basis)
      move.type<-sample(c('death','change'),1,prob=move_probs[2:3])

    # set all of this iterations values to last iteration values...we'll change them if we accept a move below
    nbasis[i]<-nbasis[i-1]
    nint[i,]<-nint[i-1,]
    dtot[i,]<-dtot[i-1,]
    degs[i,,]<-degs[i-1,,]
    vars[i,,]<-vars[i-1,,]

    if(move.type=='birth'){
      # KR:
      q0  <- sample(1:order, 1, prob=J_probs)
      wts <- make_weights(eta_vec, q0, coin_pars[[3]], coin_pars[[2]], coin_pars[[4]])
      # Delayed rejection component
      chi.cand <- 0
      delayed_reject_term <- 0
      while(sum(chi.cand) == 0 | sum(chi.cand) > order){
        chi.cand  <- stats::rbinom(p, 1, wts)
        vars.cand <- which(chi.cand == 1)
        res <- 0
        for(jj in 1:order){
          wts.j <- make_weights(eta_vec, jj, coin_pars[[3]], coin_pars[[2]], coin_pars[[4]])
          term_j <- log(J_probs[jj]) - log(Hjs) + sum(log(wts.j[vars.cand])) + sum(log(1 - wts.j[-vars.cand]))
          res <- res + exp(term_j)
        }
        delayed_reject_term <- delayed_reject_term + log(res)
      }
      #vars.cand <- which(chi.cand == 1)
      nint.cand <- length(vars.cand)
      d_probs <- seq_along(nint.cand:degree)^(-degree_penalty)
      d_probs <- d_probs / sum(d_probs)
      dtot.cand <- sample(nint.cand:degree, 1, prob=d_probs)
      degs.cand <- random_partition(dtot.cand, nint.cand)

      #==================================================
      # Modified G-Prior Stuff
      qq <- nint.cand
      dd <- dtot.cand
      g_vec.cand <- c(g_vec.curr,
                      (1 + qq*(qq + dd - 2))^(-zeta/2))
      G.cand <- 1 + tcrossprod(1/g_vec.cand) / g2[i-1]

      #==================================================

      if(nint.cand == 0) stop("Why is nint.cand zero? This shouldn't be possible.")
      basis.cand <- make_basis(vars.cand, degs.cand, X)
      B.cand <- cbind(B.curr,basis.cand) # add the new basis function to the basis functions we already have
      BtB.cand <- crossprod(B.cand)
      v.cand <- crossprod(B.cand, y)
      Sigma.cand <- tryCatch({
        safe_solve(G.cand * BtB.cand)
      }, error = function(e) {
        FALSE
      })

      # BtBi.cand <- tryCatch({
      #   safe_solve(BtB.cand)
      # }, error = function(e) {
      #   FALSE
      # })


      if(!isFALSE(Sigma.cand)){ # Otherwise reject immediately
        #Vinv.cand<-crossprod(B.cand)+diag(nbasis[i-1]+2)/tau2 # +2: one for intercept and one for birth
        #bhat.cand<-solve(Vinv.cand)%*%t(B.cand)%*%y
        #d.cand <- b_sigma + ssy - t(bhat.cand)%*%Vinv.cand%*%bhat.cand

        quad.cand <- ssy - t(v.cand) %*% Sigma.cand %*% v.cand
        Q.cand <- b_sigma + 0.5 * quad.cand
        ldet.cand <- determinant(Sigma.cand, logarithm = TRUE)$modulus

        llik.alpha <- as.numeric(
          0.5 * (ldet.cand - ldet.curr) +
            (a_sigma + n/2) * (log(Q.curr) - log(Q.cand))
        )

        # calculate the log likelihood ratio (sort of) after integrating out beta and s2
        #llik.alpha <- (
        #  0.5*log(1/tau2) +
        #    (determinant(Vinv.curr)$mod/2 - determinant(Vinv.cand)$mod/2) +
        #    (a_sigma+n/2)*(log(d.curr) - log(d.cand))
        #)

        # log prior ratio
        lprior.alpha <- (
          log(lam[i-1]) - log(nbasis[i-1] + 1) # number of basis functions
          - log(A_total)                         # Prior over vars/degs
          + log(nbasis[i-1] + 1)                 # Account for ordering (a historical artifact?)
        )

        # log proposal ratio (proposed to current via death) - (current to proposed)
        lprop.alpha <- (
          (log(move_probs[2])                    # probability of selection a death step
           + log(1/(nbasis[i-1]+1)))              # probability that this basis function is selected to kill
          - (log(move_probs[1])                  # probability of selecting a birth step
             + log(d_probs[1+dtot.cand-nint.cand])  # KR: Probability of total degree
             - lchoose(dtot.cand, nint.cand)        # KR: Probability of degree partition
             + delayed_reject_term)                 # KR: probability of nint and vars based on coin flipping (delayed reject, see above)
        )

        alpha <- llik.alpha + lprior.alpha + lprop.alpha


        if(log(stats::runif(1))<alpha){ # accept, update current values
          B.curr<-B.cand
          BtB.curr <- BtB.cand
          #BtBi.curr <- BtBi.cand
          #bhat<-bhat.cand
          #Vinv.curr<-Vinv.cand
          #d.curr<-d.cand
          v.curr <- v.cand
          G.curr <- G.cand
          g_vec.curr <- g_vec.cand
          Sigma.curr <- Sigma.cand
          Q.curr <- Q.cand
          ldet.curr <- ldet.cand


          nbasis[i]<-nbasis[i-1]+1
          nint[i,nbasis[i]]<-nint.cand
          dtot[i,nbasis[i]]<-dtot.cand
          if(!legacy){
            vars[i,nbasis[i],] <- NA
            degs[i,nbasis[i],] <- NA
          }
          vars[i,nbasis[i],1:nint.cand]<-vars.cand
          degs[i,nbasis[i],1:nint.cand]<-degs.cand

          eta_vec[vars.cand] <- eta_vec[vars.cand] + 1
          count_accept[1] <- count_accept[1] + 1
        }
      }
      # Auto reject birth
      count_propose[1] <- count_propose[1] + 1

    } else if(move.type=='death'){
      tokill<-sample(nbasis[i-1],1) # which basis function we will delete
      B.cand<-B.curr[,-(tokill+1),drop=FALSE]  # + 1 to skip the intercept
      BtB.cand <- crossprod(B.cand)
      # There really shouldn't be any issues here,
      # but for some reason we still need tryCatch
      g_vec.cand <- g_vec.curr[-(tokill+1)]
      G.cand <- 1 + tcrossprod(1/g_vec.cand) / g2[i-1]
      v.cand <- crossprod(B.cand, y)
      Sigma.cand <- tryCatch({
        safe_solve(G.cand * BtB.cand)
      }, error = function(e) {
        FALSE
      })

      # BtBi.cand <- tryCatch({
      #   safe_solve(BtB.cand)
      # }, error = function(e) {
      #   FALSE
      # })

      if(!isFALSE(Sigma.cand)){ #Otherwise reject immediately
        #Vinv.cand<-crossprod(B.cand)+diag(nbasis[i-1])/tau2
        #bhat.cand<-solve(Vinv.cand)%*%crossprod(B.cand,y)
        #d.cand <- b_sigma + ssy - crossprod(bhat.cand,Vinv.cand%*%bhat.cand)

        quad.cand <- ssy - t(v.cand) %*% Sigma.cand %*% v.cand
        Q.cand <- b_sigma + 0.5 * quad.cand
        ldet.cand <- determinant(Sigma.cand, logarithm = TRUE)$modulus

        # KR:
        vars.cand <- vars[i-1, tokill, 1:nint[i-1,tokill]]
        degs.cand <- degs[i-1, tokill, 1:nint[i-1,tokill]]
        dtot.cand <- sum(degs.cand)
        nint.cand <- length(vars.cand)
        d_probs <- seq_along(nint.cand:degree)^(-degree_penalty)
        d_probs <- d_probs / sum(d_probs)

        eta.cand <- eta_vec
        eta.cand[vars.cand] <- eta.cand[vars.cand] - 1
        proposal_nint   <- 0 # Probability of proposing the sum that we did
        proposal_nint_0 <- 0 # Probability of proposing a sum that equals 0
        proposal_nint_q <- 0 # Probability of proposing a sum that exceeds q=order
        for(jj in 1:order){
          wts.j <- make_weights(eta.cand, jj, coin_pars[[3]], coin_pars[[2]], coin_pars[[4]])
          term_j <- log(J_probs[jj]) - log(Hjs) + sum(log(wts.j[vars.cand])) + sum(log(1 - wts.j[-vars.cand]))
          proposal_nint <- proposal_nint + exp(term_j)

          # Include the delayed rejection stuff
          # P(sum(chi) = 0) = P(sum(chi) = 0 | q0=1)P(q0=1) + ... + P(sum(chi) = 0 | q0=q)P(q0=q)
          term_0_j <- log(J_probs[jj]) - log(Hjs) + sum(log(1 - wts.j))
          proposal_nint_0 <- proposal_nint_0 + exp(term_0_j)

          # if order == p, then it is not possible to overshoot
          term_q_j <- -Inf
          if(order < p){
            # Use a normal approx for this part
            mu_chi   <- sum(wts.j)
            sig_chi  <- sqrt(sum(wts.j*(1-wts.j)))
            term_q_j <- log(J_probs[jj]) - log(Hjs) +
              stats::pnorm(order+1/2,mu_chi,sig_chi,lower.tail=FALSE,log.p=TRUE)
          }
          proposal_nint_q <- proposal_nint_q + exp(term_q_j)
        }
        proposal_reject <- min(1, proposal_nint_0 + proposal_nint_q) # The 1 should never trigger, but I added it to be safe since we're using a normal approximation


        # calculate the log likelihood ratio (sort of) after integrating out beta and s2
        #llik.alpha <- (
        #  -0.5*log(1/tau2) +
        #    (determinant(Vinv.curr)$mod/2 - determinant(Vinv.cand)$mod/2) +
        #    (a_sigma+n/2)*(log(d.curr) - log(d.cand))
        #)

        llik.alpha <- as.numeric(
          0.5 * (ldet.cand - ldet.curr) +
            (a_sigma + n/2) * (log(Q.curr) - log(Q.cand))
        )

        # log prior ratio
        lprior.alpha <- (
          - log(lam[i-1]) + log(nbasis[i-1])     # number of basis functions
          + log(A_total)                         # Prior over vars/degs
          - log(nbasis[i-1] + 1)                 # Account for ordering (a historical artifact?)
        )

        # log proposal ratio (proposed to current via birth) - (current to proposed)
        lprop.alpha <- (
          (log(move_probs[1])                               # probability of selection a birth step
           + log(proposal_nint) - log(1 - proposal_reject)   # probability that we proposed the variables we did (accounting for rejection shenanigans)
           + log(d_probs[1+dtot.cand-nint.cand])             # probability that we proposed the degree that we did
           - lchoose(dtot.cand, nint.cand))                  # probability that we proposed the degree partition that we did
          - (log(move_probs[2])                             # probability of selecting a death step
             + log(1/nbasis[i-1]))                             # probability of selecting this particular basis function to kill
        )

        alpha <- llik.alpha + lprior.alpha + lprop.alpha

        if(log(stats::runif(1))<alpha){ # accept, update
          B.curr<-B.cand
          BtB.curr <- BtB.cand
          #BtBi.curr <- BtBi.cand
          #bhat<-bhat.cand
          #Vinv.curr<-Vinv.cand
          #d.curr<-d.cand
          v.curr <- v.cand
          G.curr <- G.cand
          g_vec.curr <- g_vec.cand
          Sigma.curr <- Sigma.cand
          Q.curr <- Q.cand
          ldet.curr <- ldet.cand

          nbasis[i]<-nbasis[i-1]-1

          nint[i,]<-dtot[i,]<-NA
          vars[i,,]<-degs[i,,]<-NA
          if(nbasis[i] > 0){
            nint[i,1:nbasis[i]]<-nint[i-1,(1:nbasis[i-1])[-tokill]]
            dtot[i,1:nbasis[i]]<-dtot[i-1,(1:nbasis[i-1])[-tokill]]

            vars[i,1:nbasis[i],]<-vars[i-1,(1:nbasis[i-1])[-tokill],]
            degs[i,1:nbasis[i],]<-degs[i-1,(1:nbasis[i-1])[-tokill],]
          }
          eta_vec <- eta.cand
          count_accept[2] <- count_accept[2] + 1
        }
      }
      count_propose[2] <- count_propose[2] + 1

    } else{ # KR: mutate

      # Adapt which mutation should get used.
      if(p <= 3){
        # Second mutation type isn't really needed for small p
        mutate_type <- 1
      }else{
        mutate_eps <- 0.1 # hard coded (worst case probability for either type)
        success_rates <- count_accept[3:4]/(0.01 + count_propose[3:4])
        mutate_prob <- mutate_eps + (1 - 2*mutate_eps) * stats::pnorm(-diff(pmin(5,pmax(-5,stats::qnorm(success_rates)))))
        mutate_type <- stats::rbinom(1, 1, mutate_prob)
      }

      # MUTATION TYPE 1: Re-sample the degrees
      if(mutate_type == 1){
        tochange<-sample(nbasis[i-1],1) # which basis function we will change

        # Get current info
        vars.curr <- vars[i-1,tochange,1:nint[i-1,tochange]]
        degs.curr <- degs[i-1,tochange,1:nint[i-1,tochange]]
        dtot.curr <- sum(degs.curr)    # equivalently dtot[i-1,tochange]
        nint.curr <- length(vars.curr) # equivalently nint[i-1,tochange]
        d_probs <- seq_along(nint.curr:degree)^(-degree_penalty)
        d_probs <- d_probs / sum(d_probs)

        # Generate new degree partition
        dtot.cand <- sample(nint.curr:degree, 1, prob=d_probs)
        degs.cand <- random_partition(dtot.cand, nint.curr)

        # Generate candidate basis function
        basis <- make_basis(vars.curr, degs.cand, X)
        B.cand <- B.curr
        B.cand[,tochange+1] <- basis # replace with our new basis function (+1 for intercept)
        BtB.cand <- crossprod(B.cand)

        qq <- nint.curr # This is how its defined above
        dd <- dtot.cand
        g_vec.cand <- g_vec.curr
        g_vec.cand[tochange + 1] <- (1 + qq*(qq + dd - 2))^(-zeta/2)
        G.cand <- 1 + tcrossprod(1/g_vec.cand) / g2[i-1]
        v.cand <- crossprod(B.cand, y)
        Sigma.cand <- tryCatch({
          safe_solve(G.cand * BtB.cand)
        }, error = function(e) {
          FALSE
        })

        #BtBi.cand <- tryCatch({
        #  safe_solve(BtB.cand)
        #}, error = function(e) {
        #  FALSE
        #})

        if(!isFALSE(Sigma.cand)){ #Otherwise reject immediately

          quad.cand <- ssy - t(v.cand) %*% Sigma.cand %*% v.cand
          Q.cand <- b_sigma + 0.5 * quad.cand
          ldet.cand <- determinant(Sigma.cand, logarithm = TRUE)$modulus

          # Compute acceptance ratio
          # Assume that "interaction" between the two mutation types is negligible
          llik.alpha <- as.numeric(
            0.5 * (ldet.cand - ldet.curr) +
              (a_sigma + n/2) * (log(Q.curr) - log(Q.cand))
          )

          # Variation in the degree proposal
          lprop.alpha <- log(d_probs[1+dtot.curr-nint.curr]) - log(d_probs[1+dtot.cand-nint.curr]) +
            (-lchoose(dtot.curr, nint.curr) + lchoose(dtot.cand, nint.curr))

          # Uniform priors + no dimension change <=> no term for the prior
          alpha <- llik.alpha + lprop.alpha

          if(log(stats::runif(1))<alpha){ # accept, update
            B.curr<-B.cand
            BtB.curr <- BtB.cand
            #BtBi.curr <- BtBi.cand
            #bhat<-bhat.cand
            #Vinv.curr<-Vinv.cand
            #d.curr<-d.cand
            v.curr <- v.cand
            G.curr <- G.cand
            g_vec.curr <- g_vec.cand
            Sigma.curr <- Sigma.cand
            Q.curr <- Q.cand
            ldet.curr <- ldet.cand

            vars[i,tochange,1:nint[i-1,tochange]] <- vars.curr # no change
            degs[i,tochange,1:nint[i-1,tochange]] <- degs.cand
            nint[i,tochange] <- nint.curr                      # no change
            dtot[i,tochange] <- dtot.cand

            count_accept[3] <- count_accept[3] + 1
          }
        }
        count_propose[3] <- count_propose[3] + 1

      }else{
        # MUTATION TYPE 2: Swap out a variable
        tochange<-sample(nbasis[i-1],1)           # which basis function we will change

        # Get current info
        vars.curr <- vars[i-1,tochange,1:nint[i-1,tochange]]
        degs.curr <- degs[i-1,tochange,1:nint[i-1,tochange]]
        dtot.curr <- sum(degs.curr)    # equivalently dtot[i-1,tochange]
        nint.curr <- length(vars.curr) # equivalently nint[i-1,tochange]

        # Choose a variable to delete
        ind_tochange <- sample(nint.curr, 1)
        var_tochange <- vars.curr[ind_tochange]
        newvar_probs.curr <- 1e-9 + coin_pars[[2]] + eta_vec
        newvar_probs.curr[var_tochange] <- 0
        newvar_probs.curr <- newvar_probs.curr / sum(newvar_probs.curr)
        newvar.curr <- sample(p, 1, prob=newvar_probs.curr)

        vars.cand <- vars.curr
        vars.cand[ind_tochange] <- newvar.curr

        # Construct probabilities for the reverse case
        eta.cand <- eta_vec
        eta.cand[var_tochange] <- eta.cand[var_tochange] - 1
        eta.cand[newvar.curr] <- eta.cand[newvar.curr] + 1
        newvar_probs.cand <- 1e-9 + coin_pars[[2]] + eta.cand
        newvar_probs.cand[newvar.curr] <- 0
        newvar_probs.cand <- newvar_probs.cand / sum(newvar_probs.cand)
        newvar.cand <- var_tochange

        # Generate candidate basis function
        basis <- make_basis(vars.cand, degs.curr, X)
        B.cand <- B.curr
        B.cand[,tochange+1] <- basis # replace with our new basis function (+1 for intercept)
        BtB.cand <- crossprod(B.cand)
        g_vec.cand <- g_vec.curr # No dimension change
        G.cand <- G.curr         # No dimension change
        v.cand <- crossprod(B.cand, y)
        Sigma.cand <- tryCatch({
          safe_solve(G.cand * BtB.cand)
        }, error = function(e) {
          FALSE
        })

        if(!isFALSE(Sigma.cand)){ # Otherwise reject immediately
          quad.cand <- ssy - t(v.cand) %*% Sigma.cand %*% v.cand
          Q.cand <- b_sigma + 0.5 * quad.cand
          ldet.cand <- determinant(Sigma.cand, logarithm = TRUE)$modulus

          # Compute acceptance ratio
          # Assume that "interaction" between the two mutation types is negligible
          llik.alpha <- as.numeric(
            0.5 * (ldet.cand - ldet.curr) +
              (a_sigma + n/2) * (log(Q.curr) - log(Q.cand))
          )

          # Difference in probability of the variables.
          lprop.alpha <- log(newvar_probs.cand[newvar.cand]) - log(newvar_probs.curr[newvar.curr])

          # Uniform priors + no dimension change <=> no term for the prior
          alpha <- llik.alpha + lprop.alpha

          if(log(stats::runif(1))<alpha){ # accept, update
            B.curr<-B.cand
            BtB.curr <- BtB.cand
            #BtBi.curr <- BtBi.cand
            #bhat<-bhat.cand
            #Vinv.curr<-Vinv.cand
            #d.curr<-d.cand
            v.curr <- v.cand
            G.curr <- G.cand
            g_vec.curr <- g_vec.cand
            Sigma.curr <- Sigma.cand
            Q.curr <- Q.cand
            ldet.curr <- ldet.cand

            vars[i,tochange,1:nint[i,tochange]] <- vars.cand
            degs[i,tochange,1:nint[i,tochange]] <- degs.curr # no change
            nint[i,tochange] <- nint.curr                    # no change
            dtot[i,tochange] <- dtot.curr                    # no change

            eta_vec <- eta.cand

            count_accept[4] <- count_accept[4] + 1
          }
        }
        count_propose[4] <- count_propose[4] + 1
      }
    }

    #================
    #  GIBB STEPS
    #================
    # Sample beta
    mu_n <- Sigma.curr %*% v.curr
    beta_curr <- as.vector(mvtnorm::rmvnorm(1,
                                            mu_n,
                                            s2[i-1] * Sigma.curr))
    beta[i,1:(nbasis[i]+1)] <- beta_curr

    # Residuals / variance
    yhat_curr <- B.curr %*% beta_curr
    resid <- y - yhat_curr
    sum_sq[i] <- sum(resid^2)

    # Lambda and s2
    s2[i] <- max(s2_lower,
                 1/stats::rgamma(1, n/2+a_sigma, rate = b_sigma+.5*sum_sq[i]))
    lam[i] <- stats::rgamma(1, a_M+nbasis[i], rate = b_M+1)                                               # update lambda

    # Sample g2
    if(g2_sample == "f"){
      g2[i] <- g2[i-1]
    }else if(g2_sample == "lf"){
      g2[i] <- rg0sq_laplace_full(1, a_g, b_g, g_vec.curr, BtB.curr)
    }else if(g2_sample == "lo"){
      g2[i] <- rg0sq_laplace_orth(1, a_g, b_g, g_vec.curr)
    }else if(g2_sample == "mh"){
      # Metropolis hastings using full Laplace as proposal
      lap_pars <- rg0sq_laplace_full(-1, a_g, b_g, g_vec.curr, BtB.curr)
      g2_cand <- 1/stats::rgamma(1, lap_pars$a, lap_pars$b)

      lp_cand <- log(dgsq_full(g2_cand, a_g, b_g, g_vec.curr, BtB.curr))
      lp_curr <- log(dgsq_full(g2[i-1], a_g, b_g, g_vec.curr, BtB.curr))

      lprop_cand <- dgamma(1 / g2_cand, lap_pars$a, lap_pars$b, log=TRUE) - 2 * log(g2_cand)
      lprop_curr <- dgamma(1 / g2[i-1], lap_pars$a, lap_pars$b, log=TRUE) - 2 * log(g2[i-1])

      log_alpha <- (lp_cand - lp_curr) + (lprop_curr - lprop_cand)
      if(log(runif(1)) < log_alpha){
        g2[i] <- g2_cand
      }else{
        g2[i] <- g2[i-1]
      }
    }else if(g2_sample == "mho"){
      # Metropolis hastings using full Laplace as proposal
      lap_pars <- rg0sq_laplace_orth(-1, a_g, b_g, g_vec.curr)
      g2_cand <- 1/stats::rgamma(1, lap_pars$a, lap_pars$b)

      lp_cand <- log(dgsq_full(g2_cand, a_g, b_g, g_vec.curr, BtB.curr))
      lp_curr <- log(dgsq_full(g2[i-1], a_g, b_g, g_vec.curr, BtB.curr))

      lprop_cand <- dgamma(1 / g2_cand, lap_pars$a, lap_pars$b, log=TRUE) - 2 * log(g2_cand)
      lprop_curr <- dgamma(1 / g2[i-1], lap_pars$a, lap_pars$b, log=TRUE) - 2 * log(g2[i-1])

      log_alpha <- (lp_cand - lp_curr) + (lprop_curr - lprop_cand)
      if(log(runif(1)) < log_alpha){
        g2[i] <- g2_cand
      }else{
        g2[i] <- g2[i-1]
      }
    }else if(g2_sample == "mhoo"){
      # Metropolis hastings using full Laplace as proposal
      lap_pars <- rg0sq_laplace_orth(-1, a_g, b_g, g_vec.curr)
      g2_cand <- 1/stats::rgamma(1, lap_pars$a, lap_pars$b)

      lp_cand <- log(dgsq_orth(g2_cand, a_g, b_g, g_vec.curr))
      lp_curr <- log(dgsq_orth(g2[i-1], a_g, b_g, g_vec.curr))

      lprop_cand <- dgamma(1 / g2_cand, lap_pars$a, lap_pars$b, log=TRUE) - 2 * log(g2_cand)
      lprop_curr <- dgamma(1 / g2[i-1], lap_pars$a, lap_pars$b, log=TRUE) - 2 * log(g2[i-1])

      log_alpha <- (lp_cand - lp_curr) + (lprop_curr - lprop_cand)
      if(log(runif(1)) < log_alpha){
        g2[i] <- g2_cand
      }else{
        g2[i] <- g2[i-1]
      }
    }else{
      warning("g2_sample not recognized. Fixing g2_sample.")
      g2[i] <- g2[i-1]
    }

    if(verbose & i%%1000 == 0){
      pr<-c('MCMC iteration',i,myTimestamp(),'nbasis:', nbasis[i])
      cat(pr,'\n')
    }
  }

  # Trim down data structures
  mcmc_iter <- seq(nburn+1, nmcmc, by=thin)
  max_basis_used <- max(nbasis, na.rm = TRUE)
  basis_high <- if(max_basis_used > 0) 1:max_basis_used else integer(0)
  max_inter_used <- max(nint, na.rm = TRUE)
  inter_high <- if(is.finite(max_inter_used) && max_inter_used > 0) 1:max_inter_used else integer(0)


  vars <- vars[mcmc_iter, basis_high, inter_high, drop=FALSE]
  degs <- degs[mcmc_iter, basis_high, inter_high, drop=FALSE]
  nint <- nint[mcmc_iter, basis_high, drop=FALSE]
  dtot <- dtot[mcmc_iter, basis_high, drop=FALSE]

  beta <- beta[mcmc_iter, 1:(1+max(nbasis))]
  nbasis <- nbasis[mcmc_iter]
  s2  <- s2[mcmc_iter]
  lam <- lam[mcmc_iter]
  g2  <- g2[mcmc_iter]
  sum_sq <- sum_sq[mcmc_iter]


  out <- list(B=B.curr,
              vars=vars,degs=degs,
              nint=nint,dtot=dtot,
              nbasis=nbasis,beta=beta,
              s2=s2,lam=lam,g2=g2,sum_sq=sum_sq,
              eta=eta_vec,
              count_accept=count_accept,
              count_propose=count_propose,
              prior_type="gprior",
              X=X, y=y)
  class(out) <- "adaptive_khaos"
  return(out)
}

#' @export
#' @param ... additional arguments passed to adaptive_khaos
#' @rdname adaptive_khaos_gprior
adaptive_khaos2 <- function(...) {
  .Deprecated("adaptive_khaos_gprior")
  adaptive_khaos_gprior(...)
}

test_that("simple emulation test (adaptive)", {
  cat('simple emulation test')
  f <- function(x) 10.391*((x[1]-0.4)*(x[2]-0.6) + 0.36)
  n <- 200
  p <- 3
  X <- matrix(runif(n*p), ncol=p)
  y <- apply(X, 1, f) + rnorm(n, 0, 0.1)
  fit <- adaptive_khaos(X, y)
  yhat <- colMeans(predict(fit))
  d1 <- sqrt(mean((y-yhat)^2))/sd(y)
  expect_that(d1, is_less_than(0.2))
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adaptive_khaos2.R
\name{adaptive_khaos}
\alias{adaptive_khaos}
\title{Bayesian Adaptive Polynomial Chaos Expansion}
\usage{
adaptive_khaos(
  X,
  y,
  prior_type = "ridge",
  nmcmc = 10000,
  nburn = 9000,
  thin = 1,
  legacy = TRUE,
  verbose = TRUE,
  ...
)
}
\arguments{
\item{X}{A data frame or matrix of predictors scaled to lie between 0 and 1.}

\item{y}{A numeric response vector of length \code{nrow(X)}.}

\item{prior_type}{Character string specifying the coefficient-prior formulation.
Currently one of \code{"ridge"} or \code{"gprior"}.}

\item{nmcmc}{Number of MCMC iterations.}

\item{nburn}{Number of initial MCMC iterations to discard.}

\item{thin}{Keep every \code{thin}-th retained MCMC sample.}

\item{legacy}{Logical. If \code{TRUE}, reproduces the historical storage behavior
of earlier implementations, which may retain stale values in unused portions of
the internal \code{vars} and \code{degs} arrays. If \code{FALSE}, these unused
entries are cleared more carefully.}

\item{verbose}{Logical; if \code{TRUE}, progress messages are printed.}

\item{...}{Additional arguments passed to the prior-specific fitting routine.}
}
\value{
An object of class \code{"adaptive_khaos"} containing the fitted model.
The object is a list including sampled basis functions, coefficient samples,
variance parameters, and additional MCMC output. The specific structure depends
on the chosen \code{prior_type}.
}
\description{
A wrapper for adaptive Bayesian polynomial chaos expansion models with
different coefficient prior formulations.
}
\details{
Depending on \code{prior_type}, this function calls either
\code{adaptive_khaos_ridge()} or \code{adaptive_khaos_gprior()}.

This function dispatches to one of the prior-specific adaptive KHAOS fitting
routines:
\describe{
\item{\code{prior_type = "ridge"}}{Calls \code{\link{adaptive_khaos_ridge}}.}
\item{\code{prior_type = "gprior"}}{Calls \code{\link{adaptive_khaos_gprior}}.}
}

The wrapper exposes only the most commonly used arguments. Additional
prior-specific tuning parameters can be supplied through \code{...}; see
\code{\link{adaptive_khaos_ridge}} and \code{\link{adaptive_khaos_gprior}}
for full documentation.
}
\examples{
\donttest{
X <- lhs::maximinLHS(100, 2)
f <- function(x) 10.391 * ((x[1] - 0.4) * (x[2] - 0.6) + 0.36)
y <- apply(X, 1, f) + stats::rnorm(100, 0, 0.1)

fit1 <- adaptive_khaos(X, y)
fit2 <- adaptive_khaos(X, y, prior_type = "gprior")
}
}
\references{
Francom, Devin, and Bruno Sanso. "BASS: An R package for fitting and performing
sensitivity analysis of Bayesian adaptive spline surfaces."
\emph{Journal of Statistical Software} 94.LA-UR-20-23587 (2020).

Rumsey, K.N., Francom, D., Gibson, G., Tucker, J. D., and Huerta, G. (2026).
Bayesian Adaptive Polynomial Chaos Expansions. \emph{Stat}, 15(1), e70151.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse_chaos.R
\name{print.sparse_khaos}
\alias{print.sparse_khaos}
\title{Print Method for class sparse_khaos}
\usage{
\method{print}{sparse_khaos}(x, ...)
}
\arguments{
\item{x}{An object returned by the \code{sparse_khaos()} function.}

\item{...}{additional arguments passed to or from other methods.}
}
\value{
The input object \code{x}, invisibly.
}
\description{
See \code{sparse_khaos()} for details.
}
\details{
Print function for sparse_khaos object.
}
\examples{
X <- lhs::maximinLHS(100, 2)
f <- function(x) 10.391*((x[1]-0.4)*(x[2]-0.6) + 0.36)
y <- apply(X, 1, f) + stats::rnorm(100, 0, 0.1)
fit <- sparse_khaos(X, y)
print(fit)
}
\references{
Shao, Q., Younes, A., Fahs, M., & Mara, T. A. (2017). Bayesian sparse polynomial chaos expansion for global sensitivity analysis. Computer Methods in Applied Mechanics and Engineering, 318, 474-496.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adaptive_chaos.R
\name{predict.adaptive_khaos}
\alias{predict.adaptive_khaos}
\title{Predict Method for class adaptive_khaos}
\usage{
\method{predict}{adaptive_khaos}(
  object,
  newdata = NULL,
  mcmc.use = NULL,
  nugget = FALSE,
  nreps = 1,
  ...
)
}
\arguments{
\item{object}{An object returned by the \code{adaptive_khaos()} function.}

\item{newdata}{A dataframe of the same dimension as the training data.}

\item{mcmc.use}{Which posterior samples should be used for prediction?}

\item{nugget}{Logical. Should predictions include error?}

\item{nreps}{How many predictions should be taken for each mcmc sample (ignored when \code{nugget = FALSE}).}

\item{...}{Additional arguments to predict}
}
\value{
A numeric matrix of posterior predictive samples, with rows corresponding
to samples and columns corresponding to observations.
}
\description{
See \code{adaptive_khaos()} for details.
}
\details{
Predict function for adaptive_khaos object.
}
\examples{
\donttest{
X <- lhs::maximinLHS(100, 2)
f <- function(x) 10.391*((x[1]-0.4)*(x[2]-0.6) + 0.36)
y <- apply(X, 1, f) + stats::rnorm(100, 0, 0.1)
fit <- adaptive_khaos(X, y)
predict(fit)
}
}
\references{
Francom, Devin, and Bruno Sanso. "BASS: An R package for fitting and performing sensitivity analysis of Bayesian adaptive spline surfaces." Journal of Statistical Software 94.LA-UR-20-23587 (2020).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sobol_khaos.R
\name{sobol_khaos}
\alias{sobol_khaos}
\title{Compute Sobol Sensitivity Indices for khaos Models}
\usage{
sobol_khaos(obj, plot_it = TRUE, samples = 1000)
}
\arguments{
\item{obj}{An object of class \code{"adaptive_khaos"} returned by \code{adaptive_khaos()} or of class \code{"sparse_khaos"} returned by \code{sparse_khaos()}.}

\item{plot_it}{Logical; if \code{TRUE}, boxplots of the sensitivity indices are shown.}

\item{samples}{How many samples from the posterior (only used for sparse class).}
}
\value{
A list with the following components:
\describe{
\item{S}{A data frame of Sobol indices for all main effects and interactions (MCMC samples).}
\item{T}{A data frame of total-effect Sobol indices (MCMC samples).}
\item{leftover}{A numeric vector of unexplained variance proportions (residuals).}
}
}
\description{
Computes first-order, total-effect, and full interaction Sobol indices
from an adaptive KHAOS model fit. Returns MCMC samples of sensitivity
indices, including all active interaction terms used in the model.
}
\examples{
\donttest{
X <- lhs::maximinLHS(100, 2)
f <- function(x) 10.391*((x[1]-0.4)*(x[2]-0.6) + 0.36)
y <- apply(X, 1, f) + stats::rnorm(100, 0, 0.1)
fit <- adaptive_khaos(X, y)
sob <- sobol_khaos(fit)
head(sob$S)
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse_chaos.R
\name{sparse_khaos}
\alias{sparse_khaos}
\title{Bayesian Sparse Polynomial Chaos Expansion}
\usage{
sparse_khaos(
  X,
  y,
  degree = c(2, 2, 16),
  order = c(1, 1, 5),
  prior = c(1, 1, 1),
  lambda = 1,
  rho = 0,
  regularize = TRUE,
  max_basis = 1e+06,
  enrichment = "adaptive",
  adaptive_ladder = c(10000, 1e+05),
  verbose = TRUE
)
}
\arguments{
\item{X}{A data frame or matrix of predictors scaled to lie between 0 and 1.}

\item{y}{A numeric response vector of length \code{nrow(X)}.}

\item{degree}{Integer vector of length 3 giving the polynomial-degree schedule:
\code{c(d_init, d_increment, d_max)}.}

\item{order}{Integer vector of length 3 giving the interaction-order schedule:
\code{c(q_init, q_increment, q_max)}.}

\item{prior}{Numeric vector of prior hyperparameters
\code{c(n0, v0, s0)}, where \code{n0} controls shrinkage of the regression
coefficients, \code{v0} is the degrees-of-freedom parameter for the inverse-gamma
prior on \eqn{\sigma^2}, and \code{s0} is the scale parameter for that prior.}

\item{lambda}{Nonnegative penalty parameter used in the KIC calculation. Larger
values favor sparser models.}

\item{rho}{Basis functions are discarded after partial-correlation screening if
their squared partial correlation is smaller than \code{rho}.}

\item{regularize}{Logical; if \code{TRUE}, a weighted LASSO pre-screen is used
to reduce the candidate basis set before partial-correlation ranking.}

\item{max_basis}{Maximum number of candidate basis functions allowed in any
fitting round. This acts as a hard cap on computational cost.}

\item{enrichment}{Either an integer in \code{0:4} specifying a fixed enrichment
strategy, or the character string \code{"adaptive"}, which allows the strategy
to become more restrictive when the next candidate set would otherwise be too
large. See Details.}

\item{adaptive_ladder}{Numeric vector of candidate-size thresholds used only when
\code{enrichment = "adaptive"}. The ladder should have length between 1 and 4,
be sorted in increasing order, and satisfy \code{max(adaptive_ladder) <= max_basis}.
Internally, the ladder is used to determine when the algorithm should fall back
to a more restrictive enrichment strategy before attempting the next recursion
step.}

\item{verbose}{Logical; if \code{TRUE}, progress messages are printed.}
}
\value{
An object of class \code{"sparse_khaos"}.
The object is a list containing:
\describe{
\item{B}{Matrix of selected basis functions evaluated at the training inputs.}
\item{nbasis}{Number of selected basis functions (including intercept).}
\item{vars}{Matrix encoding the polynomial multi-indices for each basis function.}
\item{beta_hat}{MAP estimates of the regression coefficients.}
\item{s2}{Estimated noise variance.}
\item{X, y}{Training data (with \code{y} on the original scale).}
\item{mu_y, sigma_y}{Centering and scaling constants used for \code{y}.}
\item{BtB, BtBi}{Cross-product matrix and its inverse.}
\item{G}{Prior scaling factors for coefficients.}
\item{prior}{Prior hyperparameters.}
\item{KIC}{Kashyap information criterion values for candidate models.}
}
}
\description{
A Bayesian sparse polynomial chaos expansion (PCE) method based on the
forward-selection framework of Shao et al. (2017), with optional enrichment
strategies that adapt the candidate basis set across degree and interaction-order
expansions.
}
\details{
For fixed maximum degree and interaction order, the method constructs a candidate
library of polynomial basis functions, ranks them using marginal correlations and
sequential partial correlations, and evaluates nested models using the Kashyap
information criterion (KIC). The best-ranked model is retained at that stage.

If the selected model contains a basis function at the current maximum degree or
interaction order, the algorithm attempts another fitting round with increased
degree and/or interaction order. In later rounds, the candidate basis set can be
rebuilt using one of several enrichment strategies rather than constructing the
full basis library from scratch.

The \code{enrichment} argument controls how the candidate set is generated after
the first fitting round:
\describe{
\item{\code{0} (local enrichment)}{
Starts from the previously selected basis functions and applies local
modifications to each term. These include deletion, simplification,
complication, and promotion moves. This is typically the fastest strategy,
but it is also the most local.}
\item{\code{1} (active-variable rebuild)}{
Rebuilds the full candidate library using only variables that were active
in the previously selected model. This is computationally efficient, but
inactive variables cannot re-enter once dropped.}
\item{\code{2} (full active rebuild plus sparse inactive enrichment)}{
Rebuilds the candidate library on the currently active variables, then
enriches around the previously selected basis functions by allowing inactive
variables to re-enter through complication and promotion moves. This is a
compromise between speed and flexibility.}
\item{\code{3} (full active rebuild plus inactive enrichment)}{
Rebuilds the candidate library on the active variables, then enriches that
larger active-variable library by allowing inactive variables to re-enter
through complication and promotion moves. This is more expansive than
\code{enrichment = 2}, but can be substantially more expensive.}
\item{\code{4} (full rebuild)}{
Rebuilds the full candidate library over all variables at each enrichment
step. This is the most exhaustive strategy.}
\item{\code{"adaptive"}}{
Begins with a relatively expansive enrichment strategy and, before the next
recursion step, uses an upper bound on the candidate-set size to decide
whether a more restrictive strategy should be used instead. This recovers
the computational benefit of avoiding construction of candidate sets that
are known in advance to be too large.}
}

When \code{enrichment = "adaptive"}, the function uses \code{adaptive_ladder}
together with an upper bound on the next candidate-set size. If the current
enrichment choice appears too expensive, the algorithm steps down to a more
restrictive enrichment rule before proceeding. If even the most restrictive
admissible next step is predicted to exceed the allowed budget, the current
fitted model is returned.

Strategies \code{2}, \code{3}, and \code{"adaptive"} are designed to address a
limitation of active-variable-only rebuilds: variables that are inactive in one
round are allowed to re-enter the model in later rounds.
}
\examples{
X <- lhs::maximinLHS(100, 2)
f <- function(x) 10.391 * ((x[1] - 0.4) * (x[2] - 0.6) + 0.36)
y <- apply(X, 1, f) + stats::rnorm(100, 0, 0.1)

fit <- sparse_khaos(X, y)

}
\references{
Shao, Q., Younes, A., Fahs, M., and Mara, T. A. (2017).
Bayesian sparse polynomial chaos expansion for global sensitivity analysis.
\emph{Computer Methods in Applied Mechanics and Engineering}, \strong{318},
474--496.

Rumsey, K. N., Francom, D., Gibson, G., Tucker, J. D., and Huerta, G. (2026).
Bayesian adaptive polynomial chaos expansions.
\emph{Stat}, \strong{15}(1), e70151.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sparse_chaos.R
\name{predict.sparse_khaos}
\alias{predict.sparse_khaos}
\title{Predict Method for class sparse_khaos}
\usage{
\method{predict}{sparse_khaos}(object, newdata = NULL, samples = 1000, nugget = FALSE, ...)
}
\arguments{
\item{object}{An object returned by the \code{sparse_khaos()} function.}

\item{newdata}{A dataframe of the same dimension as the training data.}

\item{samples}{How many posterior samples should be taken at each test point? If 0 or FALSE, then the MAP estimate is returned.}

\item{nugget}{logical; should predictions include error? If FALSE, predictions will be for mean.}

\item{...}{Additional arguments to predict}
}
\value{
A numeric matrix of predictive samples (or a vector if \code{samples = FALSE}),
with rows corresponding to samples and columns corresponding to observations.
}
\description{
See \code{sparse_khaos()} for details.
}
\details{
Predict function for sparse_khaos object.
}
\examples{
X <- lhs::maximinLHS(100, 2)
f <- function(x) 10.391*((x[1]-0.4)*(x[2]-0.6) + 0.36)
y <- apply(X, 1, f) + stats::rnorm(100, 0, 0.1)
fit <- sparse_khaos(X, y)
predict(fit)

}
\references{
Shao, Q., Younes, A., Fahs, M., & Mara, T. A. (2017). Bayesian sparse polynomial chaos expansion for global sensitivity analysis. Computer Methods in Applied Mechanics and Engineering, 318, 474-496.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ordinal_khaos.R
\name{predict.ordinal_khaos}
\alias{predict.ordinal_khaos}
\title{Predict Method for class \code{ordinal_khaos}}
\usage{
\method{predict}{ordinal_khaos}(
  object,
  newdata = NULL,
  mcmc.use = NULL,
  type = c("class", "prob", "latent"),
  aggregate = FALSE,
  ...
)
}
\arguments{
\item{object}{An object of class \code{ordinal_khaos} returned by \code{ordinal_khaos()}.}

\item{newdata}{A data frame or matrix of predictors on the same scale and with
the same columns as used for fitting. If \code{NULL}, uses \code{object$X}.}

\item{mcmc.use}{Integer vector of posterior draw indices to use. If \code{NULL},
uses all stored draws (i.e., \code{seq_along(object$nbasis)}).}

\item{type}{Character; one of \code{"prob"}, \code{"class"}, or \code{"latent"}.
\code{"prob"} returns category probabilities \eqn{P(Y=k \mid x)} for
\eqn{k=1,\dots,K}. \code{"class"} returns MAP classes. \code{"latent"} returns
the latent mean \eqn{f(x)} (i.e., \eqn{B(x)\beta}) for each draw.}

\item{aggregate}{Logical (default \code{FALSE}); if \code{TRUE}, averages results across the
selected MCMC draws (probabilities are averaged; classes are aggregated by
simple voting; latent means are averaged). If \code{FALSE}, returns per-draw
arrays.}

\item{...}{Unused; included for method compatibility.}
}
\value{
If \code{type = "prob"} and \code{aggregate = TRUE}, an \code{nrow(newdata) x K}
numeric matrix of posterior mean category probabilities. If \code{aggregate = FALSE},
a 3-D array of dimension \code{(#draws) x nrow(newdata) x K}.

If \code{type = "class"} and \code{aggregate = TRUE}, an integer vector of length
\code{nrow(newdata)} with MAP classes aggregated by voting across draws. If
\code{aggregate = FALSE}, a \code{(#draws) x nrow(newdata)} integer matrix of
per-draw MAP classes.

If \code{type = "latent"}, returns the latent mean(s) \eqn{f(x)}: a numeric vector
of length \code{nrow(newdata)} if \code{aggregate = TRUE}, or a
\code{(#draws) x nrow(newdata)} numeric matrix otherwise.

A numeric matrix of posterior predictive samples, with rows corresponding
to samples and columns corresponding to observations.
}
\description{
Compute posterior predictive quantities for an ordinal probit BA-PCE model.
For each selected MCMC draw, the method builds the polynomial chaos basis,
forms the linear predictor \eqn{f(x) = B(x)\beta}, reconstructs the ordered
cutpoints, and returns either category probabilities, MAP classes, or latent
means. Results can be averaged across draws or returned per-draw.
}
\details{
Let \eqn{Z = f_\beta(X) + \varepsilon} with \eqn{\varepsilon \sim \mathcal{N}(0,1)} and
ordered cutpoints \eqn{\gamma_0=-\infty < \gamma_1 < \cdots < \gamma_{K-1} < \gamma_K=\infty}.
For a new input \eqn{x}, the category probabilities are
\deqn{P(Y=k\mid x) = \Phi(\gamma_k - f(x)) - \Phi(\gamma_{k-1} - f(x)).}
The function reconstructs the full cutpoint vector internally. If the fitted
object stored \eqn{\gamma_1,\dots,\gamma_{K-1}} then these are used directly;
if it stored only \eqn{\gamma_2,\dots,\gamma_{K-1}} (with \eqn{\gamma_1 \equiv 0})
the missing fixed cutpoint is inserted automatically.
}
\examples{
\donttest{
# Toy example (X scaled to [0,1])
set.seed(1)
X <- lhs::maximinLHS(200, 3)
# Suppose y has K=4 ordered categories:
fit <- ordinal_khaos(X, sample(1:4, 200, replace=TRUE),
                     nmcmc = 2000, nburn = 1000, thin = 2, verbose = FALSE)

# Posterior mean probabilities for each class
p_hat <- predict(fit, newdata = X, type = "prob", aggregate = TRUE)

# MAP classes (aggregated across draws)
y_hat <- predict(fit, newdata = X, type = "class")

# Latent mean f(x)
f_hat <- predict(fit, newdata = X, type = "latent")
}

}
\references{
Hoff, Peter D. (2009). \emph{A First Course in Bayesian Statistical Methods}.
Springer. (See ordinal probit data augmentation.)

Francom, Devin, and Bruno Sanso (2020). "BASS: An R package for fitting and
performing sensitivity analysis of Bayesian adaptive spline surfaces."
\emph{Journal of Statistical Software} 94. LA-UR-20-23587.
}
\seealso{
\code{\link{ordinal_khaos}} for model fitting.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ordinal_khaos.R
\name{plot.ordinal_khaos}
\alias{plot.ordinal_khaos}
\title{Plot Method for class \code{ordinal_khaos}}
\usage{
\method{plot}{ordinal_khaos}(x, ...)
}
\arguments{
\item{x}{An object of class \code{ordinal_khaos}.}

\item{...}{additional arguments passed to \code{plot}}
}
\value{
Invisibly returns \code{NULL}.
}
\description{
Produces a 2x2 diagnostic plot: (1) trace of the number of basis functions;
(2) jittered predicted vs. actual classes with point size indicating posterior
support; (3) posterior mean class probabilities on \code{x$X} (lines if 1D,
otherwise barplot of averages); (4) histogram of latent means \eqn{f(x)} with
posterior-average thresholds overlaid.
}
\details{
Panel (2) uses \code{predict(x, type="class", aggregate=FALSE)} to compute
per-draw MAP classes, plotting the modal class vs. observed class; point size
reflects the fraction of draws supporting the modal class. Panel (3) uses
\code{predict(x, type="prob")} to plot posterior mean class probabilities.
Panel (4) uses \code{predict(x, type="latent", aggregate=FALSE)} to pool
latent means across draws and overlays posterior-average cutpoints (handles
either stored \eqn{\gamma_1,\dots,\gamma_{K-1}} or \eqn{\gamma_2,\dots,\gamma_{K-1}}
with \eqn{\gamma_1\equiv 0}).
}
\examples{
\donttest{
set.seed(1)
X <- matrix(runif(200), ncol = 1)
y <- sample(1:4, 200, replace = TRUE)
fit <- ordinal_khaos(X, y, nmcmc = 2000, nburn = 1000, thin = 2, verbose = FALSE)
plot(fit)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adaptive_chaos.R
\name{adaptive_khaos_ridge}
\alias{adaptive_khaos_ridge}
\title{Bayesian Adaptive Polynomial Chaos Expansion (Ridge Prior)}
\usage{
adaptive_khaos_ridge(
  X,
  y,
  degree = 15,
  order = 5,
  nmcmc = 10000,
  nburn = 9000,
  thin = 1,
  max_basis = 1000,
  tau2 = 10^5,
  s2_lower = 0,
  g1 = 0,
  g2 = 0,
  h1 = 4,
  h2 = 40/length(y),
  move_probs = rep(1/3, 3),
  coin_pars = list(function(j) 1/j, 1, 2, 3),
  degree_penalty = 0,
  legacy = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{X}{A data frame or matrix of predictors scaled to be between 0 and 1}

\item{y}{a response vector}

\item{degree}{Maximum polynomial degree for each basis function.}

\item{order}{Maximum order of interaction for each basis function.}

\item{nmcmc}{Number of MCMC iterations.}

\item{nburn}{Number of initial MCMC iterations to discard.}

\item{thin}{Keep every \code{thin} samples.}

\item{max_basis}{Maximum number of basis functions.}

\item{tau2}{Prior variance for coefficients}

\item{s2_lower}{Lower bound on process variance (numerically useful for deterministic functions).}

\item{g1, g2}{Shape/rate parameters for the IG prior on process variance (default is Jefffrey's prior)}

\item{h1, h2}{Shape/rate parameters for the Gamma prior on expected number of basis functions.}

\item{move_probs}{A 3-vector with probabilities for (i) birth, (ii) death, and (iii) mutation.}

\item{coin_pars}{A list of control parameters for coinflip proposal}

\item{degree_penalty}{Increasing this value encourages lower order polynomial terms (0 is no penalization).}

\item{legacy}{Logical. If TRUE, mimics original implementation behavior (from the emulator comparison paper),
which may retain invalid basis function structures across iterations. Defaults to FALSE.}

\item{verbose}{Logical. Should progress information be printed?}
}
\value{
An object of class \code{"adaptive_khaos"}.
The object is a list containing:
\describe{
\item{B}{Matrix of basis functions evaluated at the training inputs.}
\item{vars, degs}{Arrays encoding the variables and polynomial degrees for each basis function across MCMC iterations.}
\item{nint, dtot}{Interaction order and total degree for each basis function.}
\item{nbasis}{Number of basis functions per MCMC iteration.}
\item{beta}{Posterior samples of regression coefficients.}
\item{s2}{Posterior samples of the noise variance.}
\item{lam}{Posterior samples of the Poisson rate parameter controlling model size.}
\item{sum_sq}{Residual sum of squares at each iteration.}
\item{eta}{Variable inclusion counts used in the proposal distribution.}
\item{count_accept, count_propose}{Acceptance and proposal counts for RJMCMC moves.}
\item{prior_type}{Character string indicating the ridge prior.}
\item{X, y}{Training data.}
}
}
\description{
Adaptive Bayesian polynomial chaos expansion using a fixed Gaussian (ridge)
prior on coefficients. This implementation corresponds to a simplified
version of the adaptive Bayesian PCE methodology described in Rumsey et al. (2026)
}
\details{
Implements the RJMCMC algorithm described by Francom & Sanso (2020) for BMARS, modifying it for polynomial chaos basis functions.
As an alternative the NKD procedure of Nott et al. (2005), we use a coinflipping procedure to identify useful variables. See writeup (coming soon) for details.
The coin_pars argument is an ordered list with elements:
\enumerate{
\item{a function giving proposal probability for q_0, the expected order of interaction,}
\item{epsilon, the base weight for each variable (epsilon=Inf corresponds to uniform sampling),}
\item{alpha, the exponent-learning rate,}
\item{num_passes a numerical parameter only.}
}

Other proposal notes:
Expected order q0 is chosen from 1:order with weights \code{coin_pars[[1]](1:order)}
Degree is chosen from q0:degree with weights \code{(1:(degree-q0+1))^(-degree_penalty)} and a random partition is created (see helper function).

Two types of change steps are possible (and equally likely)
A re-ordering of the degrees
A single variable is swapped out with another one
}
\examples{
\donttest{
X <- lhs::maximinLHS(100, 2)
f <- function(x) 10.391*((x[1]-0.4)*(x[2]-0.6) + 0.36)
y <- apply(X, 1, f) + stats::rnorm(100, 0, 0.1)
fit <- adaptive_khaos(X, y)
}
}
\references{
Francom, Devin, and Bruno Sanso. "BASS: An R package for fitting and performing sensitivity analysis of Bayesian adaptive spline surfaces." Journal of Statistical Software 94.LA-UR-20-23587 (2020).

Nott, David J., Anthony YC Kuk, and Hiep Duc. "Efficient sampling schemes for Bayesian MARS models with many predictors." Statistics and Computing 15 (2005): 93-101.

Rumsey, K.N., Francom, D., Gibson, G., Derek Tucker, J. and Huerta, G., 2026. Bayesian Adaptive Polynomial Chaos Expansions. Stat, 15(1), p.e70151.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adaptive_khaos2.R
\name{adaptive_khaos_gprior}
\alias{adaptive_khaos_gprior}
\alias{adaptive_khaos2}
\title{Bayesian Adaptive Polynomial Chaos Expansion (Modified g-Prior)}
\usage{
adaptive_khaos_gprior(
  X,
  y,
  degree = 15,
  order = 5,
  nmcmc = 10000,
  nburn = 9000,
  thin = 1,
  max_basis = 1000,
  a_g = 0.001,
  b_g = 1000,
  zeta = 1,
  g2_sample = "mh",
  g2_init = NULL,
  s2_lower = 0,
  a_sigma = 0,
  b_sigma = 0,
  a_M = 4,
  b_M = 4/length(y),
  move_probs = rep(1/3, 3),
  coin_pars = list(function(j) 1/j, 1, 2, 3),
  degree_penalty = 0,
  legacy = TRUE,
  verbose = TRUE
)

adaptive_khaos2(...)
}
\arguments{
\item{X}{A data frame or matrix of predictors scaled to be between 0 and 1}

\item{y}{a response vector}

\item{degree}{Maximum polynomial degree for each basis function.}

\item{order}{Maximum order of interaction for each basis function.}

\item{nmcmc}{Number of MCMC iterations.}

\item{nburn}{Number of initial MCMC iterations to discard.}

\item{thin}{Keep every \code{thin} samples.}

\item{max_basis}{Maximum number of basis functions.}

\item{a_g, b_g}{Prior parameters for global g-prior term (on precision scale)}

\item{zeta}{Hyperparameter for modified g-prior. Setting zeta = 0 reduces to the usual g-prior.}

\item{g2_sample}{Character string specifying the method used to sample or update the prior scaling parameter \code{g0^2}.}

\item{g2_init}{Initial calue of g2 (global precision for g-prior). Becomes the only value when \code{g2_sample = "fixed"}.}

\item{s2_lower}{Lower bound on process variance (numerically useful for deterministic functions).}

\item{a_sigma, b_sigma}{Shape/rate parameters for the IG prior on process variance (default is Jefffrey's prior)}

\item{a_M, b_M}{Shape/rate parameters for the Gamma prior on expected number of basis functions.}

\item{move_probs}{A 3-vector with probabilities for (i) birth, (ii) death, and (iii) mutation.}

\item{coin_pars}{A list of control parameters for coinflip proposal}

\item{degree_penalty}{Increasing this value encourages lower order polynomial terms (0 is no penalization).}

\item{legacy}{Logical. If TRUE, mimics original implementation behavior (from the emulator comparison paper),
which may retain invalid basis function structures across iterations. Defaults to FALSE.}

\item{verbose}{Logical. Should progress information be printed?}

\item{...}{additional arguments passed to adaptive_khaos}
}
\value{
An object of class \code{"adaptive_khaos"}.
The object is a list containing:
\describe{
\item{B}{Matrix of basis functions evaluated at the training inputs.}
\item{vars, degs}{Arrays encoding the variables and polynomial degrees for each basis function across MCMC iterations.}
\item{nint, dtot}{Interaction order and total degree for each basis function.}
\item{nbasis}{Number of basis functions per MCMC iteration.}
\item{beta}{Posterior samples of regression coefficients.}
\item{s2}{Posterior samples of the noise variance.}
\item{lam}{Posterior samples of the Poisson rate parameter controlling model size.}
\item{g2}{Posterior samples of the global g-prior scaling parameter.}
\item{sum_sq}{Residual sum of squares at each iteration.}
\item{eta}{Variable inclusion counts used in the proposal distribution.}
\item{count_accept, count_propose}{Acceptance and proposal counts for RJMCMC moves.}
\item{prior_type}{Character string indicating the g-prior formulation.}
\item{X, y}{Training data.}
}
}
\description{
Adaptive Bayesian polynomial chaos expansion using a modified g-prior. This
implementation corresponds to the adaptive Bayesian PCE methodology described
in Rumsey et al. (2026)
}
\details{
Implements the RJMCMC algorithm described by Francom & Sanso (2020) for BMARS, modifying it for polynomial chaos basis functions.
As an alternative the NKD procedure of Nott et al. (2005), we use a coinflipping procedure to identify useful variables. See writeup (coming soon) for details.
The coin_pars argument is an ordered list with elements:
\enumerate{
\item{a function giving proposal probability for q_0, the expected order of interaction,}
\item{epsilon, the base weight for each variable (epsilon=Inf corresponds to uniform sampling),}
\item{alpha, the exponent-learning rate,}
\item{num_passes a numerical parameter only.}
}

Sampling method for g0^2 depends on \code{g2_sample}
\describe{
\item{\code{"f"}}{Fixed: Do not update \code{g0^2}; carry forward the previous value (\code{g2[i] <- g2[i-1]}).}
\item{\code{"lf"}}{Laplace-Full: Use a Laplace approximation to sample \code{g0^2} under the full marginal likelihood (no orthogonality assumption).}
\item{\code{"lo"}}{Laplace-Orthogonal: Use a Laplace approximation to sample \code{g0^2} assuming the orthogonality approximation applies.}
\item{\code{"mh"}}{Metropolis-Hastings with full Laplace approximation as proposal. Target is the full posterior density of \code{g0^2}.}
\item{\code{"mho"}}{Metropolis-Hastings with orthogonal Laplace approximation as proposal. Target is the full posterior density of \code{g0^2}. under the orthogonality assumption.}
\item{\code{"mhoo"}}{Metropolis-Hastings with orthogonal Laplace approximation as proposal. Target is the posterior density of \code{g0^2} under the orthogonality assumption.}
}

Other proposal notes:
Expected order q0 is chosen from 1:order with weights \code{coin_pars[[1]](1:order)}
Degree is chosen from q0:degree with weights \code{(1:(degree-q0+1))^(-degree_penalty)} and a random partition is created (see helper function).

Two types of change steps are possible (and equally likely)
A re-ordering of the degrees
A single variable is swapped out with another one (only used when \code{ncol(X) > 3})
}
\examples{
\donttest{
X <- lhs::maximinLHS(100, 2)
f <- function(x) 10.391*((x[1]-0.4)*(x[2]-0.6) + 0.36)
y <- apply(X, 1, f) + stats::rnorm(100, 0, 0.1)
fit <- adaptive_khaos(X, y)
}
}
\references{
Francom, Devin, and Bruno Sanso. "BASS: An R package for fitting and performing sensitivity analysis of Bayesian adaptive spline surfaces." Journal of Statistical Software 94.LA-UR-20-23587 (2020).

Nott, David J., Anthony YC Kuk, and Hiep Duc. "Efficient sampling schemes for Bayesian MARS models with many predictors." Statistics and Computing 15 (2005): 93-101.

Rumsey, K.N., Francom, D., Gibson, G., Derek Tucker, J. and Huerta, G., 2026. Bayesian Adaptive Polynomial Chaos Expansions. Stat, 15(1), p.e70151.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/adaptive_chaos.R
\name{plot.adaptive_khaos}
\alias{plot.adaptive_khaos}
\title{Plot Method for class adaptive_khaos}
\usage{
\method{plot}{adaptive_khaos}(x, ...)
}
\arguments{
\item{x}{An object returned by the \code{adaptive_khaos()} function.}

\item{...}{Additional arguments for plotting}
}
\value{
No return value; called for its side effects (plotting).
}
\description{
See \code{adaptive_khaos()} for details.
}
\details{
Plot function for adaptive_khaos object.
}
\examples{
\donttest{
X <- lhs::maximinLHS(100, 2)
f <- function(x) 10.391*((x[1]-0.4)*(x[2]-0.6) + 0.36)
y <- apply(X, 1, f) + stats::rnorm(100, 0, 0.1)
fit <- adaptive_khaos(X, y)
plot(fit)
}
}
\references{
Francom, Devin, and Bruno Sanso. "BASS: An R package for fitting and performing sensitivity analysis of Bayesian adaptive spline surfaces." Journal of Statistical Software 94.LA-UR-20-23587 (2020).
}
